//============================================================================
// Author      : Sanjib Sharma
// Version     :
// Copyright   : Copyright (c) Sanjib Sharma 2010
//============================================================================
#ifndef FUNCTIONS_H_
#define FUNCTIONS_H_


#include<algorithm>
#include "sutils.h"

template<class T>
void printm(const char* str,T* matrix,int dim1,int dim2)
{
std::cout<<str<<std::endl;
for(int j=0;j<dim1;++j)
{
	for(int k=0;k<dim2;++k)
		std::cout<<matrix[j*dim1+k]<<" ";
	std::cout<<std::endl;
}
}

template<class T>
void printa(const char* str,T* array_begin,T* array_end)
{
	T* temp;
	std::cout<<str<<" ";
for( temp=array_begin;temp<array_end;temp++)
{
		std::cout<<(*temp)<<" ";
}
std::cout<<std::endl;
}

template<class T> struct index_cmp {
index_cmp(const T arr1) : arr(arr1) {}
bool operator()(const size_t a, const size_t b) const
{ return arr[a] < arr[b]; }
const T arr;
};


template <class T>
void sortIndices(T xb,T xe,std::vector<int> &ind)
{
	ind.clear();
	ind.resize(xe-xb,0);
	for(size_t i=0;i<ind.size();++i)
		ind[i]=i;
	sort(ind.begin(),ind.end(),index_cmp<T>(xb));
}

struct Ran {
	uint64_t u,v,w;
	Ran(uint64_t j) : v(4101842887655102017LL), w(1) {
		u = j ^ v; int64();
		v = u; int64();
		w = v; int64();
	}
	inline uint64_t int64() {
		u = u * 2862933555777941757LL + 7046029254386353087LL;
		v ^= v >> 17; v ^= v << 31; v ^= v >> 8;
		w = 4294957665U*(w & 0xffffffff) + (w >> 32);
		uint64_t x = u ^ (u << 21); x ^= x >> 35; x ^= x << 4;
		return (x + v) ^ w;
	}
	inline double doub() { return 5.42101086242752217E-20 * int64(); }
	inline uint32_t int32() { return (uint32_t)int64(); }
};
struct Normaldev : Ran {
	double mu,sig;
	Normaldev(double mmu, double ssig, uint64_t i)
	: Ran(i), mu(mmu), sig(ssig){}
	double dev() {
		double u,v,x,y,q;
		do {
			u = doub();
			v = 1.7156*(doub()-0.5);
			x = u - 0.449871;
			y = fabs(v) + 0.386595;
			q = x*x + y*(0.19600*y-0.25472*x);
		} while (q > 0.27597
			&& (q > 0.27846 || (v*v) > -4.*log(u)*(u*u)));
		return mu + sig*v/u;
	}
};





float gammln(float xx);
double volume_d(int dim);
float interpolate(float r,std::vector<float> &tablex,std::vector<float> &tabley);
template<class T>
void printv(T x1,T x2)
{
	for(T x=x1;x<x2;x++)
		std::cout<<(*x)<<" ";
	std::cout<<std::endl;
};

template<class T>
void interpolate(std::vector<T> &x,std::vector<T> &y,std::vector<T> &x1,std::vector<T> &y1)
{
	int p;
	y1.resize(x1.size());
	for(unsigned int i=0;i<x1.size();++i)
	{
		if(x1[i]>=x.back())
			y1[i]=y.back();
		else 		if(x1[i]<x.front())
			y1[i]=y.front();
		else
		{
			p=int(upper_bound(x.begin(),x.end(),x1[i])-x.begin()-1);
			y1[i]=y[p]+(y[p+1]-y[p])*(x1[i] -x[p])/(x[p+1]-x[p]);
		}
//		cout<<i<<" "<<x1[i]<<" "<<y1[i]<<" "<<x.front()<<" "<<x.back()<<" "<<y.front()<<" "<<y.back()<<endl;
	}
};

template<class T>
T interpolate(std::vector<T> &x,std::vector<T> &y, T x1)
{

	int p;
	if(x1>=x.back())
	{
		return y.back();
	}
	if(x1<x.front())
	{
		return y.front();
	}
	p=int(upper_bound(x.begin(),x.end(),x1)-x.begin()-1);
	return y[p]+(y[p+1]-y[p])*(x1 -x[p])/(x[p+1]-x[p]);
};


template<class T>
void mysort2(T arr, T arr1, T brr)
{
#define SWAP(a,b) temp=(a);(a)=(b);(b)=temp;
#define M 7
#define NSTACK 50
	uint64_t n=arr1-arr;
	arr=arr-1;
	brr=brr-1;

	uint64_t i,ir=n,j,k,l=1;
        int *istack,jstack=0;
        double a,b,temp;

        istack=new int[1+NSTACK];
        for (;;) {
                if (ir-l < M) {
                        for (j=l+1;j<=ir;j++) {
                                a=arr[j];
                                b=brr[j];
                                for (i=j-1;i>=l;i--) {
                                        if (arr[i] <= a) break;
                                        arr[i+1]=arr[i];
                                        brr[i+1]=brr[i];
                                }
                                arr[i+1]=a;
                                brr[i+1]=b;
                        }
                        if (!jstack) {
                                delete[] istack;
                                return;
                        }
                        ir=istack[jstack];
                        l=istack[jstack-1];
                        jstack -= 2;
                } else {
                        k=(l+ir) >> 1;
                        SWAP(arr[k],arr[l+1])
                        SWAP(brr[k],brr[l+1])
                        if (arr[l] > arr[ir]) {
                                SWAP(arr[l],arr[ir])
                                SWAP(brr[l],brr[ir])
                        }
                        if (arr[l+1] > arr[ir]) {
                                SWAP(arr[l+1],arr[ir])
                                SWAP(brr[l+1],brr[ir])
                        }
                        if (arr[l] > arr[l+1]) {
                                SWAP(arr[l],arr[l+1])
                                SWAP(brr[l],brr[l+1])
                        }
                        i=l+1;
                        j=ir;
                        a=arr[l+1];
                        b=brr[l+1];
                        for (;;) {
                                do i++; while (arr[i] < a);
                                do j--; while (arr[j] > a);
                                if (j < i) break;
                                SWAP(arr[i],arr[j])
                                SWAP(brr[i],brr[j])
                        }
                        arr[l+1]=arr[j];
                        arr[j]=a;
                        brr[l+1]=brr[j];
                        brr[j]=b;
                        jstack += 2;
                        if (jstack > NSTACK) exit(1);
                        if (ir-i+1 >= j-l) {
                                istack[jstack]=ir;
                                istack[jstack-1]=i;
                                ir=j-1;
                         } else {
                                 istack[jstack]=j-1;
                                 istack[jstack-1]=l;
                                 l=i;
                         }
                 }
         }
#undef M
#undef NSTACK
#undef SWAP
 }


template<class T,class T1>
void mysorti2(T arr, T arr1, T1 brr)
{
#define SWAP(a,b) temp=(a);(a)=(b);(b)=temp;
#define SWAPI(a1,b1) tempi=(a1);(a1)=(b1);(b1)=tempi;
#define M 7
#define NSTACK 50
	uint64_t n=arr1-arr;
	arr=arr-1;
	brr=brr-1;
	uint64_t i,ir=n,j,k,l=1;
        int *istack,jstack=0;
        double a,temp;
        int b,tempi;

        istack=new int[1+NSTACK];
        for (;;) {
                if (ir-l < M) {
                        for (j=l+1;j<=ir;j++) {
                                a=arr[j];
                                b=brr[j];
                                for (i=j-1;i>=l;i--) {
                                        if (arr[i] <= a) break;
                                        arr[i+1]=arr[i];
                                        brr[i+1]=brr[i];
                                }
                                arr[i+1]=a;
                                brr[i+1]=b;
                        }
                        if (!jstack) {
                                delete[] istack;
                                return;
                        }
                        ir=istack[jstack];
                        l=istack[jstack-1];
                        jstack -= 2;
                } else {
                        k=(l+ir) >> 1;
                        SWAP(arr[k],arr[l+1])
                        SWAPI(brr[k],brr[l+1])
                        if (arr[l] > arr[ir]) {
                                SWAP(arr[l],arr[ir])
                                SWAPI(brr[l],brr[ir])
                        }
                        if (arr[l+1] > arr[ir]) {
                                SWAP(arr[l+1],arr[ir])
                                SWAPI(brr[l+1],brr[ir])
                        }
                        if (arr[l] > arr[l+1]) {
                                SWAP(arr[l],arr[l+1])
                                SWAPI(brr[l],brr[l+1])
                        }
                        i=l+1;
                        j=ir;
                        a=arr[l+1];
                        b=brr[l+1];
                        for (;;) {
                                do i++; while (arr[i] < a);
                                do j--; while (arr[j] > a);
                                if (j < i) break;
                                SWAP(arr[i],arr[j])
                                SWAPI(brr[i],brr[j])
                        }
                        arr[l+1]=arr[j];
                        arr[j]=a;
                        brr[l+1]=brr[j];
                        brr[j]=b;
                        jstack += 2;
                        if (jstack > NSTACK) exit(1);
                        if (ir-i+1 >= j-l) {
                                istack[jstack]=ir;
                                istack[jstack-1]=i;
                                ir=j-1;
                         } else {
                                 istack[jstack]=j-1;
                                 istack[jstack-1]=l;
                                 l=i;
                         }
                 }
         }
#undef M
#undef NSTACK
#undef SWAP
 }



#endif /*FUNCTIONS_H_*/
