//============================================================================
// Author      : Sanjib Sharma
// Version     :
// Copyright   : Copyright (c) Sanjib Sharma 2010
//============================================================================

#ifndef MATRIX_H_
#define MATRIX_H_
#include<iostream>

template<class T>
class Matrix
{
private:
	int nrows;
	int ncols;
	T **v;
	inline void allocate(int nrows1,int ncols1);
	inline void deallocate()
	{
		if (v != NULL)
		{
			delete[] (v[0]);
			delete[] (v);
		}
	}
public:
	Matrix():nrows(0), ncols(0), v(NULL){}
	Matrix(int nrows1, int ncols1) {allocate(nrows1,ncols1);} // uninitialized
	Matrix(int n, int m, const T &a); // initialized to constant
	Matrix(int n, int m, const T *a); // initialize with an array
	Matrix(const Matrix &b); // Copy
	Matrix & operator=(const Matrix &b); //assignment
	inline T* operator[](const int i) {return v[i];} //subscript operator
	inline const T* operator[](const int i) const {return v[i];}
	inline int rows() const {return nrows;}
	inline int cols() const {return ncols;}
	void resize(int nrows1, int ncols1); // deletes content
	void assign(int nrow1, int ncol1, const T &a); // resize and assign a constant value
	inline void print()
	{
		for (int j = 0; j < nrows; ++j)
		{
			std::cout << j << " ";
			for (int k = 0; k < ncols; ++k)
				std::cout << v[j][k] << " ";
			std::cout << std::endl;
		}

	}
	inline void transpose()
	{
		T temp;
		for (int j = 0; j < nrows; ++j)
			for (int k = j; k < nrows; ++k)
			{
				if (j != k)
				{
					temp = v[k][j];
					v[k][j] = v[j][k];
					v[j][k] = temp;
				}
			}
	}
	template<class T1>
	void mult(T1 x, T1 y)
	{
		for (int j = 0; j < nrows; ++j)
		{
			y[j] = 0.0;
			for (int k = 0; k < ncols; ++k)
			{
				y[j] += v[j][k] * x[k];
			}
		}
	}
	template<class T1>
	void Tmult(T1 x, T1 y)
	{
		for (int j = 0; j < nrows; ++j)
		{
			y[j] = 0.0;
			for (int k = 0; k < ncols; ++k)
			{
				y[j] += v[k][j] * x[k];
			}
		}
	}
	~Matrix(){deallocate();}
};

//template<class T>
//Matrix<T>::Matrix() :
//	nrows(0), ncols(0), v(NULL)
//{
//}

template<class T>
inline void Matrix<T>::allocate(int nrows1, int ncols1)
{
	nrows=nrows1;
	ncols=ncols1;

	if(nrows > 0)
		v=new T*[nrows];
	else
		v=NULL;
	int elements = ncols * nrows;

	if (v!=NULL)
	{
		if (elements > 0)
			v[0] = new T[elements];
		else
			v[0] = NULL;
	}

	for (int i = 1; i < nrows; i++)
		v[i] = v[i - 1] + ncols;

}

template<class T>
Matrix<T>::Matrix(int nrows1, int ncols1, const T &a)
{
	allocate(nrows1,ncols1);
	for (int i = 0; i < nrows; i++)
		for (int j = 0; j < ncols; j++)
			v[i][j] = a;
}

template<class T>
Matrix<T>::Matrix(int nrows1, int ncols1, const T *a)
{
	allocate(nrows1,ncols1);
	for (int i = 0; i < nrows; i++)
		for (int j = 0; j < ncols; j++)
			v[i][j] = *a++;
}

template<class T>
Matrix<T>::Matrix(const Matrix &b)
{
	allocate(b.nrows,b.ncols);
	for (int i = 0; i < nrows; i++)
		for (int j = 0; j < ncols; j++)
			v[i][j] = b[i][j];
}

template<class T>
Matrix<T> & Matrix<T>::operator=(const Matrix<T> &b)
{
	if (this != &b)
	{
		resize(b.nrows,b.ncols);
		for (int i = 0; i < nrows; i++)
			for (int j = 0; j < ncols; j++)
				v[i][j] = b[i][j];
	}
	return *this;
}


template<class T>
void Matrix<T>::resize(int nrows1, int ncols1)
{
	if (nrows1 != nrows || ncols1 != ncols)
	{
		deallocate();
		allocate(nrows1,ncols1);
	}
}

template<class T>
void Matrix<T>::assign(int nrows1, int ncols1, const T& a)
{
	resize(nrows1,ncols1);
	for (int i = 0; i < nrows; i++)
		for (int j = 0; j < ncols; j++)
			v[i][j] = a;
}


#endif /* MATRIX_H_ */
