//============================================================================
// Author      : Sanjib Sharma
// Version     :
// Copyright   : Copyright (c) Sanjib Sharma 2010
//============================================================================
#ifndef PARTICLE_H_
#define PARTICLE_H_

#include "sutils.h"
#include "Matrix.h"

class Potmodel
{
public:
	Potmodel()
	{
		c=40.0;
		Mhalo=1.4e2/(log(c+1)-c/(c+1));
		Rhalo=282.0/c;
		Mdisk=1.0e1;
		Rdisk=6.5;
		Zdisk=0.26;
		Msphere=3.4;
		Rsphere=0.7;
		G=43007.1;
	}
	~Potmodel()
	{

	}
	//	double energy(double* pos,double* vel)
	template<class T>
	double energy(T pos,T vel)
	{
	    rc=sqrt(pos[0]*pos[0]+pos[1]*pos[1]);
	    r=sqrt(pos[0]*pos[0]+pos[1]*pos[1]+pos[2]*pos[2]);
	    double temp=(Rdisk+sqrt(pos[2]*pos[2]+Zdisk*Zdisk));
	    temp=sqrt(rc*rc+temp*temp);
	    phi=-(Mhalo/r)*log(1+r/Rhalo)-(Msphere)/(r+Rsphere)-(Mdisk)/(temp);
	    return G*phi+0.5*(vel[0]*vel[0]+vel[1]*vel[1]+vel[2]*vel[2]);
	//	return 0.0;
	}
private:
	double c,Mhalo,Rhalo,Mdisk,Rdisk,Zdisk,Msphere,Rsphere,G;
	double r,rc,phi;
};


class Particle
{
public:
	Particle();
	~Particle();
    double* Pos;       // particle position
    inline double& Mass(int k) {return Pos[k-3];}
    inline double& Density(int k) {return Pos[k-2];}
    inline int64_t& ID(int k) {return *(int64_t *)&Pos[k-1];}
//    double distance(const double* xyz,const double* metric,int dim);
    inline double distance(const double* xyz,const double* metric,int dim)
    {
    	double r=0.0;
    	for(int j=0;j<dim;++j)
    		r+=(Pos[j]-xyz[j])*(Pos[j]-xyz[j])*metric[j]*metric[j];
    	return r;
    }
    inline double distance(const double* xyz,const Matrix<double>& mt,int dim)
    {
    	double r=0.0;
    	double temp=0.0;
    	for(int j=0;j<dim;++j)
    	{
    		temp=(Pos[0]-xyz[0])*mt[j][0];
    		for(int k=1;k<dim;++k)
    			temp+=(Pos[k]-xyz[k])*mt[j][k];
    		r+=temp*temp;
    	}
    	return r;
    }
//    double distance_pq(const double* xyz,const double* metric,int dim,double pqR);
    inline double periodic(double temp,double l)
    {
    	if((2*temp)>l)
    		temp-=l;
    	else if((2*temp)<(-l))
    		temp+=l;
    	return temp;
    }

    inline  double distance_pq(const double* xyz,const double* metric,int dim,double pqR,double* lp)
    {

    	double temp=periodic(Pos[0]-xyz[0],lp[0])*metric[0];
		double r=temp*temp;
		for(int j=1;j<dim;++j)
    	{
    		if(r>pqR)
    			break;
    		temp=periodic(Pos[j]-xyz[j],lp[j])*metric[j];
    		r+=temp*temp;
    	}
    	return r;
    }

    inline  double distance_pq(const double* xyz,const double* metric,int dim,double pqR)
    {
//    	double r=0.0;
//    	double temp=(Pos[0]-xyz[0])*metric[0];
//    		r+=temp*temp;
//    		if(r>pqR)
//    			return r;
//    		temp=(Pos[1]-xyz[1])*metric[1];
//    		r+=temp*temp;
//    		if(r>pqR)
//    			return r;
//    		temp=(Pos[2]-xyz[2])*metric[2];
//    		r+=temp*temp;
//    			return r;

    	double temp=(Pos[0]-xyz[0])*metric[0];
		double r=temp*temp;
		for(int j=1;j<dim;++j)
    	{
    		if(r>pqR)
    			break;
    		temp=(Pos[j]-xyz[j])*metric[j];
    		r+=temp*temp;
    	}
    	return r;

//    	    	double temp=(Pos[0]-xyz[0]);
//    	    	double	r=temp*temp;
//    	    		if(r>pqR)
//    	    			return r;
//    	    		temp=(Pos[1]-xyz[1]);
//    	    		r+=temp*temp;
//    	    		if((r>pqR)||(dim==2))
//    	    			return r;
 //    	    		temp=(Pos[2]-xyz[2]);
//    	    		r+=temp*temp;
//    	    		if((r>pqR)||(dim==3))
//    	    			return r;
//

    }
    inline  double distance_pq(const double* xyz,const Matrix<double>& mt,int dim,double pqR)
    {
    	double r=0.0;
    	double temp=0.0;
    	for(int j=0;j<dim;++j)
    	{
    		if(r>pqR)
    			break;
    		temp=(Pos[0]-xyz[0])*mt[j][0];
    		for(int k=1;k<dim;++k)
    			temp+=(Pos[k]-xyz[k])*mt[j][k];
    		r+=temp*temp;
    	}
    	return r;

    }

    //	inline  int  pID (Particle *pFirst )  const {return (pLast-pFirst);};
};



class data_less_than: std::unary_function<class Particle, bool>
{
	int kdim;
	double val;
public:
	explicit data_less_than(const int& kdim1,const double& val1):kdim(kdim1), val(val1){}
//	explicit data_less_than(const int& kdim1,const double& val1):kdim(kdim1){}
	bool operator() (const Particle& Pa) const {return Pa.Pos[kdim]<val;}
};

class par_less_than: std::binary_function<Particle, Particle,bool>
{
	int kdim;
public:
	explicit par_less_than(const int& kdim1):kdim(kdim1){}
//	explicit data_less_than(const int& kdim1,const double& val1):kdim(kdim1){}
	bool operator() (const Particle& Pa1,const Particle& Pa2) const {return Pa1.Pos[kdim]<Pa2.Pos[kdim];}
};



class ParticleTag
{
public:
	ParticleTag(std::string name1,int dim1,int id1,int datatype1,int status1):name(name1),dim(dim1),id(id1), datatype(datatype1),status(status1){}
	std::string name;
	int dim;
	int id;
	int datatype;
	int status;
	bool operator== (const ParticleTag& a)
	{
		return (a.name==name);
	}
};

class ParticleTag_eq: public std::unary_function<ParticleTag,bool>
{
	std::string s;
	int status;
public:
	explicit ParticleTag_eq(const std::string& ss,const int &st): s(ss), status(st){}
	bool operator() (const ParticleTag &t) const {return ((t.name==s)&&(t.status==status));}
};


class ParticleData
{
public:
	ParticleData();
	ParticleData(const char* fname,int dim1,const char *s);
	~ParticleData();
    void readFromFile(const char *fname,int dim1,const char *s);
    void AddError(const char* fname,int ErrorOption,const char *s,const char* psuffix1,double mu0,double sigma_mu,double sigma_vlb,double sigma_vr,double sigma_r);
//    void readGadgetFile(const char *fname);
    void readAsciiFile(const char *fname);
    void readCustomFile(const  char *fname);
    void readEbfFile(const  char *fname);
    void initialize(int size1,int option);
    void setTags(int option);
    inline Particle& operator[] (unsigned i) const {return part[i];}
    inline int size()  const {return nsize;}
    inline Particle* begin() const {return &part[0];}
    inline Particle* end() const {return &part[nsize];}
    int64_t posid(int64_t i) {return (part[i].Pos-data)/items;}
    void kdOrder();
//    void convert(double *PosC);
    int convert_p3v3(int dim1,double mu0);
    void print();
    std::vector<int> npart,npartc;
    double posC[6];
    int nsize,dim,items;
    Particle* part;
    std::vector<ParticleTag> tags;
    std::vector<std::string> sv;
private:
    void read_tag(const char *fname,ParticleTag &tag, int n_size);
	double* data;
	Potmodel BJmodel;
};


//class Data
//{
//public:
//	Data(int size);
//	~Data();
//    std::vector<int> Part;       // particle position
//    void readFromFile();
//};



#endif /*PARTICLE_H_*/
