//============================================================================
// Author      : Sanjib Sharma
// Version     :
// Copyright   : Copyright (c) Sanjib Sharma 2010
//============================================================================

#ifndef SPLITTER_H_
#define SPLITTER_H_

//#include "Particle.h"
#include "Functions.h"
#include "Mat.h"

class Subspace
{
public:
	Subspace()
	{

	}
	~Subspace();
	template<class T>
	Subspace(T d1,T d2)
	{
		d.resize(d2-d1);
		enC.resize(d2-d1,0.0);
		enS=0.0;
		for(T d3=d1;d3<d2;++d3)
		{
			d[d3-d1]=*d3;
		}
	}
	template<class T>
	Subspace(std::vector<T> &d1)
	{
		d=d1;
		enC.resize(d1.size(),0.0);
		enS=0.0;
	}
	Subspace(int dim1)
	{
		d.resize(dim1);
		enC.resize(dim1,0.0);
		enS=0.0;
		for(int i=0;i<dim1;++i)
			d[i]=i;
	}
	template<class T>
	void set(T d1,T d2)
	{
		d.resize(d2-d1);
		enC.resize(d2-d1,0.0);
		enS=0.0;
		for(T d3=d1;d3<d2;++d3)
		{
			d[d3-d1]=*d3;
		}
	}
	template<class T>
	void set(std::vector<T> &d1)
	{
		d=d1;
		enC.resize(d1.size(),0.0);
		enS=0.0;
	}
	void shuffle(std::vector<int> &d1)
	{
		for(size_t i=0;i<d.size();++i)
		{
			d[i]=d1[d[i]];
		}

	}
	void set(int dim1)
	{
		d.resize(dim1);
		enC.resize(dim1,0.0);
		enS=0.0;
		for(int i=0;i<dim1;++i)
			d[i]=i;
	}
	std::vector<int> d;
	double enS;
	std::vector<double> enC;
	bool operator<( const Subspace &s ) const    {	return enS < s.enS;    }
	void print()
	{
		std::cout<<"Subspace En="<<enS<<" dim="<<d.size()<<std::endl;
//		for (int i = 0; i < d.size(); ++i)
//					cout<<d[i]<<" ";
//		cout<<endl;;
		for (int i = 0; i < int(d.size()); ++i)
			std::cout<<d[i]<<" "<<enC[i]<<" "<<std::endl;;
	}
};


class ParticleStats1
{
public:
	ParticleStats1(int dim1);
	~ParticleStats1();
	void print();
	void computeMinMaxSig(Particle *pFirst,Particle *pLast,int jbegin,int jend);
	void computeCovar(Particle *pFirst,Particle *pLast);
//	void computeAll(Particle *pFirst,Particle *pLast);
	std::vector<double>	xmin,xmax,xmean,xsig,covar;
	double**	 covar2;
	int dim;
};

class Splitter:public ParticleStats1
{
public:
Splitter(int dim1,int maxPart1,int treeType1,Particle* pRoot1,int items1):	ParticleStats1(dim1),mt(dim1,0),randomS(17l)
	{
	treeType=treeType1;
	maxPart=maxPart1;
	partBoundary=7;
	epsilon=(10*std::numeric_limits<double>::epsilon());
	pRoot=pRoot1;
	items=items1;
	rotationOn=0;
	}
	virtual ~Splitter();
	virtual void computeSplitParameters(Particle *pFirst,Particle *pLast,double* bnd,int prevK);
	void setRotation(){rotationOn=1;}
	void save(Particle *pFirst,Particle *pLast);
	void load(Particle *pFirst,Particle *pLast);
//	void print();
	void check_identical_coord(Particle *pFirst,Particle *pLast,int prevK)
	{
		int j;
		double temp=(std::max(fabs(xmin[kSplit]),fabs(xmax[kSplit])));
		if(temp>1e-5)
			temp=fabs(xmax[kSplit]-xmin[kSplit])/temp;
		if(temp<1e-4)
		{
//			cout<<"correcting "<<" "<<temp<<" "<<kSplit<<" "<<xSplit<<" "<<xmax[kSplit]<<" "<<xmin[kSplit]<<" "<<pLast-pFirst<<endl;
			for(int k=0;k<dim;++k)
			{
				j=(prevK+k+1)%dim;
				temp=(std::max(fabs(xmin[j]),fabs(xmax[j])));
				if(temp>1e-5)
					temp=fabs(xmax[j]-xmin[j])/temp;
				if(temp>1e-4)
				{
				   kSplit=j;
				   xSplit=computexSplit(pFirst,pLast);
//				   cout<<"new val"<<" "<<temp<<" "<<kSplit<<" "<<xSplit<<" "<<pLast-pFirst<<endl;
				   break;
				}

			}
		}
//		else
//		{
//			cout<<"No correction needed"<<endl;
////			exit(1);
//		}
	}
	Particle* pRoot;
	double xSplit;
	int kSplit,rotationOn;
	Mat mt;
	Ran randomS;
protected:
	int maxPart,treeType,partBoundary,items;
	float epsilon;
	std::vector<int> rho;
	std::vector<double> dx;
	std::vector<double> en;
	std::vector<float> Pdat;
	void correctBoundary(double* bnd,int count);
	void correctParticles(Particle *pFirst,Particle *pLast,double* bnd);
	double computexSplit(Particle *pFirst,Particle *pLast);
	void compute1dEntropy(Particle *pFirst,Particle *pLast,double* bnd);
};

//class Entropy1d:public Splitter
//{
//public:
//	Entropy1d(int dim1,int maxPart1,int treeType1):Splitter(dim1,maxPart1,treeType1)
//	{
//		dx.resize(dim1,0.0);
//		rho.resize(maxPart1);
//		en.resize(dim1);
//	}
//	~Entropy1d();
//	void compute1dEntropy(Particle *pFirst,Particle *pLast,double* bnd);
//};

class SplitterEnCubic:public Splitter
{
public:
	SplitterEnCubic(int dim1,int maxPart1,int treeType1,Particle* pRoot1,int items1):Splitter(dim1,maxPart1,treeType1,pRoot1,items1)
	{
		dx.resize(dim1,0.0);
		rho.resize(maxPart1);
		enSub.resize(dim1);
	}
	~SplitterEnCubic();
	virtual void computeSplitParameters(Particle *pFirst,Particle *pLast,double* bnd,int prevK);
private:
	std::vector<double> enSub;
	void compute3dEntropy(Particle *pFirst,Particle *pLast,double* bnd);
};


class SplitterEn:public Splitter
{
public:
	SplitterEn(int dim1,int maxPart1,int treeType1,Particle* pRoot1,int items1):Splitter(dim1,maxPart1,treeType1,pRoot1,items1)
	{
		dx.resize(dim1,0.0);
		rho.resize(maxPart1);
		en.resize(dim1);
	}
	~SplitterEn();
	virtual void computeSplitParameters(Particle *pFirst,Particle *pLast,double* bnd,int prevK);
};

class SplitterEnC:public Splitter
{
public:
	SplitterEnC(int dim1,int maxPart1,int treeType1,Particle* pRoot1,int items1):Splitter(dim1,maxPart1,treeType1,pRoot1,items1)
	{
		dx.resize(dim1,0.0);
		rho.resize(maxPart1);
		en.resize(dim1);
		status_peak=0;
		troughX.resize(dim1);
		troughF.resize(dim1);
		enC.resize(dim1);
		x.resize(maxPart1);
		ind.resize(maxPart1);
		le=new int* [maxPart1];
		for (int i=0; i<maxPart1; i++)
			le[i]=new int [dim1];
	}
	~SplitterEnC();
	virtual void computeSplitParameters(Particle *pFirst,Particle *pLast,double* bnd,int prevK);
private:
	std::vector<double> enC;
	std::vector<double> x;
	std::vector<int> ind;
	int **le;
	std::vector<double> troughX;
	std::vector<double> troughF;
	int status_peak;
	void calculate_peak_split(std::vector<double> &x, double &e, double &ksplit,int option);
	void computeComplEntropy(Particle *pFirst,Particle *pLast,double* bnd);
};

class Trough
{
public:
	Trough();
	~Trough();
	void set(int k1,int binsize1,std::vector<double> &tempx);
	double sigl,sigr,denl0,denl1,denl2,denr0,denr1,denr2,xSplit;
	int kSplit,statusl,statusr,il0,il1,il2,ir0,ir1,ir2,binsize,tempx_size;
	void calculate(std::vector<double> &tempx,int &status1);
	double get_xSplit(std::vector<double> &tempx);
	void print(std::vector<double> &tempx)
	{
//		cout<<"Trough "<<kSplit<<" "<<xSplit<<" "<<tempx_size<<endl;
//		cout<<"left "<<statusl<<" "<<sigl<<" "<<il0<<" "<<il1<<" "<<il0<<" "<<il2<<" "<<tempx[il0]<<endl;
//		cout<<denl0<<" "<<denl1<<" "<<denl2<<endl;
//		cout<<"right "<<statusr<<" "<<sigr<<" "<<tempx_size-ir0<<" "<<ir1<<" "<<ir0<<" "<<ir2<<" "<<tempx[ir0]<<endl;
//		cout<<denr1<<" "<<denr0<<" "<<denr2<<endl;
	}
	bool operator<( const Trough &a ) const   {	return std::max(il0*statusl,(tempx_size-ir0)*statusr) < std::max(a.il0*a.statusl,(a.tempx_size-a.ir0)*a.statusr);    }
};


class SplitterEnSub:public Splitter
{
public:
	SplitterEnSub(int dim1,int maxPart1,int treeType1,Particle* pRoot1,int items1):Splitter(dim1,maxPart1,treeType1,pRoot1,items1)
	{
		dx.resize(dim1,0.0);
		rho.resize(maxPart1);
		en.resize(dim1);
		tr.resize(dim1);
		pind.resize(maxPart1);
		if(dim>2)
		{
			generate_subs_all(sAll,dim1);
			shuffle_subs_all(sAll,dim1);
		}
//		pind1.resize(maxPart1);
		pln.resize(dim1*maxPart1);
		templ.resize(maxPart1);
		tempx.resize(maxPart1);
	}
	~SplitterEnSub();
private:
	int status_peak;
	virtual void computeSplitParameters(Particle *pFirst,Particle *pLast,double* bnd,int prevK);
	void calculate_peak_split(std::vector<double> &tempx,int j);
		std::vector<std::vector<Particle> >  pl;
		std::vector<int>  templ;
		std::vector<double>  tempx;
		std::vector<int>  pln;
		std::vector<int>  pind;
		std::vector<std::vector<Subspace> > sAll;
		std::vector<Subspace> Sv;
		std::vector<Trough> tr;
		void identifySubspace(Particle *pFirst,Particle *pLast,int prevK);
		void compute_enS(Subspace &S,Particle *pFirst,Particle *pLast);
		void compute_enC(Subspace &S,Particle *pFirst,Particle *pLast);
		int next_comb(std::vector<int> &comb, int n);
		void correct_odd(Subspace &s,int n);
		void generate_subs_k(std::vector<Subspace> &sdv, int n,int kmax);
		void generate_subs_all(std::vector<std::vector<Subspace> > &sv, int n);
		void shuffle_subs_all(std::vector<std::vector<Subspace> > &sv,int n);
		void update_subspace(std::vector<Subspace> &sv,Subspace &S);
		void swap_pind(int d_old,int d_new,int count,int bins_n,int ib);
		void calculate_peak_split_all(Particle *pFirst,Particle *pLast,int prevK);
};


#endif /* SPLITTER_H_ */
