//============================================================================
// Author      : Sanjib Sharma
// Version     :
// Copyright   : Copyright (c) Sanjib Sharma 2010
//============================================================================
#ifndef TREE_H_
#define TREE_H_

#include<stack>
#include<queue>
#include "Node.h"
#include "NodeA.h"
#include "Functions.h"
#include "ParticlePriorityQueue.h"
#include "ebfvector.hpp"
#include "Parameters.h"
#include "Image.h"

template<class T>
void printv(const char* s,T x1,T x2)
{
	std::cout<<s;
	for(T x=x1;x<x2;x++)
		std::cout<<(*x)<<" ";
	std::cout<<std::endl;
};


template<class T>
void make_tree(Particle* pBegin1,Particle* pEnd1,T*& root1,int leafSize,int dim1,int treeType,int items)
{

	T *currentNode, *lastCreatedNode;

	//	ParticleStats pStat(dim1, (pEnd1-pBegin1),treeType);
//	ParticleStats *pStatp;
//		pStatp= new ParticleStats(dim1, (pEnd1-pBegin1),treeType);
	Splitter *pStatp=NULL;
	if(treeType<2)
		pStatp= new Splitter(dim1, (pEnd1-pBegin1),treeType,pBegin1,items);
	if(treeType==2)
		pStatp= new SplitterEn(dim1, (pEnd1-pBegin1),treeType,pBegin1,items);
	if(treeType==3)
		pStatp= new SplitterEnCubic(dim1, (pEnd1-pBegin1),treeType,pBegin1,items);
//	if(treeType==4)
//		pStatp= new SplitterEnC(dim1, (pEnd1-pBegin1),treeType,pBegin1,items);
	if(treeType==4)
		pStatp= new SplitterEnSub(dim1, (pEnd1-pBegin1),treeType,pBegin1,items);

	pStatp->computeMinMaxSig(pBegin1,pEnd1,0,dim1);
//	pStatp->print();

	//	std::cout<<"Building Tree ....";  std::cout.flush();
	int maxNodes=2*(pEnd1-pBegin1);
	int nodesCreated=1;
	if(leafSize>4)
		maxNodes=2*(pEnd1-pBegin1)*2/(leafSize/2);
//	int maxNodes=2*(pEnd1-pBegin1);
//	root1= new T[maxNodes]; //for np particles nodes needed is 2*np-1 we keep one extra

//  Try this experimental
//	while(nodesCreated<=maxNodes)
//	{
//	nodesCreated=1;
//	if (root1 != NULL)
//	{
//		if (root1->type() == 0)
//		{
//			if (root1->bnd != NULL)
//				free(root1->bnd);
//		}
//		free(root1);
//		root1=NULL;
//	}

	root1= (T*) malloc(maxNodes*sizeof(T)); //for np particles nodes needed is 2*np-1 we keep one extra
	root1->setUpRoot(pBegin1,pEnd1,*pStatp,maxNodes);
	lastCreatedNode=root1;
//	std::cout<<"Max Nodes="<<maxNodes<<" "<<sizeof(root1[0])<<std::endl;
	std::deque<T*> nodeStack;
	nodeStack.push_back(root1);
//	int lastLevel=0;
	while (nodeStack.size()>0)
	{
//		if((lastCreatedNode-root1)>128)
//			exit(1);

		//		if((lastCreatedNode-root1)%1000==0)
		currentNode=nodeStack.front();
//		if(currentNode->level>lastLevel)

//		currentNode=nodeStack.back();
//		std::cout<<" Here "<<(lastCreatedNode-root1)<<" "<<2*(pEnd1-pBegin1)<<" "<<currentNode->count()<<std::endl;
		if ((currentNode->pLast-currentNode->pFirst)>leafSize)
		{
			nodesCreated+=2;
			if((lastCreatedNode+2)>=(root1+maxNodes))
			{
				std::cout<<"No nodes too large"<<std::endl;
// try this experimental
//				maxNodes*=2;
//               continue;
				exit(1);
			}
			currentNode->setChildren(lastCreatedNode+1);
//			currentNode->left=lastCreatedNode+1;
			lastCreatedNode+=2;
			currentNode->split(*pStatp);
//			currentNode->print(root1,pBegin1);
//			pStatp->print();

		}
		else
		{
			currentNode->setChildren(NULL);
//			currentNode->left=NULL;
		}

//		nodeStack.pop_back();
//		if (currentNode->leftN()!=NULL)
//			nodeStack.push_back(currentNode->rightN());
//		if (currentNode->leftN()!=NULL)
//			nodeStack.push_back(currentNode->leftN());
		nodeStack.pop_front();
		if (currentNode->leftN()!=NULL)
			nodeStack.push_back(currentNode->leftN());
		if (currentNode->leftN()!=NULL)
			nodeStack.push_back(currentNode->rightN());
	}

// try this experimental
//}

//	std::cout<<"Node container initial capacity="<<maxNodes<<" final size="<<lastCreatedNode-root1+1<<std::endl;

	T * rootP=root1;
	double* bndP=root1->bnd;
	maxNodes=lastCreatedNode-root1+1;
	root1=(T *) realloc(root1,sizeof(T)*maxNodes);
	std::cout<<std::setw(16)<<"TreeNodes Size"<<maxNodes*sizeof(T)/1e6<<" MB ("<<sizeof(T)<<"x"<<maxNodes<<")"<<std::endl;
	root1->bnd=(double *) realloc(root1->bnd,sizeof(double)*maxNodes*2*pStatp->dim);
	if(rootP!=root1)
	for(int i=0;i<maxNodes;++i)
	{
		root1[i].update_node_realloc(rootP,root1);
	}
	if(bndP!=root1->bnd)
	for(int i=1;i<maxNodes;++i)
	{
		root1[i].update_bnd_realloc(bndP,root1->bnd);
	}
//	std::cout<<"Realloc stat"<<rootP<<" "<<root1<<" "<<bndP<<" "<<bndP<<" "<<std::endl;

	delete pStatp;
	//	std::cout<<" Done"<<std::endl;
}



template<class T>
void make_tree_bt(Particle* pBegin1,Particle* pEnd1,Particle* pBegin2,Particle* pEnd2,T*& root1,int leafSize,int dim1,int treeType,int items,int rotationOn)
{
//	std::cout<<"here0"<<std::endl;

	T *currentNode, *lastCreatedNode;

	Splitter *pStatp=NULL;
	if(treeType<2)
		pStatp= new Splitter(dim1, (pEnd2-pBegin2),treeType,pBegin1,items);
	if(treeType==2)
		pStatp= new SplitterEn(dim1, (pEnd2-pBegin2),treeType,pBegin1,items);
	if(treeType==3)
		pStatp= new SplitterEnCubic(dim1, (pEnd2-pBegin2),treeType,pBegin1,items);
//	if(treeType==4)
//		pStatp= new SplitterEnC(dim1, (pEnd2-pBegin2),treeType,pBegin1,items);
	if(treeType==4)
		pStatp= new SplitterEnSub(dim1, (pEnd2-pBegin2),treeType,pBegin1,items);

	pStatp->computeMinMaxSig(pBegin1,pEnd1,0,dim1);
	if(rotationOn==1)
	{
		pStatp->setRotation();
		pStatp->save(pBegin2,pEnd2);
	}

//	std::cout<<"here1"<<std::endl;
//	pStatp->print();

//	ParticleStats pStat(dim1, (pEnd1-pBegin1),treeType);
	//	std::cout<<"Building Tree ....";  std::cout.flush();
	int maxNodes=2*(pEnd1-pBegin1);
//	if(leafSize>4)
//		maxNodes=2*(pEnd1-pBegin1)*2/leafSize;
	root1= (T*) malloc(maxNodes*sizeof(T)); //for np particles nodes needed is 2*np-1 we keep one extra
//	root1= new T[maxNodes]; //for np particles nodes needed is 2*np-1 we keep one extra
	root1->setUpRoot(pBegin1,pEnd1,pBegin2,pEnd2,*pStatp);
	lastCreatedNode=root1;

//	std::cout<<"here2"<<std::endl;

	std::deque<T*> nodeStack;
	nodeStack.push_back(root1);
	while (nodeStack.size()>0)
	{
//		if((lastCreatedNode-root1)%1000==0)
//			std::cout<<" Here "<<(lastCreatedNode-root1)<<" "<<2*(pEnd1-pBegin1)<<std::endl;
//				if((lastCreatedNode-root1)>100)
//					exit(1);

		currentNode=nodeStack.front();
		if ((currentNode->pLast-currentNode->pFirst)>leafSize)
		{
			if((lastCreatedNode+2)>=(root1+maxNodes))
			{
				std::cout<<"No nodes too large"<<std::endl;
				exit(1);
			}
			currentNode->setChildren(lastCreatedNode+1);
//			currentNode->left=lastCreatedNode+1;
			lastCreatedNode+=2;
			currentNode->split(*pStatp,leafSize);
		}
		else
		{
			currentNode->setChildren(NULL);
//			currentNode->left=NULL;
		}
		nodeStack.pop_front();
		if (currentNode->leftN()!=NULL)
			nodeStack.push_back(currentNode->leftN());
		if (currentNode->leftN()!=NULL)
			nodeStack.push_back(currentNode->rightN());

//		if((lastCreatedNode-root1)<10)
//			std::cout<<currentNode->count()<<" "<<pStatp->kSplit<<" "<<pStatp->xSplit<<std::endl;

	}
	if(rotationOn==1)
		pStatp->load(pBegin2,pEnd2);


	delete pStatp;

//	std::cout<<"Node container initial capacity="<<maxNodes<<" final size="<<lastCreatedNode-root1+1<<std::endl;

	T * rootP=root1;
	maxNodes=lastCreatedNode-root1+1;
	root1=(T *) realloc(root1,sizeof(T)*maxNodes);


	if(rootP!=root1)
	for(int i=0;i<maxNodes;++i)
	{
		root1[i].update_node_realloc(rootP,root1);
	}


//	std::cout<<"Realloc stat"<<rootP<<" "<<root1<<std::endl;



	//	std::cout<<" Done"<<std::endl;
}


template<class T>
void make_tree_bt_par(Particle* pBegin1, Particle* pEnd1, Particle* pBegin2,
		Particle* pEnd2, T*& root1, int leafSize, int dim1, int treeType,
		int items, int rotationOn)
{
//	T *currentNode, *lastCreatedNode;
//	Splitter *pStatp = NULL;
//	if (treeType < 2)
//		pStatp
//				= new Splitter(dim1, (pEnd2 - pBegin2), treeType, pBegin1, items);
//	if (treeType == 2)
//		pStatp
//				= new SplitterEn(dim1, (pEnd2 - pBegin2), treeType, pBegin1, items);
//	if (treeType == 3)
//		pStatp
//				= new SplitterEnCubic(dim1, (pEnd2 - pBegin2), treeType, pBegin1, items);
//	if (treeType == 4)
//		pStatp
//				= new SplitterEnSub(dim1, (pEnd2 - pBegin2), treeType, pBegin1, items);
//
//	pStatp->computeMinMaxSig(pBegin1, pEnd1, 0, dim1);
//	if (rotationOn == 1)
//	{
//		pStatp->setRotation();
//		pStatp->save(pBegin2, pEnd2);
//	}
//
//	int maxNodes = 2 * (pEnd1 - pBegin1);
//	root1 = (T*) malloc(maxNodes * sizeof(T)); //for np particles nodes needed is 2*np-1 we keep one extra
//	root1->setUpRoot(pBegin1, pEnd1, pBegin2, pEnd2, *pStatp);
//	lastCreatedNode = root1;
//
//	deque<T*> nodeStack;
//	nodeStack.push_back(root1);
//	while (nodeStack.size() > 0)
//	{
//		if (nodeStack.size() >64)
//			break;
//		currentNode = nodeStack.front();
//		if ((currentNode->pLast - currentNode->pFirst) > leafSize)
//		{
//			if ((lastCreatedNode + 2) >= (root1 + maxNodes))
//			{
//				std::cout << "No nodes too large" << std::endl;
//				exit(1);
//			}
//			currentNode->setChildren(lastCreatedNode + 1);
//			lastCreatedNode += 2;
//			currentNode->split(*pStatp, leafSize);
//		}
//		else
//		{
//			currentNode->setChildren(NULL);
//		}
//		nodeStack.pop_front();
//		if (currentNode->leftN() != NULL)
//			nodeStack.push_back(currentNode->leftN());
//		if (currentNode->leftN() != NULL)
//			nodeStack.push_back(currentNode->rightN());
//
//
//	}
//	delete pStatp;
//
//	std::vector<T*> lastCreatedNode1(nodeStack.size());
//	lastCreatedNode1[0] = lastCreatedNode;
//
//	for (int j = 1; j < int(nodeStack.size()); ++j)
//		lastCreatedNode1[j] = lastCreatedNode1[j - 1] + (2* nodeStack[j - 1]->count() - 2);
//
//	int i;
//	#pragma omp parallel for default(none) private(i) shared(nodeStack,lastCreatedNode1,dim1,treeType,items,leafSize,maxNodes,rotationOn)
//	for (i = 0; i < int(nodeStack.size()); ++i)
//	{
//		T *currentNode1;
//		T *cn = nodeStack[i];
//		Splitter *pStatp1 = NULL;
//		if (treeType < 2)
//			pStatp1
//					= new Splitter(dim1, (cn->pLast2 - cn->pFirst2), treeType, cn->pFirst, items);
//		if (treeType == 2)
//			pStatp1
//					= new SplitterEn(dim1, (cn->pLast2 - cn->pFirst2), treeType, cn->pFirst, items);
//		if (treeType == 3)
//			pStatp1
//					= new SplitterEnCubic(dim1, (cn->pLast2 - cn->pFirst2), treeType, cn->pFirst, items);
//		if (treeType == 4)
//			pStatp1
//					= new SplitterEnSub(dim1, (cn->pLast2 - cn->pFirst2), treeType, cn->pFirst, items);
//
//		pStatp1->computeMinMaxSig(cn->pFirst, cn->pLast, 0, dim1);
//
//		deque<T*> nodeStack1;
//		nodeStack1.push_back(cn);
//		while (nodeStack1.size() > 0)
//		{
//			currentNode1 = nodeStack1.front();
//			if ((currentNode1->pLast - currentNode1->pFirst) > leafSize)
//			{
//				if ((lastCreatedNode1[i] + 2) >= (root1 + maxNodes))
//				{
//					exit(1);
//				}
//				currentNode1->setChildren(lastCreatedNode1[i] + 1);
//				lastCreatedNode1[i] += 2;
//				currentNode1->split(*pStatp1, leafSize);
//			}
//			else
//			{
//				currentNode1->setChildren(NULL);
//			}
//			nodeStack1.pop_front();
//			if (currentNode1->leftN() != NULL)
//				nodeStack1.push_back(currentNode1->leftN());
//			if (currentNode1->leftN() != NULL)
//				nodeStack1.push_back(currentNode1->rightN());
//		}
//		delete pStatp1;
//
//	}

}


void calculate_metric(const char* fname,Particle* pFirst2, Particle* pLast2,int leafSize,int treeType,int rotationOn,int dim,int items);

class node_list
{
public:
	inline void initialize (Node* cp1,double r) {cp=cp1; r2=r;}
	Node* cp;
	double r2;
	bool operator<(const class node_list &a) const
	{
		return r2 > a.r2;
	}
	bool operator>(const class node_list &a) const
	{
		return r2 < a.r2;
	}
};

class pqStats
{
public:
	void initialize(int dim);
	std::vector<double> xmin,xmax,metric;
	std::vector<std::vector<double> > searchx;
	std::vector<std::vector<double> > searchcenter;
};


class pinfo
{
    public:
    int pi;
    float density;
    std::vector<int> ngbList;
    std::vector<double> rList;
    void print()
    {
    	std::cout<<density<<"  "<<pi<<" ";
    	for(int unsigned j=0;j<ngbList.size();++j)
    		std::cout<<ngbList[j]<<" ";
    	std::cout<<std::endl;
    	for(int unsigned j=0;j<ngbList.size();++j)
    		std::cout<<rList[j]<<" ";
    	std::cout<<std::endl;
    }
    void printToFile(ebf::EbfFile &ebf,int nMerge)
    {
//    	sort2(rList.size(),&rList[0]-1,&ngbList[0]-1);
    	mysorti2(&rList[0],&rList[rList.size()],&ngbList[0]);
    	ebf.Write(&pi,1);
    	ebf.Write(&ngbList[0],nMerge);
    }

    bool operator<( const pinfo &a ) const
	{
	    return pi > a.pi;
	}
    bool operator>( const  pinfo &a ) const
	{
	    return pi < a.pi;
	}
};


class ScaleInfo
{
public:
	double *metric;
	double *matrix;
};

class Tree
{
public:
	Tree();
	void initialize(Particle* pBegin1,Particle* pEnd1,int desNumNgb1,int leafSize,int treeType,int dim1,int items1);
	virtual ~Tree();
//	void make_tree(Particle* pBegin1,Particle* pEnd1,Node*& root1,int dim1,int treeType);
//	void buildTreeA(Particle *pBegin,Particle* pEnd);
	void buildTree(int option );
	void print( );
	int computeSplitParameters(Node *lastCreatedNode);
	void knn_scatter(double* xyz,int *nlist);
	void ngb_treesearch_sphere_metric_im(const double *searchcenter,const double *metric,std::vector<class node_list> &box_list,ParticlePriorityQueue &pq);
	void ngb_treesearch_sphere_matrix_im(const double *searchcenter,const double *metric,std::vector<node_list> &box_list,ParticlePriorityQueue &pq);
//	void ngb_treesearch_sphere_metric_im(const double *searchcenter,const double *metric,std::vector<class node_list> &box_list,void (ParticlePriorityQueue::*funct)(Particle*));
	void knn_bruteForce(const double *searchcenter,const double *metric,ParticlePriorityQueue &pq);
//	void densityAll(int optionMetric,int scatterOn,int ngbListOn,int typeOfKernel,int biasCorrectionOn,int volCorrOn,int cubicCellsOn,FILE* fp);
	void densityAll(Parameters &All,ebf::EbfFile &ebf_density,ebf::EbfFile &ebf_metric,std::vector<float>  &pMetric);
	void density(const std::string& targetfile,const std::string& outfile, Parameters &All,std::vector<float>  &pMetric);
	void getMetric(Particle* Part,double *metric);
	void setNormalMetric(double *metric);
	void readGlobalMetricFromFile(double* metric,Parameters &All);
	void computeMetricIsochrone(double x1,double x2,std::vector<double> &xv);
	void saveMetric(double* metric,double factor,ebf::EbfFile &ebf_metric);
	void computeMetricGeneral( );
	void computeMetric(double* PartPos,double* metric,std::vector<class node_list> &box_list,pqStats &pq_stats);
	void makeCubic(double* metric);
	int checkVol(double *Pos,double* metric,ParticlePriorityQueue &pq,pqStats &pq_stats);
	int checkVol2(double *Pos,double* metric,ParticlePriorityQueue &pq,pqStats &pq_stats);
	int checkVol1(double *Pos,double* metric,ParticlePriorityQueue &pq);
	void smooth_quant(const double *searchcenter,int qBegin,int qEnd,std::vector<double> &quantMean,std::vector<class node_list> &box_list);
	void smoothMetric(double *metric,std::vector<float>  &pMetric,ParticlePriorityQueue &pq);
	void smoothMatrix(double *metric,std::vector<float>  &pMetric,ParticlePriorityQueue &pq);
	void clear();
	Node *root;
//	NodeA *rootA;
	std::vector<double> lp;
	int dim,items,numBucket,desNumNgb;
//	ParticlePriorityQueue pq;
private:
	float dummy[10];
//	ScaleInfo *scale;
//	pqStats pq_stats;
//	double *boundary;

};

#endif /*TREE_H_*/
