from setuptools import setup, Extension

USE_CYTHON = False

if USE_CYTHON:
    ext = '.pyx'
else:
    ext='.cpp'
    
source_files = ['enlink'+ext,'src/ebf.cpp', 'src/Functions.cpp', 'src/Image.cpp', 'src/NodeA.cpp', 'src/Particle.cpp', 'src/SurveyError.cpp', 'src/ebftable.cpp', 'src/gf.cpp', 'src/Kernel.cpp', 'src/Node.cpp', 'src/ParticlePriorityQueue.cpp', 'src/Timer.cpp', 'src/Enlink.cpp', 'src/gfextra.cpp', 'src/Mat.cpp', 'src/Parameters.cpp', 'src/Splitter.cpp', 'src/Tree.cpp']

    
extensions = [Extension("enlink",source_files,include_dirs=["include"],language="c++")]

if USE_CYTHON:
    from Cython.Build import cythonize
    extensions = cythonize(extensions)

setup(name='enlink',
      version='0.1.0',
      author='Sanjib Sharma',
      author_email='bugsanjib@gmail.com',
      classifiers=[
        "Programming Language :: Python",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering"],
      ext_modules = extensions,
      install_requires=['ebfpy'],
      long_description="""\
Python module for doing density based hierarchical clustering 
--------------------------------------------------------------
EnLink is a density based hierarchical clustering algorithm for
multi-dimensional data sets. Can identify clsuters of arbitrary
shape and size. It is fast and has a low memory footprint,
making it ideal for analyzing large data sets.  
"""
)

