//============================================================================
// Author      : Sanjib Sharma
// Version     :
// Copyright   : Copyright (c) Sanjib Sharma 2010
//============================================================================

#include "sutils.h"
#include "Parameters.h"
#include "Timer.h"
#include "Tree.h"
#include "gf.h"
#include "gfextra.h"
#include <sstream>


bool fileExists(const char* filename)
{
	std::ifstream infile;
	infile.open (filename);
	if (infile.is_open())
	{
		infile.close();
		return 1;
	}
	else
	{
		return 0;
	}
}

void printMetric2(std::vector<float> &pMetric,int count,int dim)
{
std::vector<double> metric1(dim,0.0);
std::vector<double> metric2(dim,0.0);
std::vector<double> metric3(dim,0.0);
std::vector<double> metric4(dim,0.0);
float temp;
for(int k=0;k<dim;++k)
{
	temp=pMetric[k];
//	temp=1.0/sqrt(pMetric[k]);
		metric1[k]=temp;
		metric2[k]=temp;
		metric3[k]+=temp*temp;
		metric4[k]+=temp*temp;
}
for(int i=1;i<(count);++i)
	{
		for(int k=0;k<dim;++k)
		{
			temp=pMetric[i*dim+k];
//			temp=1.0/sqrt(pMetric[i*dim+k]);
			if(metric1[k]>temp)
				metric1[k]=temp;
			if(metric2[k]<temp)
				metric2[k]=temp;
			metric3[k]+=temp;
			metric4[k]+=temp*temp;
		}
	}
for(int k=0;k<dim;++k)
{
	metric3[k]/=count;
	metric4[k]/=count;
	metric4[k]=sqrt(metric4[k]-metric3[k]*metric3[k]);
}
std::cout<<"Printing metric stats"<<std::endl;
std::cout<<"Dim"<<" "<<"Min"<<" "<<"Max"<<" "<<"Mean"<<" "<<"Sig"<<std::endl;
	for(int k=0;k<dim;++k)
		std::cout<<k<<" "<<metric1[k]<<" "<<metric2[k]<<" "<<metric3[k]<<" "<<metric4[k]<<std::endl;
	std::cout<<std::endl;
}


void printMetric(std::vector<float> &pMetric,int count,int dim)
{
std::vector<double> metric1(dim,0.0);
std::vector<double> metric2(dim,0.0);
for(int k=0;k<dim;++k)
{
		metric1[k]=1.0/sqrt(pMetric[k]);
		metric2[k]=1.0/sqrt(pMetric[k]);
}
for(int i=0;i<(count);++i)
	{
		for(int k=0;k<dim;++k)
		{
			if(metric1[k]>(1.0/sqrt(pMetric[i*dim+k])))
				metric1[k]=1.0/sqrt(pMetric[i*dim+k]);
			if(metric2[k]<(1.0/sqrt(pMetric[i*dim+k])))
				metric2[k]=1.0/sqrt(pMetric[i*dim+k]);

		}
	}
	for(int k=0;k<dim;++k)
		std::cout<<k<<" "<<metric1[k]<<" "<<metric2[k]<<std::endl;
	std::cout<<std::endl;
//exit(1);
}


// int run_enlink(const char*c)
// {
//     std::string str(c);
//     std::string buf;                 // Have a buffer string
//     std::stringstream ss(str);       // Insert the string into a stream

//     std::vector<std::string> tokens; // Create vector to hold our words

//     while (ss >> buf)
//       {
//         tokens.push_back(buf);
// 	std::cout<<buf<<std::endl;
//       }

//     return 0;
// }


int run_enlink(const char* c,const char* cmetric)
{
    int status;
    ParticleData Pa;
    Timer timer1,timer0;
    getTypeCheck();
    srand48(3);
    //------------------------------------------------
    // Set Parameters
    Parameters All;
    status=All.setFromArguments_from_string(c,cmetric);
    if(status != 0)
      return status;
  
  
    
    //------------------------------------------------
    //  Read data
    timer1.start();

    All.checkCompilation();
    All.print();

    if(All.DensityOn)
    {
    	Pa.readFromFile(All.InitCondFile,All.Dim,All.PaSuffix);
    	All.Dim=Pa.dim;
    }

    if((All.Dim==-1)&&(Pa.npart.size()>0))
    {
    	std::cout<<"Dim not set"<<std::endl;
    	return 1;
    }

    std::string s(All.InitCondFile);
    if(s.find(".ebf")!=s.npos)
    	s.erase(s.begin()+s.find(".ebf"),s.end());
    else
      if(s.find(".ascii")!=s.npos)
	s.erase(s.begin()+s.find(".ascii"),s.end());
      else
        {
	  std::cout<<"file extension not recognized"<<std::endl;
	  exit(1);
        }

    strcpy(All.InitCondFile,s.c_str());
    timer1.print("Data Reading Time");

    //---------------------------------------------------------
    //  Setup output files
    ebf::EbfFile ebf_density;
    ebf::EbfFile ebf_metric;
    // Start new metric density files
    if(All.DensityOn)
      {
	s.clear(); s.assign(All.InitCondFile);		 s+=All.DeSuffix; s+="_den.ebf";
	FILE* fd=fopen(s.c_str(),"w");
	fclose(fd);
	s.clear(); s.assign(All.InitCondFile);		 s+=All.DeSuffix; s+="_metric.ebf";
	fd=fopen(s.c_str(),"w");
	fclose(fd);
      }


    if(All.NgblistOn)
      {
	s.clear(); s.assign(All.InitCondFile);		 s+=All.DeSuffix; s+="_den.ebf";
	int nMerge=(All.DesNumNgb>9) ? 9: (All.DesNumNgb-1);
	//		 int nMerge=8;
	nMerge++;
	int64_t dims[2];
	dims[0]=Pa.size();
	dims[1]=nMerge;
	ebf_density.Open(s,"/NGB_LIST","w",2,"",2,&dims[0]);
      }

    if(All.MetricOn)
      {
	s.clear(); s.assign(All.InitCondFile);		 s+=All.DeSuffix; s+="_metric.ebf";
	int64_t dims[2];
	dims[0]=Pa.size();
	dims[1]=Pa.dim;
	ebf_metric.Open(s,"/H_smooth","w",4,"",2,&dims[0]);
      }
    //-------------------------------------------------------------
    // Build tree
    //	 if(All.HammerOn==1)
    //	 {
    //		 Pa.hammer_aitoff();
    //	 }

    //	 Tree *T=new Tree[Pa.npart.size()];
    std::vector<Tree> T(Pa.npart.size());
    //	 treeType=2;
    timer1.start();
    //	for(unsigned int i=0;i<Pa.npart.size();++i)
    s.clear(); s.assign(All.InitCondFile);		 s+=All.DeSuffix; s+="_metric.dat";
    for(unsigned int i=0;i<Pa.npart.size();++i)
      {
	//		 int leafSize=All.DesNumNgb/2;
	//		 		 std::cout<<"here"<<std::endl;
	//		    	std::cout<<"here "<<Pa.npartc[i]<<" "<<Pa.npartc[i+1]<<" "<<Pa.npart[i]<<std::endl;

	if(All.BootstrapOn==1)
	  {
	    int tempi=std::max(Pa.dim+2,10);
	    //				 tempi=32;
	    //				 calculate_metric(s.c_str(),Pa.begin()+Pa.npartc[i],Pa.begin()+Pa.npartc[i+1],Pa.dim+2,All.TreeType,All.RotationOn,Pa.dim,Pa.items);
	    calculate_metric(s.c_str(),Pa.begin()+Pa.npartc[i],Pa.begin()+Pa.npartc[i+1],tempi,All.TreeType,All.RotationOn,Pa.dim,Pa.items);
	    T[i].initialize (Pa.begin()+Pa.npartc[i],Pa.begin()+Pa.npartc[i+1],All.DesNumNgb,0,All.TreeType,Pa.dim,Pa.items);
	  }
	else
	  {
	    //	    		All.TreeType=1;
	    //		    	if(All.TypeOfSmoothing==3)
	    //int Tree.cpp numBucket=DesNumNgb/2  so leafSize must be less than equal to All.DesNumNgb/2
	    // if All.TypeOfSmoothing==3 and All.BootstrapOn==0 then leafSize must be 1 to allow smoothing
	    if(All.TypeOfSmoothing==2)
	      {
		//		    	leafSize=max(8,All.DesNumNgb/2);
		T[i].initialize (Pa.begin()+Pa.npartc[i],Pa.begin()+Pa.npartc[i+1],All.DesNumNgb,0,1,Pa.dim,Pa.items);
	      }
	    else
	      {
		//	    		leafSize=1;
		T[i].initialize (Pa.begin()+Pa.npartc[i],Pa.begin()+Pa.npartc[i+1],All.DesNumNgb,1,All.TreeType,Pa.dim,Pa.items);
	      }
	  }
	//		 T[i].initialize (Pa.begin()+Pa.npartc[i],Pa.begin()+Pa.npartc[i+1],All.DesNumNgb,4,Pa.dim,Pa.items);
      }
    timer1.print("Tree building Time",Pa.size()*2);
    //-------------------------------------------------------------
	 // Order particle data
    timer1.start();
    if(All.DensityOn)
      Pa.kdOrder();
    timer1.print("Particle ordering Time",Pa.size()*1.0);
    //-------------------------------------------------------------
	 // Calculate density metric ngblist
    timer1.start();
    if(All.DensityOn)
      {
    	std::vector<float> pMetric;
    	if((All.TypeOfSmoothing==3)&&(All.BootstrapOn==1))
	  {
	    s.clear(); s.assign(All.InitCondFile);		 s+=All.DeSuffix; s+="_metric.dat";
	    std::ifstream fin(s.c_str(), std::ios::in | std::ios::binary);
	    int info[10];
	    fin.read(reinterpret_cast<char *>(info),40);
	    pMetric.resize(info[1]*info[2],0.0);
	    fin.read(reinterpret_cast<char *>(&pMetric[0]),info[info[0]+2]*info[info[0]+3]);
	    fin.close();
	    //    		 printMetric2(pMetric,info[1],info[2]);
	  }
    	if (All.TargetFile.size() ==0)
	  {
	    for(unsigned int i=0;i<Pa.npart.size();++i)
	      {
		T[i].densityAll(All,ebf_density,ebf_metric,pMetric);
	      }
	  }
    	else
	  {
	    s.clear(); s.assign(All.InitCondFile);		 s+=All.DeSuffix; s+="_den.ebf";
	    T[0].density(All.TargetFile,s,All,pMetric);
	  }

      }
    timer1.print("Time Density calculation ",Pa.size());

    if(All.NgblistOn)
      {
	ebf_density.Close();
      }
    if(All.MetricOn)
      ebf_metric.Close();
    if(All.DensityOn)
      {
	for(unsigned int i=0;i<Pa.npart.size();++i)
	  {
	    T[i].clear();
	  }
      }
	 //-----------------------------------------------------------------
    //  Write Density Mass Ngblist and sort ngblist according to density
//	char buf1[100];
    if((All.DensityOn==1)&&(All.TargetFile.size() ==0))
      {
    	std::vector<float> Density(Pa.size(),0.0);
    	for(int i=0;i<Pa.size();++i)
	  {
	    Density[Pa[i].ID(Pa.items)]=Pa[i].Density(Pa.items);
	  }
    	if(All.massDensityOn==1)
	  {
	    for(int i=0;i<Pa.size();++i)
	      Density[Pa[i].ID(Pa.items)]=Pa[i].Mass(Pa.items);
	  }

	//    	std::cout<<"Densities"<<std::endl;
	//    	for(int i=0;i<Pa.size();i+=Pa.size()/100)
	//    	{
	//    		std::cout<<i<<" "<<Density[i]<<std::endl;
	//    	}

	//    	sprintf(buf1,"%s%s%s",All.InitCondFile,All.DeSuffix,".ebf");
    	s.clear(); s.assign(All.InitCondFile); s+=All.DeSuffix; s+="_den.ebf";
	//    	if(All.NgblistOn)
	//    		ebf_density.open(s,"a","/DENSITY",4,Pa.size());
	//    	else
	//    		ebf_density.open(s,"w","/DENSITY",4,Pa.size());
	//    	ebf_density.sync(&Density[0]);
    	if(All.NgblistOn)
	  ebf::Write(s,"/DENSITY",&Density[0],"a","",Pa.size());
    	else
	  ebf::Write(s,"/DENSITY",&Density[0],"w","",Pa.size());


    	if(All.asciiOn==1)
	  {
	    std::cout<<"Writing Ascii density file"<<std::endl;
	    std::string s1=All.InitCondFile; s1+=All.DeSuffix; s1+="_den.txt";
	    std::ofstream fout(s1.c_str());
	    fout<<"# "<<Pa.size()<<" "<<1<<std::endl;
	    fout.setf(std::ios::scientific);    		fout<<std::setprecision(6);
	    for(int i=0;i<Pa.size();++i)
	      {
		fout<<Density[i]<<std::endl;
	      }
	  }

    	for(int i=0;i<Pa.size();++i)
    	{
    		Density[Pa[i].ID(Pa.items)]=Pa[i].Mass(Pa.items);
    	}
    	ebf::Write(s,"/MASS",&Density[0],"a","",Pa.size());
//    	ebf_density.open(s,"a","/MASS",4,Pa.size());
//    	ebf_density.sync(&Density[0]);
        timer1.start();
	if(((All.ClusterOn)||(All.NgblistOn)))
	  sort_plist(s.c_str());
	ebf::Write(s,"/dim",&All.Dim,"a","",1);
	timer1.print("Time Sorting ngblist");
      	std::cout<<std::setw(16)<<"File written:"<<s<<"[.txt]"<<std::endl;
      }

    //-----------------------------------------------------------------
    //  Write metric to density file
    if(All.MetricOn==1)
      {
	s.clear(); s.assign(All.InitCondFile);s+=All.DeSuffix; s+="_metric.ebf";
	std::vector<float> hm(Pa.dim*Pa.size());
	//		 ebf_metric.open(s.c_str(),"r","/H_smooth",4,Pa.size(),Pa.dim);
	//		 ebf_metric.sync(&hm[0]);
	ebf::Read(s,"/H_smooth",&hm[0],hm.size());

	if((All.CubicCellsOn==1)||(Pa.dim==3))
	  {
	    std::vector<float> hm1(Pa.dim*Pa.size()/3,0.0);
	    //			 float *hm1=  new float[Pa.dim*Pa.size()/3];
	    for(int i=0;i<Pa.size();++i)
	      {
		for(int k=0;k<Pa.dim/3;++k)
		  hm1[Pa[i].ID(Pa.items)*Pa.dim/3+k]=hm[i*Pa.dim+k*3];
		//					 hm1[Pa[i].ID(Pa.items)*Pa.dim/3+k]=hm[Pa[i].ID(Pa.items)*Pa.dim+k*3];
		//				 std::cout<<hm1[Pa[i].ID(Pa.items)*Pa.dim/3+0]<<std::endl;
	      }
	    s.clear(); s.assign(All.InitCondFile); s+=All.DeSuffix; s+="_den.ebf";
	    //			 ebf_density.open(s,"a","/H_cubic",4,Pa.size(),Pa.dim/3);
	    //			 ebf_density.sync(&hm1[0]);
	    if(Pa.dim==3)
	      ebf::Write(s,"/H_cubic",&hm1[0],"a","",Pa.size());
	    else
	      ebf::Write(s,"/H_cubic",&hm1[0],"a","",Pa.size(),Pa.dim/3);

	  }

	if(All.CubicCellsOn!=1)
	  {
	    std::vector<float> hm1(Pa.dim*Pa.size(),0.0);
	    //			 float *hm1=  new float[Pa.dim*Pa.size()];
	    for(int i=0;i<Pa.size();++i)
	      {
		for(int k=0;k<Pa.dim;++k)
		  hm1[Pa[i].ID(Pa.items)*Pa.dim+k]=hm[i*Pa.dim+k];
		//					 hm1[Pa[i].ID(Pa.items)*Pa.dim+k]=hm[Pa[i].ID(Pa.items)*Pa.dim+k];
	      }
	    s.clear(); s.assign(All.InitCondFile);s+=All.DeSuffix; s+="_den.ebf";
	    //			 ebf_density.open(s,"a","/H_smooth",4,Pa.size(),Pa.dim);
	    //			 ebf_density.sync(&hm1[0]);
	    ebf::Write(s,"/H_smooth",&hm1[0],"a","",Pa.size(),Pa.dim);
	  }
	std::cout<<std::setw(16)<<"File appended with metric:"<<s<<std::endl;
      }

    //-----------------------------------------------------------------
    //  Clustering
// 	sprintf(buf1,"%s%s%s",All.InitCondFile,All.DeSuffix,".ebf");
    if(All.ClusterOn)
      {
        std::cout<<"-----------------------------------------------------------------------------"<<std::endl;
        timer1.start( );
    	s.clear(); s.assign(All.InitCondFile);s+=All.DeSuffix; s+="_den.ebf";
	ebf::Read(s,"/dim",&All.Dim);
    	gf(s.c_str(),All);
    	timer1.print("Time Group Finding");
        std::cout<<"-----------------------------------------------------------------------------"<<std::endl;
      }
    //-----------------------------------------------------------------


//    std::cout<<fixed<<"\n\t\t\tTotal\tTime\t"<<right<<std::setw(12)<<timer1.elapsedTime()<<"s \n"<<std::endl;


    //    if(All.MetricOn==0)
    {
      s.clear(); s.assign(All.InitCondFile);		 s+=All.DeSuffix; s+="_metric.ebf";
      if (fileExists(s.c_str()) == 1)
    	{
	  std::cout<<"Deleting metric file "<<s<<std::endl;
	  remove(s.c_str());
    	}
      s.clear(); s.assign(All.InitCondFile);		 s+=All.DeSuffix; s+="_metric.dat";
      if (fileExists(s.c_str()) == 1)
    	{
	  std::cout<<"Deleting metric file "<<s<<std::endl;
	  remove(s.c_str());
    	}
    }
//    {
//		 FILE* fd=fopen(s.c_str(),"w");
//		 fclose(fd);
//    }

    std::cout<<std::endl;
    timer0.stop();    timer0.print("Status-->SUCCESS     Total Time");
    std::cout<<std::endl;

    return 0;
}


