//============================================================================
// Author      : Sanjib Sharma
// Version     :
// Copyright   : Copyright (c) Sanjib Sharma 2010
//============================================================================
#include "Kernel.h"
#include "Functions.h"

Kernel::Kernel(int size1, int dim1, int TypeOfKernel1,bool biasCorrectionOn1,int scatterOn1)
{
	scatterOn=scatterOn1;
	size=size1;
	ky.resize(size+2, 0.0);
	dim=dim1;
	epsilon=1e-6;
	biasCorrectionOn=biasCorrectionOn1;
	setKernel(TypeOfKernel1);
	if(biasCorrectionOn==1)
		w_corr=computeK(dim/(1.0+dim))-computeK(0.0);
	else
		w_corr=0.0;
}

Kernel::~Kernel()
{
}

//double Kernel::computeK(double u)
//{
//	int i=int(u*(ky.size()-2));
//	return (ky[i]+(ky[i+1]-ky[i])*(u-i*dx));
//}


double Kernel::computeDensityGen(ParticlePriorityQueue &pq,double* metric,int items)
{
	double hinv_d=1.0;
	double density=0.0;
	double w=1.0/sqrt(pq.top().r);
	for(int i=0;i<dim;i++)
		hinv_d*=w;
	hinv_d*=fd*pq.hinvd;
	if(scatterOn==0)
	{
		for(Neighbor* ngb=pq.begin();ngb<pq.end();ngb++)
		{
			w=computeK(sqrt(ngb->r/pq.top().r));
			density+=w*ngb->Pa->Mass(items);
		}
		density*=hinv_d;
	}
	return density;

}


void Kernel::computeDensity(Particle* Part,ParticlePriorityQueue &pq,double* metric,int items)
{
	double hinv_d=1.0;
	double w=1.0/sqrt(pq.top().r);
//	for(int i=0;i<dim;i++)
//		hinv_d*=w*metric[i];
//	hinv_d*=fd;
	for(int i=0;i<dim;i++)
		hinv_d*=w;
	hinv_d*=fd*pq.hinvd;

	if(scatterOn==0)
	{
		Part->Density(items)=0.0;
		for(Neighbor* ngb=pq.begin();ngb<pq.end();ngb++)
		{
			w=computeK(sqrt(ngb->r/pq.top().r));
			(Part->Density(items))+=w*ngb->Pa->Mass(items);
		}
		(Part->Density(items))*=hinv_d;
	}

	if(scatterOn==1)
	{
		hinv_d*=Part->Mass(items);
		for(Neighbor* ngb=pq.begin();ngb<pq.end();ngb++)
		{
			w=computeK(sqrt(ngb->r/pq.top().r));
			(ngb->Pa->Density(items))+=w*hinv_d;
		}
		hinv_d/=Part->Mass(items);
	}

	if(scatterOn==2)
	{
		hinv_d*=0.5;
		for(Neighbor* ngb=pq.begin();ngb<pq.end();ngb++)
		{
			w=computeK(sqrt(ngb->r/pq.top().r));
			(Part->Density(items))+=w*ngb->Pa->Mass(items)*hinv_d;
			(ngb->Pa->Density(items))+=w*Part->Mass(items)*hinv_d;
		}
		hinv_d*=2;
	}
		(Part->Density(items))+=w_corr*Part->Mass(items)*hinv_d;

}


void Kernel::computeDensityS(Particle* Part,ParticlePriorityQueue &pq,double* metric,int items)
{
	double hinv_d=1.0;
	double w=1.0/sqrt(pq.top().r);
//	for(int i=0;i<dim;i++)
//		hinv_d*=w*metric[i];
//	hinv_d*=fd;
	for(int i=0;i<dim;i++)
		hinv_d*=w;
	hinv_d*=fd*pq.hinvd;

	if(scatterOn==0)
	{
		Part->Density(items)=0.0;
		for(Neighbor* ngb=pq.begin();ngb<pq.end();ngb++)
		{
			(Part->Density(items))+=ngb->Pa->Mass(items);
		}
		Part->Density(items)/=pq.size();
	}


}


void Kernel::computeDensityN(Particle* Part,ParticlePriorityQueue &pq,double* metric,int items,int N)
{
	int kden=N;
	assert(kden<=pq.size());
	Neighbor* ngblist=pq.ngblist;
	std::vector<Neighbor> ngblistv;
	double rmax2=pq.top().r;
	if(kden<pq.size())
	{
		for(Neighbor* ngb=pq.begin();ngb<pq.end();ngb++)
			ngblistv.push_back(*ngb);
		sort(ngblistv.begin(),ngblistv.end());
		ngblist=&ngblistv[0];
		rmax2=(ngblist[N-1].r);
	}
//	std::cout<<kden<<" "<<pq.size()<<" "<<scatterOn<<" "<<std::endl;


	double hinv_d=1.0;
	double w=1.0/sqrt(rmax2);
	for(int i=0;i<dim;i++)
		hinv_d*=w;
	hinv_d*=fd*pq.hinvd;

	if(scatterOn==0)
	{
		for(int i=0;i<kden;i++)
		{
			w=computeK(sqrt(ngblist[i].r/rmax2));
			(Part->Density(items))+=w*ngblist[i].Pa->Mass(items);
		}
		(Part->Density(items))*=hinv_d;
	}

	if(scatterOn==1)
	{
		hinv_d*=Part->Mass(items);
		for(int i=0;i<kden;i++)
		{
			w=computeK(sqrt(ngblist[i].r/rmax2));
			(ngblist[i].Pa->Density(items))+=w*hinv_d;
		}
		hinv_d/=Part->Mass(items);
	}

	if(scatterOn==2)
	{
		hinv_d*=0.5;
		for(int i=0;i<kden;i++)
		{
			w=computeK(sqrt(ngblist[i].r/rmax2));
			(Part->Density(items))+=w*ngblist[i].Pa->Mass(items)*hinv_d;
			(ngblist[i].Pa->Density(items))+=w*Part->Mass(items)*hinv_d;
		}
		hinv_d*=2;
	}
		(Part->Density(items))+=w_corr*Part->Mass(items)*hinv_d;

}



void Kernel::setKernel(int TypeOfKernel)
{
	// more accurate 11/16/05
	float fsp[]=
	{ 1.3333369, 1.8189136, 2.5464790, 3.6606359, 5.4037953, 8.1913803,
			12.748839, 20.366416, 33.380983, 56.102186, 96.621159, 170.39909,
			307.49826, 567.37865, 1069.6362, 2058.8172, 4043.1010, 8095.3490,
			16515.921, 34312.457 };
	float fbw[]=
	{ 0.93750176, 0.95492964, 1.0444543, 1.2158542, 1.4960706, 1.9350925,
			2.6191784, 3.6957561, 5.4191207, 8.2347774, 12.937213, 20.969717,
			35.003422, 60.073937, 105.84864, 191.22182, 353.77415, 669.54824,
			1295.0185, 2557.4981 };

	float fcic[]=
	{ 1.0000020, 0.95492963, 0.95492963, 1.0132118, 1.1398631, 1.3545643,
			1.6932052, 2.2174514, 3.0316711, 4.3134446, 6.3690677, 9.7358945,
			15.373955, 25.030601, 41.945765, 72.238385, 127.67576, 231.29376,
			428.98102, 813.72530 };

	float fep[]=
	{ 0.75000113, 0.63661975, 0.59683102, 0.60792705, 0.66492015, 0.77403670,
			0.95242788, 1.2319173, 1.6674189, 2.3527875, 3.4499109, 5.2424031,
			8.2360444, 13.349647, 22.283674, 38.243824, 67.384374, 121.73344,
			225.21478, 426.23651 };

	float ftsc[]=
	{ 1.6875038, 1.9833154, 2.4171656, 3.0521959, 3.9950015, 5.4207369,
			7.6205882, 11.087829, 16.674494, 25.880863, 41.399134, 68.151573,
			115.30531, 200.24626, 356.54234, 650.15322, 1212.9379, 2312.9804,
			4504.4024, 8951.2220 };

	double PI=3.141596;
	double vd=2.0*pow(PI,dim/2.0)/(dim*exp(gammln(dim/2.0)));
	double sigma=1/4.0;
	double f1=0.0;
	std::string kernelName;

	switch (TypeOfKernel)
	{
	case 0:
		f1=fsp[dim-1];
		kernelName="Spline";
		break; //Spline
	case 1:
		f1=1.0/vd;
		kernelName="Top-Hat";
		break; //Top Hat
	case 2:
		f1=fbw[dim-1];
		kernelName="Bi-weight";
		break; //Bi-weight
	case 3:
		f1=fep[dim-1];
		kernelName="Epanechnikov";
		break; //Epanechikov
	case 4:
		f1=fcic[0];
		kernelName="CIC";
		break; //CIC
	case 5:
		f1=ftsc[0];
		kernelName="TSC";
		break; //TSC
	case 6:
		f1=1.0/pow(sqrt(2*PI)*sigma, dim);
		kernelName="Gaussian";
		break; //Gaussian
	default:
		std::cout<<"specify the Kernel Type"<<std::endl;
		exit(1);
	}

	fd=f1;

	if ((dim<1)||(dim>20))
	{
		std::cout<<"Specify the Normalization constant for dimensions > 20"<<std::endl;
		exit(1);
	}
	else
	{
		//	if(All.Debug)
		{
			std::cout<<std::left<<std::setw(36)<<"Kernel Type"<<std::setw(12)<<kernelName
					<<std::endl;
			std::cout<<std::left<<std::setw(36)<<"Kernel Normalization constant"<<"fd("<<dim<<")="<<f1
					<<std::endl;
		}
	}

	dx=1.0/size;
	std::vector<double> x(size+1,0.0);
	for (int i=0; i<=size; i++)
		x[i]=i*dx;

	if (TypeOfKernel==0)
	{
		for (int i=0; i<=size; i++)
		{
			if (x[i]<=0.5)
			{
				ky[i] = f1*(1-6*x[i]*x[i]*(1-x[i]));
			}
			else
			{
				ky[i] = f1 * 2*(1-x[i])*(1-x[i])*(1-x[i]);
			}
		}
	}

	if (TypeOfKernel==1)
	{
		for (int i=0; i<=size; i++)
			ky[i] = f1;
	}

	if (TypeOfKernel==2)
	{
		for (int i=0; i<=size; i++)
			ky[i] = f1 *(1-x[i]*x[i])*(1-x[i]*x[i]);
	}
	if (TypeOfKernel==3)
	{
		for (int i=0; i<=size; i++)
			ky[i] = f1 *(1-x[i]*x[i]);
	}
	if (TypeOfKernel==4)
	{
		for (int i=0; i<=size; i++)
			ky[i] = f1 *(1-x[i]);
	}
	if (TypeOfKernel==5)
	{
		for (int i=0; i<=size; i++)
		{
			if (x[i]<1.0/3.0)
			{
				ky[i] = f1 *(2.0/3.0-2*x[i]*x[i]);
			}
			else
			{
				ky[i] = f1 *(1-x[i])*(1-x[i]);
			}
		}
	}
	if (TypeOfKernel==6)
	{
		for (int i=0; i<=size; i++)
			ky[i] = f1 *exp(-x[i]*x[i]/(2*sigma*sigma));
	}

}



