//============================================================================
// Author      : Sanjib Sharma
// Version     :
// Copyright   : Copyright (c) Sanjib Sharma 2010
//============================================================================
#include "Parameters.h"
#include <sstream>

Parameters::Parameters()
{
//    double PI=3.14159265358979323846;
//---General  options--------------------------------------------
    Debug                 =  0;
    Dim                   =  -1;
//    ICFormat              =   1;    // O)ASCII 1)Gadget 5)Hydra 6)Hydra1
    strcpy(DeSuffix,"");
    strcpy(GrSuffix,"");
    strcpy(PaSuffix,"");
    strcpy(InitCondFile,"");
    TargetFile="";
    ImageFile="";
    DensityOn             =  0;
    NgblistOn             =  0;
    MetricOn              =  0;
    ClusterOn             =  0;
//----Group Finder Parameters-------------------------------------
    MinGpSize             =  10;
    PeakHeight            = -1.0;
    RhoB                  = 0.0;    // 0.52e-6  2.72899e-5 1e-1 1e-3 1e-8
    Delta                 = 0.0;    // 2.808e-6 2.776e-6 36.85e-5
    SepCriteria           =   0;    // 0) attach on contact 1) Distance 2)Volume based
    ContactRatio          = 0.5;    //
    GAnisotropyOn         =   0;

//----Density options  AM=adaptive metric Ker=Kernel Sp=Spherical Pr=Product
    BootstrapOn       =   0; // 0) None 1)Fiestas 2)Ker Normal 3)Ker Sp AM 4)KerPr AM 5)Periodic
    RotationOn       =   0; // 0) None 1)Fiestas 2)Ker Normal 3)Ker Sp AM 4)KerPr AM 5)Periodic
    DesNumNgb             =  10; // 2 for Fiestas and 25-100 for
    DesNumNgbA             =  10; // 2 for Fiestas and 25-100 for
    TypeOfSmoothing       =   2; // 0) None 1)Fiestas 2)Ker Normal 3)Ker Sp AM 4)KerPr AM 5)Periodic
    DensityScatter        =   0; // 0) Gather 1) Scatter 2) Both
    VolCorrOn             =   1; // 0) Disbale 1) Enable
    TypeOfKernel          =   3; // 0) B-Spline 1)top hat 2)Bi_weight (1-x^2)^2 3)Epanechikov 4)CIC 5)TSC
    KernelBiasCorrectionOn=   1; // 0)none 1)shift central particle
    NormalEuclideanMetricOn=0;
//----Tree related options-----------------------------------------
    SpatialScale          = 1.0;    //  x->x/SpatialScale and v->v
 //   PartBoundary          =   7;    // Min particles in a node to do boundary correction
    NodeSplittingCriterion=   0;    // 0)Alternate 1) Min Entropy
    CubicCellsOn          =   0;    // use 1 in spherically symmetric systems
    TreeType              =   0;
    MedianSplittingOn     =   0;
    AnisotropyOn          =   0;    // fix minimum c/a minor/major axis ratio
    PeriodicBoundaryOn    =   0;
    if(PeriodicBoundaryOn==1)
    	p_length.resize(Dim);
    ErrorOption=0;
    mu0=8.0;
//    Extra=0;
    VoxelsOn=0;
//   error
    sigma_mu=200.0;
    sigma_vr=10.0;
    sigma_vlb=10.0;
    sigma_r=0.15;
    massDensityOn=0;
//    checkCompilation();


}


void Parameters::usage(void )
{
	//[-dlgmvecxar]
    std::cout<<"EnLink version 2.3 Copyright (c) Sanjib Sharma 2010"<<std::endl<<std::endl;
    std::cout<<"USAGE "<<std::endl;
    std::cout<<"\t Enlink\t [OPTIONS] file"<<std::endl;

    std::cout<<std::endl<<"EXAMPLES"<<std::endl;
    std::cout<<" Enlink -dlg --dim=3 file.ascii"<<std::endl;
    std::cout<<"\t Basic run with default settings, Constant Euclidean metric"<<std::endl<<std::endl;
    std::cout<<" Enlink -dlg --gmetric=1 --dim=3 file.ascii"<<std::endl;
    std::cout<<"\t Constant Euclidean metric, but normalized using variance of data"<<std::endl<<std::endl;
    std::cout<<" Enlink -dlg --gmetric=2 --dim=3 file.ascii"<<std::endl;
    std::cout<<"\t custom constant from file ./metric_default.txt"<<std::endl<<std::endl;
    std::cout<<" Enlink -dlgsbr --dim=3 file.ascii"<<std::endl;
    std::cout<<"\t Recommended option. Does the best possible metric computation."<<std::endl<<std::endl;
    std::cout<<" Enlink -d --dim=3 file.ascii"<<std::endl;
    std::cout<<"\t only to output density"<<std::endl<<std::endl;
    std::cout<<" Enlink -dl --dim=3 file.ascii"<<std::endl;
    std::cout<<"\t to output density+ neighbor-list"<<std::endl<<std::endl;
    std::cout<<" Enlink -dlg --dim=3 --ngb=30  --dsuffix=_n30 --ph=4.0 --gsuffix=_g4.0  file.ebf"<<std::endl;
    std::cout<<"\t Run with user specifed settings"<<std::endl<<std::endl;
    std::cout<<" Enlink -g   --dim=3 --ngb=30  --dsuffix=_n30 --ph=4.5 --gsuffix=_g4.5  file.ascii"<<std::endl;
    std::cout<<"\t Run only group finder with new value of peak height/significance"<<std::endl;
    std::cout<<"\t threshold, ph , on the previously calculated densities and neighbor-list"<<std::endl;

    std::cout<<std::endl<<"OUPUT FILES"<<std::endl;
    std::cout<<std::left<<std::setw(40)<<" file(suffixd)_den.txt"<<"(densities + neighbor-list of particles)"<<std::endl;
    std::cout<<std::setw(40)<<" file(suffixpd)_den(suffixg)_gid.txt"<<"(group id of particles)"<<std::endl;
    std::cout<<std::setw(40)<<" file(suffixpd)_den(suffixg)_grp.txt"<<"(group properties and its hierarchy)"<<std::endl;
    std::cout<<std::endl<<" Brackets are inserted for convenience only and are not present in actual "<<std::endl;
    std::cout<<" output file names. Also, suffixpd=suffixp+suffixd"<<std::endl;

    std::cout<<std::endl<<"OPTIONS"<<std::endl;
    std::cout<<"Main parameters:"<<std::endl;
    std::cout<<std::left<<std::setw(24)<<"\t -d"<<"compute density"<<std::endl;
    std::cout<<std::setw(24)<<"\t -l"<<"calculate neighbour list "<<std::endl;
    std::cout<<std::setw(24)<<"\t -g"<<"calculate the groups "<<std::endl;
    std::cout<<std::setw(24)<<"\t -v"<<"verbose mode"<<std::endl;

    std::cout<<std::endl<<"File Related Parameters:"<<std::endl;
    std::cout<<std::setw(24)<<"\t --psuffix=suffixp"<<"suffix for density output files"<<std::endl;
    std::cout<<std::setw(24)<<"\t --dsuffix=suffixd"<<"suffix for density output files"<<std::endl;
    std::cout<<std::setw(24)<<"\t --gsuffix=suffixg"<<"suffix for group output files"<<std::endl;

    std::cout<<std::endl<<"Density and Tree Parameters:"<<std::endl;
    std::cout<<std::setw(24)<<"\t --dim=DIM"<<"no of dimensions"<<std::endl;
    std::cout<<std::setw(24)<<"\t -s"<<"entropy based adaptive metric, subspace scheme"<<std::endl;
    std::cout<<std::setw(24)<<"\t -b"<<"boot strap technique for smoothing the metric"<<std::endl;
    std::cout<<std::setw(24)<<"\t -r"<<"anisotopic metric"<<std::endl;

    std::cout<<std::endl<<"Group Finder Parameters:"<<std::endl;
    std::cout<<std::setw(24)<<"\t --mng=int "<<"minimum number of particles in a group"<<std::endl;
    std::cout<<std::setw(24)<<"\t --ph=float  "<<"min significance level of Groups"<<std::endl;
    std::cout<<std::setw(24)<<"\t --gmetric=int  "<<"1) variance normalized 2)Custom from file "<<std::endl;

    std::cout<<std::setw(24)<<"\t --ascii=1  "<<" for ascii output of density and image"<<std::endl;
    std::cout<<std::setw(24)<<"\t --ifile=filename  "<<" for calculating densities on a grid/image"<<std::endl;
    std::cout<<std::setw(24)<<"\t                   "<<"change im.set in Tree.cpp for setting grid bounds"<<std::endl;

    std::cout<<std::endl<<"Report bugs to <bugsanjib@gmail.com>."<<std::endl;
}



void Parameters::print( )
{
		std::cout.precision(6);
		std::cout<<"----------------------------------"<<std::endl;
		std::cout<<"Parameter Details"<<std::endl;
		std::cout<<std::left<<std::setw(20)<<"DensityOn "<<std::setw(16)<<DensityOn<<std::endl;
		std::cout<<std::left<<std::setw(20)<<"NgbListOn "<<std::setw(16)<<NgblistOn<<std::endl;
		std::cout<<std::left<<std::setw(20)<<"MetricOn "<<std::setw(16)<<MetricOn<<std::endl;
		std::cout<<std::left<<std::setw(20)<<"ClusterOn "<<std::setw(16)<<ClusterOn<<std::endl;
		std::cout<<std::left<<std::setw(20)<<"TreeType "<<std::setw(16)<<TreeType<<std::endl;
		std::cout<<std::left<<std::setw(20)<<"TypeOfSmoothing "<<std::setw(16)<<TypeOfSmoothing<<std::endl;
		std::cout<<std::left<<std::setw(20)<<"CubicCellsOn "<<std::setw(16)<<CubicCellsOn<<std::endl;
		std::cout<<std::left<<std::setw(20)<<"DesNumNgb "<<std::setw(16)<<DesNumNgb<<std::endl;
//		std::cout<<std::left<<std::setw(12)<<"ImageOn "<<std::setw(12)<<ImageOn<<std::endl;
		std::cout<<"----------------------------------"<<std::endl;
}

int Parameters::checkCompilation( )
{


//#ifdef PERIODIC
//    if(PeriodicBoundaryOn!=1)
//    {
//	std::cout<<"Code compiled with PERIODIC option, Set parameter PeriodicBoundaryOn to 1"<<std::endl;
//	endrun(10);
//    }
//#else
//    if(PeriodicBoundaryOn!=0)
//    {
//	std::cout<<"Set parameter PeriodicBoundaryOn to 0"<<std::endl;
//	exit(10);
//    }
//#endif

//    if((TypeOfSmoothing==3)&&(NodeSplittingCriterion==0))
//    {
//    	std::cout<<"For adaptive metric NodeSplittingCriterion should not be zero"<<std::endl;
//	    exit(10);
//    }

    if((DensityOn==0)&&(ClusterOn==0))
    {
    	std::cout<<std::endl<<"One of the following option -dlgm is mandatory"<<std::endl<<std::endl;
    	usage();
	return 1;
    }

    if (CubicCellsOn == 1)
	{

		if (Dim != 6)
		//	if((Dim!=6)&&(Dim!=3))
		{
			std::cout
					<< "Cubic Cells not allowed for dimensions other than  6 here Dim="
					<< Dim << std::endl;
//			std::cout << "Choose another option or modify routine" << std::endl;
			return 1;
		}
		if (RotationOn == 1)
		{
			std::cout<<"Rotational trnasformation (option -r) not allowed for cubic cells "<<std::endl;
			std::cout<<std::endl;
			 return 1;
		}
	}




    if((BootstrapOn!=0)&&(TypeOfSmoothing==2))
    {
    	BootstrapOn=0;
    }
    if((BootstrapOn!=1)&&(RotationOn!=0))
    {
    	RotationOn=0;
    }

	TreeType=NodeSplittingCriterion;
	if(VoxelsOn==1)
	{
		if((DensityOn !=0)&&(NgblistOn !=0))
		{
			std::cout<<"Density and neighbor list calculation not allowed for voxels"<<std::endl;
			return 1;
		}
	}

	// if(Dim<1)
	// {
	// 	std::cout<<"Dimensionality of data not specified should be greater than 0 "<<std::endl;
	// 	return 1;
	// }

//    if(TypeOfSmoothing==3)
//    {
//    	TreeType=1;
//    	if(NodeSplittingCriterion==1)
//    	{
//    		if(CubicCellsOn==1)
//    			TreeType=3;
//    		else
//    			TreeType=2;
//    		if(AnisotropyOn==1)
//    			TreeType=4;
//
//    	}
//    }

//	DesNumNgbA=DesNumNgb;
//	if(DesNumNgb<10)
//		DesNumNgb=10;
	return 0;
}


int Parameters::setFromArguments_from_string(std::string command, std::string metric)
{
	int i = 1;
	float Omega = 0.0;
	std::string c1; 
	int argc;

	std::string buf;                 // Have a buffer string
	std::stringstream ss1(command);       // Insert the string into a stream	
	std::vector<std::string> argv; // Create vector to hold our words
	while (ss1 >> buf)
 	{
	  argv.push_back(buf);
	}

	std::stringstream ss2(metric);       // Insert the string into a stream	
    	metric_default.clear();
	std::cout<<"Default metric=[";
	while (ss2 >> buf)
 	{
	  metric_default.push_back(atof(buf.c_str()));
	  std::cout<<metric_default.back()<<" ";
	}
	  std::cout<<"]"<<std::endl;


	argc=argv.size();

	//   double PI=3.14159265358979323846;
	while (i < argc)
	{
		//			std::cout<<"a here"<<argv[i]<<std::endl;

	  if (argv[i].compare(0,2,"--") == 0)
		{

		  c1=argv[i].substr(argv[i].find_first_of("=")+1);

			if (argv[i].compare("--version") == 0)
			{
				std::cout << "Version 2.3 Copyright (c) Sanjib Sharma 2010"<<std::endl;
				return 1;
			}
			else if (argv[i].compare("--help") == 0)
			{
				usage();
				return 1;
			}
			else if (argv[i].compare(0,6,"--rhob") == 0)
			{
			  RhoB = atof(c1.c_str());
			}
			else if (argv[i].compare(0,9,"--sepcrit") == 0)
			{
				SepCriteria=atoi(c1.c_str());
			}
			else if (argv[i].compare(0,9,"--psuffix") == 0)
			{
			  strcpy(PaSuffix, c1.c_str());
			}
			else if (argv[i].compare(0,7,"--tfile") == 0)
			{
				TargetFile=c1;
			}
			else if (argv[i].compare(0,7,"--ifile") == 0)
			{
				ImageFile=c1;
			}
			else if (argv[i].compare(0,9,"--gsuffix") == 0)
			{
				strcpy(GrSuffix, c1.c_str());
			}
			else if (argv[i].compare(0,9,"--dsuffix") == 0)
			{
				strcpy(DeSuffix, c1.c_str());
			}
			else if (argv[i].compare(0,5,"--dim") == 0)
			{
				Dim = atoi(c1.c_str());
			}
			else if (argv[i].compare(0,5,"--ngb") == 0)
			{
				DesNumNgb = atoi(c1.c_str());
			}
			else if (argv[i].compare(0,4,"--ph") == 0)
			{
				PeakHeight = atof(c1.c_str());
			}
			else if (argv[i].compare(0,5,"--mng") == 0)
			{
				MinGpSize = atoi(c1.c_str());
			}
			else if (argv[i].compare(0,9,"--gmetric") == 0)
			{
				NormalEuclideanMetricOn = atoi(c1.c_str());
			}
			else if (argv[i].compare(0,7,"--error") == 0)
			{
				 ErrorOption= atoi(c1.c_str());
			}
			else if (argv[i].compare(0,10,"--sigma_mu") == 0)
			{
				sigma_mu = atof(c1.c_str());
			}
			else if (argv[i].compare(0,11,"--sigma_vlb") == 0)
			{
				sigma_vlb = atof(c1.c_str());
			}
			else if (argv[i].compare(0,10,"--sigma_vr") == 0)
			{
				sigma_vr = atof(c1.c_str());
			}
			else if (argv[i].compare(0,9,"--sigma_r") == 0)
			{
				sigma_r = atof(c1.c_str());
			}
			else if (argv[i].compare(0,6,"--mden") == 0)
			{
				massDensityOn = atoi(c1.c_str());
			}
			else if (argv[i].compare(0,7,"--ascii") == 0)
			{
			  asciiOn = atoi(c1.c_str());
			}
			else
			  {
			    usage();
			    return 1;
			  }
		}
	  else if (argv[i].compare(0,1,"-") == 0)
		{

		  for (size_t j = 1; j < argv[i].size(); ++j)
			{
			  switch (argv[i][j])
				{
				case 'd':
					DensityOn = 1;
					break;
				case 'l':
					NgblistOn = 1;
					break;
				case 'g':
					ClusterOn = 1;
					break;
				case 'm':
					MetricOn = 1;
					break;
				case 'v':
					Debug = 1;
					break;
				case 'n':
					NormalEuclideanMetricOn = 1;
					break;
				case 'a':
					NodeSplittingCriterion = 1;
					TypeOfSmoothing = 3;
					break;
				case 'e':
					NodeSplittingCriterion = 2;
					TypeOfSmoothing = 3;
					break;
				case 'c':
					NodeSplittingCriterion = 3;
					TypeOfSmoothing = 3;
					CubicCellsOn = 1;
					break;
				case 's':
					NodeSplittingCriterion = 4;
					TypeOfSmoothing = 3;
					break;
				case 'x':
					DensityScatter = 1;
					break;
				case 'b':
					BootstrapOn = 1;
					break;
				case 'r':
					RotationOn = 1;
					break;
				default:
					usage();
					return 1;
				}
			}
		}
		else
			strcpy(InitCondFile, argv[i].c_str());
		i++;
	}

	if (Omega > 0.0)
	{
		if (RhoB > 0.0)
		{
			std::cout << "RhoB must be set to zero to use Omega" << std::endl;
			return 1;
		}
		Delta *= (2.7755e-8 * Omega);
		RhoB *= (2.7755e-8 * Omega);
	}

    if((MetricOn==1)||(NgblistOn==1))
    	DensityOn=1;

    return checkCompilation();
}



void Parameters::setFromArguments(int argc, char **argv)
{
	int i = 1;
	float Omega = 0.0;
	char *c1;
	//   double PI=3.14159265358979323846;
	while (i < argc)
	{
		//			std::cout<<"a here"<<argv[i]<<std::endl;


		if ((argv[i][0] == '-') && (argv[i][1] == '-'))
		{

			c1 = strchr(argv[i], '=') + 1;

			if (strcmp(argv[i], "--version") == 0)
			{
				std::cout << "Version 2.3 Copyright (c) Sanjib Sharma 2010"<<std::endl;
				exit(1);
			}
			else if (strcmp(argv[i], "--help") == 0)
			{
				usage();
			}
			else if (strncmp(argv[i], "--rhob", 6) == 0)
			{
				RhoB = atof(c1);
			}
			else if (strncmp(argv[i], "--sepcrit", 9) == 0)
			{
				SepCriteria=atoi(c1);
			}
			else if (strncmp(argv[i], "--psuffix", 9) == 0)
			{
				strcpy(PaSuffix, c1);
			}
			else if (strncmp(argv[i], "--tfile", 7) == 0)
			{
				TargetFile=c1;
			}
			else if (strncmp(argv[i], "--ifile", 7) == 0)
			{
				ImageFile=c1;
			}
			else if (strncmp(argv[i], "--gsuffix", 9) == 0)
			{
				strcpy(GrSuffix, c1);
			}
			else if (strncmp(argv[i], "--dsuffix", 9) == 0)
			{
				strcpy(DeSuffix, c1);
			}
			else if (strncmp(argv[i], "--dim", 5) == 0)
			{
				Dim = atoi(c1);
			}
			else if (strncmp(argv[i], "--ngb", 5) == 0)
			{
				DesNumNgb = atoi(c1);
			}
			else if (strncmp(argv[i], "--ph", 4) == 0)
			{
				PeakHeight = atof(c1);
			}
			else if (strncmp(argv[i], "--mng", 5) == 0)
			{
				MinGpSize = atoi(c1);
			}
			else if (strncmp(argv[i], "--gmetric", 9) == 0)
			{
				NormalEuclideanMetricOn = atoi(c1);
			}
			else if (strncmp(argv[i], "--error", 7) == 0)
			{
				 ErrorOption= atoi(c1);
			}
			else if (strncmp(argv[i], "--sigma_mu", 10) == 0)
			{
				sigma_mu = atof(c1);
			}
			else if (strncmp(argv[i], "--sigma_vlb", 11) == 0)
			{
				sigma_vlb = atof(c1);
			}
			else if (strncmp(argv[i], "--sigma_vr", 10) == 0)
			{
				sigma_vr = atof(c1);
			}
			else if (strncmp(argv[i], "--sigma_r", 9) == 0)
			{
				sigma_r = atof(c1);
			}
			else if (strncmp(argv[i], "--mden", 6) == 0)
			{
				massDensityOn = atoi(c1);
			}
			else if (strncmp(argv[i], "--ascii", 6) == 0)
			{
				asciiOn = atoi(c1);
			}
			else
				usage();
		}
		else if (argv[i][0] == '-')
		{

			for (size_t j = 1; j < strlen(argv[i]); ++j)
			{
				switch (argv[i][j])
				{
				case 'd':
					DensityOn = 1;
					break;
				case 'l':
					NgblistOn = 1;
					break;
				case 'g':
					ClusterOn = 1;
					break;
				case 'm':
					MetricOn = 1;
					break;
				case 'v':
					Debug = 1;
					break;
				case 'n':
					NormalEuclideanMetricOn = 1;
					break;
				case 'a':
					NodeSplittingCriterion = 1;
					TypeOfSmoothing = 3;
					break;
				case 'e':
					NodeSplittingCriterion = 2;
					TypeOfSmoothing = 3;
					break;
				case 'c':
					NodeSplittingCriterion = 3;
					TypeOfSmoothing = 3;
					CubicCellsOn = 1;
					break;
				case 's':
					NodeSplittingCriterion = 4;
					TypeOfSmoothing = 3;
					break;
				case 'x':
					DensityScatter = 1;
					break;
				case 'b':
					BootstrapOn = 1;
					break;
				case 'r':
					RotationOn = 1;
					break;
				default:
					usage();
				}
			}
		}
		else
			strcpy(InitCondFile, argv[i]);
		i++;
	}

	if (Omega > 0.0)
	{
		if (RhoB > 0.0)
		{
			std::cout << "RhoB must be set to zero to use Omega" << std::endl;
			exit(1);
		}
		Delta *= (2.7755e-8 * Omega);
		RhoB *= (2.7755e-8 * Omega);
	}

    if((MetricOn==1)||(NgblistOn==1))
    	DensityOn=1;

	checkCompilation();
}





Parameters::~Parameters()
{
}


