//============================================================================
// Author      : Sanjib Sharma
// Version     :
// Copyright   : Copyright (c) Sanjib Sharma 2010
//============================================================================
//#include <hdf5.h>
#include "Particle.h"
#include "Functions.h"
#include "Timer.h"
#include "ebfvector.hpp"
#include "SurveyError.h"


template<class T,class T1>
void cross_product(T r,T v,T1 rv)
{
	double rv1[3];
	rv1[0]=r[1]*v[2]-r[2]*v[1];	rv1[1]=-(r[0]*v[2]-r[2]*v[0]);	rv1[2]=(r[0]*v[1]-r[1]*v[0]);
	rv[0]=rv1[0];		rv[1]=rv1[1];		rv[2]=rv1[2];
}



ParticleData::ParticleData()
{
	data=NULL;
	part=NULL;
	nsize=0;
}

ParticleData::ParticleData(const char* fname,int dim1,const char* s)
{
	data=NULL;
	part=NULL;
	dim=dim1;
	readFromFile(fname,dim1,s);
}

template<class T>
T radius3(const T* Pos)
{
	return sqrt(Pos[0]*Pos[0]+Pos[1]*Pos[1]+Pos[2]*Pos[2]);
}
template<class T>
T radius3(const T* Pos,const T*PosC)
{
	return sqrt((Pos[0]-PosC[0])*(Pos[0]-PosC[0])+(Pos[1]-PosC[1])*(Pos[1]-PosC[1])+(Pos[2]-PosC[2])*(Pos[2]-PosC[2]));
}

int ParticleData::convert_p3v3(int dim1,double mu0)
{
//	int t1=0;
	Normaldev randn(0.0, 1.0, 3l);
	SurveyError SuE(1,0.0,0.0,0.0,0.0);
	double pos[6];
	double pos1[6],l[3];
	int p, k;
	p = dim;
	if (sv.size() > 0)
		for (int i = 0; i < nsize; ++i)
		{
			SuE.xyz_to_lbr(part[i].Pos, pos, dim1);

//			double temp1=pos[2],temp2,sig=0.25;
////			temp2=(pos[2]+100)*100;
//			if(pos[2]<=0)
//				pos[2]=drand48();
//			temp1=pos[2];
//			temp2=-1;
//			while(temp2<=0)
//				temp2=(pos[2]*(1+sig*randn.dev()));
//			temp2=5*log10(temp2*100)-8;
////			temp2=temp2+500;
//			while(temp2<=0)
//			{
//				temp2=drand48();
//			}
//			for (int j = 0; j < (3); ++j)
//				part[i].Pos[j] *= temp2/temp1;
//			SuE.xyz_to_lbr(part[i].Pos, pos, dim1);

			for (int j = 0; j < dim1; ++j)
				pos1[j] = part[i].Pos[j]+posC[j];

			p = 0;
			k = 0;
			if (sv[k] == std::string("p3"))
			{
				p = p + 3;
				k++;
			}
			else if (sv[k] == std::string("s3"))
			{
				double temp = 5* log10 (pos[2] * 100) - (mu0);
				if(temp<0.0)
					temp=fabs(0.3*randn.dev());
				// for 2mass temp = pos[2] - 8; or temp=pos[2]
				for (int j = 0; j < 3; ++j)
					part[i].Pos[p+j] *= (fabs(temp) / pos[2]);
				//			for (int j = p; j < (p+3); ++j)
				//				part[i].Pos[j]=pos[j];
				//			part[i].Pos[2]=5*log10(pos[2]*100);
				p = p + 3;
				k++;
			}
			else if (sv[k] == std::string("h3"))
			{
//				float temp;
				for (int j = 0; j < 3; ++j)
					part[i].Pos[p+j] = pos[j];
				SuE.lbr_to_hammer_aitoff(part[i].Pos);
				//			part[i].Pos[2]=5*log10(pos[2]*100)-8;
				part[i].Pos[p+2]=pos[2]*pos[2]*pos[2];
				p = p + 3;
				k++;
			}
			else if(sv[k] == std::string("en"))
			{
				part[i].Pos[p]=BJmodel.energy(pos1,pos1+3);
				p++;
				k++;
			}

			if (k < int(sv.size()))
			{
				if (sv[k] == std::string("v3"))
				{
					p = p + 3;
					k++;
				}
				else if ((sv[k] == std::string("u3")) || (sv[k] == std::string("u2")))
				{
					for (int j = 0; j < 3; ++j)
						part[i].Pos[p+j] = pos[3+j];
					if (sv[k] == std::string("u3"))
						p = p + 3;
					if (sv[k] == std::string("u2"))
						p = p + 2;
					k++;
				}
				else if ((sv[k] == std::string("m3")) || (sv[k] == std::string("m2")))
				{
					for (int j = 0; j < 2; ++j)
						part[i].Pos[p+j] = pos[3+j] / (pos[2]);
					part[i].Pos[p+2] = pos[5];
					if (sv[k] == std::string("m3"))
						p = p + 3;
					if (sv[k] == std::string("m2"))
						p = p + 2;
					k++;
				}
				else if (sv[k] == std::string("vr"))
				{
					part[i].Pos[p] = pos[5];
					p = p + 1;
					k++;
				}
				else if (sv[k] == std::string("a3"))
				{
					for (int j = 0; j < 2; ++j)
						part[i].Pos[p+j] = pos[3+j] / (pos[2]);
					part[i].Pos[p+2] = pos[5];
					p = p + 3;
					if (dim == 6)
					{
						part[i].Pos[2] = pos[2] * pos[2] * pos[2] * pos[2]
								* pos[2];
					}
					else
					{
						std::cout << "Problem dim must be 6  --psuffix=p3a3" << std::endl;
						exit(1);
					}
					k++;
				}
				else if (sv[k] == std::string("b3"))
				{
					if (dim == 6)
					{
						for (int j = 0; j < 3; ++j)
							part[i].Pos[p+j] = randn.dev();
						p = p + 3;
						k++;
					}
					else
					{
						std::cout << "Problem dim must be 6  --psuffix=p3a3" << std::endl;
						exit(1);
					}
				}
				else if (sv[k] == std::string("l3"))
				{
					cross_product(pos1,pos1+3,l);
					part[i].Pos[p]=l[0];
					part[i].Pos[p+1]=l[1];
					part[i].Pos[p+2]=l[2];
					p=p+3;
					k++;
				}
				else if (sv[k] == std::string("l2"))
				{
					cross_product(pos1,pos1+3,l);
					part[i].Pos[p]=l[2];
					part[i].Pos[p+1]=sqrt(l[0]*l[0]+l[1]*l[1]+l[2]*l[2]);
					p=p+2;
					k++;
				}
				else if (sv[k] == std::string("lz"))
				{
					cross_product(pos1,pos1+3,l);
					part[i].Pos[p]=l[2];
					p=p+1;
					k++;
				}

			}

		}
	return p;
}

// sanjib 4/28/2010 changed to handle if feh alpha not present
void ParticleData::AddError(const char* fname, int ErrorOption,const char *suff,const char* psuffix1,double mu0,double sigma_mu,double sigma_vlb,double sigma_vr,double sigma_r)
{
	std::vector<float> Mag0, Mag1, feh, alpha,Rad;
	StarParticle Star;
	SurveyError SuE(ErrorOption,sigma_mu,sigma_vlb,sigma_vr,sigma_r);
	std::string psuffix=psuffix1;
	Star.initialize();


	if ((ErrorOption > 0) || (std::string::npos != psuffix.find("feal", 0)))
	{
		std::cout << "adding error" << std::endl;
		if (ebf::ContainsKey(fname,"/FeH") == 1)
			ebf::Read(fname,"/FeH",feh);
		else
		{
			std::cout << "FeH not found" << std::endl;
		}
		if (ebf::ContainsKey(fname,"/Alpha") == 1)
			ebf::Read(fname,"/Alpha",alpha);
		else
		{
			std::cout << "Alpha not found" << std::endl;
		}


//		ebf::EbfFile ebf1;
//		ebf1.open(fname, "r", "/FeH", 4);
//		if (ebf1.status == 1)
//		{
//			feh.resize(ebf1.dim(0));
//			ebf1.sync(&feh[0]);
//		}
//		else
//		{
//			std::cout << "FeH not found" << std::endl;
//		}

//		ebf1.open(fname, "r", "/Alpha", 4);
//		if (ebf1.status == 1)
//		{
//			alpha.resize(ebf1.dim(0));
//			ebf1.sync(&alpha[0]);
//		}
//		else
//		{
//			std::cout << "Alpha not found" << std::endl;
//		}

	}
//	std::cout << "adding error" << std::endl;

	int dim1;
	if(sv.size()==0)
	{
			dim1=dim;
	}
	else
	{
		dim1=6;
		if( (sv.size()==1) && ((sv[0]=="p3")||(sv[0]=="s3")||(sv[0]=="h3")) )
				dim1=3;
		if(dim1==3)
			assert(items>=6);
		if(dim1==6)
			assert(items>=9);
	}

//	assert(items>=6);
//	int dim1=6;
//	if( (sv.size()==1) && ((sv[0]=="p3")||(sv[0]=="s3")||(sv[0]=="h3")) )
//			dim1=3;
//	else if(sv.size()==0)
//		dim1=dim;
//	int dim1=6;
//	if(dim==3)
//		dim1=3;
	Ran rand(1l);
	Rad.resize(nsize);

//	std::cout << "adding error" << std::endl;
	if (ErrorOption > 0)
	{
		assert((dim1==6)||(dim1==3));

//		ExtBinFormat ebf1;

// Sanjib Oct 18 2010 /Mag1 and /Mag0 reversed for new format files
		if (ebf::ContainsKey(fname,"/Mag1") == 1)
			ebf::Read(fname,"/Mag1",Mag0);
		else
		{
			std::cout << "Mag0 not found" << std::endl;
		}
		if (ebf::ContainsKey(fname,"/Mag0") == 1)
			ebf::Read(fname,"/Mag0",Mag1);
		else
		{
			std::cout << "Mag0 not found" << std::endl;
		}

//		ebf1.open(fname, "r", "/Mag1", 4);
//		if (ebf1.status == 1)
//		{
//			Mag0.resize(ebf1.dim(0));
//			ebf1.sync(&Mag0[0]);
//		}
//		else
//		{
//			std::cout << "Mag0 not found" << std::endl;
//			exit(1);
//		}
//
//		ebf1.open(fname, "r", "/Mag0", 4);
//		if (ebf1.status == 1)
//		{
//			Mag1.resize(ebf1.dim(0));
//			ebf1.sync(&Mag1[0]);
//		}
//		else
//		{
//			std::cout << "Mag1 not found" << std::endl;
//			exit(1);
//		}


		for (int i = 0; i < nsize; ++i)
		{
//			std::cout<<"a "<<part[i].Pos[0]<<" "<<part[i].Pos[1]<<" "<<part[i].Pos[2]<<" "<<std::endl;
			for (int j = 0; j < dim1; ++j)
				Star.Pos[j] = part[i].Pos[j];
			Star.mag(0) = Mag0[i];			Star.mag(1) = Mag1[i];
			if(feh.size()>0)
				Star.feh() = feh[i];
			if(alpha.size()>0)
				Star.alpha() = alpha[i];
//			printv(part[i].Pos,part[i].Pos+6);
//			printv(Star.Pos,Star.Pos+6);
			SuE.add(Star);
			for (int j = 0; j < dim1; ++j)
				part[i].Pos[j]=Star.Pos[j];
			Rad[i]= radius3(part[i].Pos);
//			std::cout<<"b "<<part[i].Pos[0]<<" "<<part[i].Pos[1]<<" "<<part[i].Pos[2]<<" "<<std::endl;
			if(feh.size()>0)
				feh[i]=Star.feh();
			if(alpha.size()>0)
				alpha[i]=Star.alpha();
		}

		std::string s(fname);	s.erase(s.begin() + s.find(".ebf"), s.end());	s += suff;	s += "_rad.ebf";
//		ExtBinFormat ebf2;
//		ebf2.open(s.c_str(), "w", "/Rade", 4, Rad.size());
//		ebf2.sync(&Rad[0]);
		ebf::Write(s,"/Rade",&Rad[0],"w","",Rad.size());
	}

	if (sv.size() > 0)
	{
		int p = convert_p3v3(dim1,mu0);

		if (std::string::npos != psuffix.find("feal", 0))
		{
			if((feh.size()==0)||(alpha.size()==0))
			{
				std::cout<<"FeH or Alpha not found"<<std::endl;
				exit(1);
			}
			else
			{
				for (int i = 0; i < nsize; ++i)
				{
					part[i].Pos[p] = feh[i];
					part[i].Pos[p + 1] = alpha[i];
				}
				p = p + 2;
			}
		}
		dim = p;
	}
//if(dim==6)
//{
//	Normaldev randn(0.0,1.0,3l);
//	for (int i = 0; i < nsize; ++i)
//		for (int j = 3; j < 6; ++j)
//			part[i].Pos[j]=(randn.dev()-0.5)*2;
//}
	if(dim<1)
	{
		std::cout<<"Incorrect input tags for data"<<std::endl;
		exit(1);
	}

}




int dims(const char  *s)
{
	if(strcmp(s+1,"3")==0)
		return 3;
	else if	(strcmp(s+1,"2")==0)
		return 2;
	else
		return 1;
}

void ParticleData::initialize(int size1,int option)
{
    setTags(option);
    items=0;
//    for (int ig = 0; ig < tags.size(); ig++)
//    	items+=tags[ig].dim;
    items=tags.back().id+tags.back().dim;
    nsize=size1;
//	items=items1;
//	dim=dim1;
	data= new double[nsize*items];
	part= new Particle[nsize];
	for(int i=0;i<nsize;i++)
			part[i].Pos=&data[i*items];
	for(int64_t i=0;i<int64_t(nsize);i++)
	{
		part[i].ID(items)=i;
		part[i].Density(items)=0.0;
		part[i].Mass(items)=1.0;
	}
	std::cout<<std::setw(12)<<"Data size"<<nsize*items*sizeof(part[0].Pos[0])/1e6<<" MB"<<std::endl;
}


ParticleData::~ParticleData()
{
	delete [] data;
	delete [] part;
}

struct io_header_1
{
    int npart[6];
    double   mass[6];
    double   time;
    double   redshift;
    int flag_sfr;
    int flag_feedback;
    int npartTotal[6];
    int flag_cooling;
    int num_files;
    double   BoxSize;
    double   Omega0;
    double   OmegaLambda;
    double   HubbleParam;
    /* extra flags other than gadget*/
    int flag_id;          /* if IDs needed in output */
    int flag_dim;         /* no of dimensions */
    /* signifies that output file contains estimated density */
    int flag_density;
    int flag_extra;
    char fill[256- 6*4- 6*8- 2*8- 2*4- 6*4- 2*4 - 4*8-4*4];  /* fills to 256 Bytes */
};

void ParticleData::print()
{
	std::cout.precision(6);
	std::cout<<"---------------------------------------------------------------"<<std::endl;
	std::cout<<"Particle Data Details"<<std::endl;
	std::cout<<std::setw(36)<<"Type of particles  \t"<<std::setw(12)<<npart.size()<<std::endl;
	std::cout<<std::setw(36)<<"Total Number of Particles "<<std::setw(12)<<nsize<<std::endl;
	std::cout<<std::setw(12)<<"Dim    "<<std::setw(12)<<dim<<std::endl;
	std::cout<<std::setw(12)<<"Items  "<<std::setw(12)<<items<<" (dims+id+density+mass)"<<std::endl;
	std::cout<<std::setw(16)<<"Dim"<<std::setw(16)<<"Min"<<std::setw(16)<<"Max"<<std::setw(16)<<"Mean"<<std::endl;
	std::cout<<"----------------------------------------------------------------"<<std::endl;
		for(int k=0;k<dim;++k)
		{
			double meanx=0.0, minx=part[0].Pos[k],maxx=part[0].Pos[k];
			for(int i=0;i<nsize;++i)
			{
				if(minx>part[i].Pos[k])
					minx=part[i].Pos[k];
				if(maxx<part[i].Pos[k])
					maxx=part[i].Pos[k];
				meanx+=part[i].Pos[k];
			}
				std::cout<<std::setw(16)<<k<<std::setw(16)<<minx<<std::setw(16)<<maxx<<std::setw(16)<<meanx/nsize<<std::endl;
		}
	std::cout<<"-----------------------------------------------------------------"<<std::endl;

}

void ParticleData::readFromFile(const char* fname,int dim1,const char*s)
{
	dim=dim1;
	int icFormat=1;
    Timer timer1;
    std::cout<<std::left<<std::setw(36)<<"Reading Data from file  "<<fname<<std::endl;
    std::string fs(fname);
    int i=fs.find_last_of(".");
    std::string suffix=fs.substr(i+1);
    std::string fbase=fs.substr(0,i);

    for(unsigned int i=0;i<(strlen(s)/2);++i)
		sv.push_back(std::string(s+2*i,2));
	for(unsigned int i=0;i<sv.size();++i)
		std::cout<<i<<" "<<sv[i]<<std::endl;


	icFormat=0;
//    if(suffix.compare("gadget")==0)
//    	icFormat=1;
    if(suffix.compare("ascii")==0)
    	icFormat=0;
    if(suffix.compare("txt")==0)
    	icFormat=0;
    if(suffix.compare("ebf")==0)
    	icFormat=2;
    if(suffix.compare("ebf2")==0)
    	icFormat=2;
//    if(suffix.compare("hdf5")==0)
//    	icFormat=4;

switch(icFormat)
    {
    	case 0:readAsciiFile(fname);break;
//    	case 1:readGadgetFile(fname);break;
    	case 2:readEbfFile(fname);break;
    	case 3:readCustomFile(fname);break;
//    	case 4:readVLFile(fname);break;
//    	case 5:readHydra1File(fname);break;
//    	case 6:loadpositions(fname);break;
    	default:std::cout<<"File format not recognized"<<std::endl;exit(1);
    }
	npartc.resize(npart.size()+1);
	int temp=0;
	npartc[0]=temp;
	for(size_t i=0;i<npart.size();++i)
	{
		temp+=npart[i];
		npartc[i+1]=temp;
	}
	//	timer1.print("Reading Time");

//	for(int i=0;i<nsize;++i)
//	{
//		std::cout<<i<<part[i].Mass(items)<<std::endl;
//	}
//	exit(1);
	print();
}


void ParticleData::readAsciiFile(const char* fname)
{
	int nsize1,dim1,dim2,scan_status,massid;
	FILE* fd=NULL;
	char st1[256],st2[256];
	if((fd=fopen(fname,"r")))
	{
		std::vector<int> dimid;
		dim2=-1;
		massid=-1;
		scan_status=fscanf(fd,"%s",st1);
		scan_status=fscanf(fd,"%i%i%i%i",&nsize1,&dim1,&massid,&dim2);
		if(scan_status != 4)
		{
			std::cout<<"data header must have at least 4 integers"<<std::endl;
			exit(1);
		}
		std::cout<<"dim1="<<dim1<<" dim2="<<dim2<<" massid="<<massid<<std::endl;
		if(dim2>0)
		{
			if(dim!=dim2)
			{
				std::cout<<"Number of dim in data file does not match the one specified in arguments "<<std::endl;
				exit(1);
			}
		dim=dim2;
		dimid.resize(dim,0);
		for(int i=0;i<dim;++i)
		{
			scan_status=fscanf(fd,"%i",&dimid[i]);
			if(scan_status==0)
			{
				std::cout<<"Input format not correct"<<std::endl;
				exit(1);
			}
		}
		}
		else
		{
			if(dim!=dim1)
			{
				std::cout<<"Number of dim in data file does not match the one specified in arguments "<<std::endl;
				exit(1);
			}
			dim=dim1;
			dimid.resize(dim);
			for(int i=0;i<dim;++i)
				dimid[i]=i;
		}

//		check duplicate dims
		{
			std::vector<int> dimid1=dimid;
			sort(dimid1.begin(),dimid1.end());
//			printv(dimid1.begin(),dimid1.end());
			for(size_t i=1;i<dimid1.size();++i)
				if(dimid1[i]==dimid1[i-1])
				{
					std::cout<<"Input format not correct duplicate dims"<<std::endl;
					exit(1);
				}

		}

		scan_status=fscanf(fd,"%s",st2);
		if((strcmp(st1,st2)!=0)||(massid >= dim1)||((*max_element(dimid.begin(),dimid.end()))>=dim1)||(dim2>dim1))
		{
			if((massid!=-1)&&((dim2+1)>dim1))
			{
				std::cout<<"Input format not correct"<<std::endl;
				exit(1);
			}

			std::cout<<std::endl<<"Input format not correct. Should be"<<std::endl;
			std::cout<<"--------------------------------------------------"<<std::endl;
			std::cout<<"<head>"<<std::endl;
			std::cout<<"rows cols massid dim dim_1 dim_2 ..dim_k"<<std::endl;
			std::cout<<"<head>"<<std::endl;
			std::cout<<"--------------------------------------------------"<<std::endl;
			std::cout<<"k should be equal to dim or dim should be 0 and no dim_i after that"<<std::endl;
			std::cout<<"massid should be an integer or -1. Here it is"<<std::endl;
			std::cout<<nsize1<<" "<<dim1<<" "<<massid<<" "<<dim2<<" "<<(*max_element(dimid.begin(),dimid.end()))<<std::endl;
			exit(1);
		}

		std::cout<<"Dims "<<dim<<" dim_id "<<std::endl;
		printv(dimid.begin(),dimid.end());

		npart.clear();
		if (npart.size() == 0)
			npart.push_back(nsize1);
		initialize(nsize1, 1);

		std::vector<float> tempv(dim1,0.0);
		for(int i=0;i<nsize1;++i)
		{
			for(int k=0;k<dim1;++k)
				scan_status=fscanf(fd,"%f",&tempv[k]);
			for(int k=0;k<dim;++k)
				part[i].Pos[k]=tempv[dimid[k]];
			if(massid>0)
				part[i].Mass(items)=tempv[massid];
//			printv(&part[i].Pos[0],&part[i].Pos[0+dim]);
		}
		fclose(fd);
		if(scan_status!=1)
		{
			std::cout<<"Error reading data from ascii file"<<std::endl;
			exit(1);
		}
	}
	else
	{
		std::cout<<"File not found"<<std::endl;
		exit(1);
	}


}



void ParticleData::readCustomFile(const char* fname)
{

}

void ParticleData::read_tag(const char *fname,ParticleTag &tag, int n_size)
{
//	ExtBinFormat ebf1;
//	ebf1.open(fname, "r", tag.name.c_str(), 4);
	//    		ebf1.initialize(fname,"r","/Pos3",4,0,0);
//	std::cout<<"Reading "<<tag.name<<" ("<<tag.id<<" "<<tag.dim<<" "<<ebf1.dim(0)<<" "<<ebf1.dim(1)<<" "<<ebf1.rank()<<")"<<std::endl;

	if (ebf::ContainsKey(fname,tag.name) == 1)
	{
//		std::cout << ebf1.ebfh.dim[0] << " " << ebf1.ebfh.dim[1] << std::endl;
		ebf::EbfVector<double> x(fname,tag.name);
//		float *x = new float[ebf.elements()];
//		ebf1.read(&x[0], ebf1.dataSize() * xsize * n_size);

		if((x.rank()>2)||(x.rank()<1))
		{
			std::cout<<"Problem: tag dim and data rank mismatch"<<std::endl;
			exit(1);
		}

		if(x.rank()==2)
//		if (tag.dim > 1)
		{
			if(tag.dim>x.dim(1))
			{
				std::cout<<"Problem:tag dim greater than data dim"<<std::endl;
				exit(1);
			}
			std::cout<<"Reading "<<tag.name<<" ("<<tag.id<<" "<<tag.dim<<" "<<x.dim(0)<<" "<<x.dim(1)<<" "<<x.rank()<<")"<<std::endl;
			for (int i = 0; i < n_size; ++i)
				for (int k = 0; k < tag.dim; ++k)
					//   	        for(int k=0;k<ebf1.ebfh.dim[1];++k)
					part[i].Pos[tag.id + k] = x[i * x.dim(1) + k];
		}
		if(x.rank()==1)
		{
			if(tag.dim>1)
			{
				std::cout<<"Problem: tag dim > 1 while data rank=1"<<std::endl;
				exit(1);
			}
			std::cout<<"Reading "<<tag.name<<" ("<<tag.id<<" "<<tag.dim<<" "<<x.dim(0)<<" "<<x.dim(1)<<" "<<x.rank()<<")"<<std::endl;
			for (int i = 0; i < n_size; ++i)
				part[i].Pos[tag.id] = x[i];

		}
	}
	else
	{
		std::cout<<"Item not found "<<tag.name<<" assigning default value of 1.0"<<std::endl;
		if((strncmp(tag.name.c_str(),"/weight",5)!=0))
			exit(1);
	}
//	ebf1.close();
//	std::cout<<"Done"<<std::endl;
}


void ParticleData::setTags(int option)
{
	int c = 0;
	int dtype = 5;
	tags.clear();

	if (sv.size() == 0)
	{
		if (option == 1)
		{
			tags.push_back(ParticleTag("/Pos", dim, c, dtype, 1));
			c += tags.back().dim;
		}
		if (option == 4)
		{
			tags.push_back(ParticleTag("/x", dim, c, dtype, 1));
			c += tags.back().dim;
		}
		if (option == 2)
		{
			if(dim>6)
			{
				std::cout<<"Dimension should be between 1 to 6 for data with tags /Pos3 /Vel3"<<std::endl;
				exit(1);
			}
			tags.push_back(ParticleTag("/Pos3", std::min(dim, 3), c, dtype, 1));
			c += tags.back().dim;
			if (dim > 3)
			{
				tags.push_back(ParticleTag("/Vel3", std::min(dim - 3, 3), c, dtype,1));
				c += tags.back().dim;
			}

		}
		if (option == 3)
		{
			if(dim>6)
			{
				std::cout<<"Dimension should be between 1 to 6 for data with tags /px /vx"<<std::endl;
				exit(1);
			}
			tags.push_back(ParticleTag("/px",1, c, dtype, 1));			c += tags.back().dim;
			tags.push_back(ParticleTag("/py",1, c, dtype, 1));			c += tags.back().dim;
			tags.push_back(ParticleTag("/pz",1, c, dtype, 1));			c += tags.back().dim;
			if (dim > 3)
			{
				tags.push_back(ParticleTag("/vx",1, c, dtype, 1));				c += tags.back().dim;
				tags.push_back(ParticleTag("/vy",1, c, dtype, 1));				c += tags.back().dim;
				tags.push_back(ParticleTag("/vz",1, c, dtype, 1));				c += tags.back().dim;
			}

		}

	}
	else
	{
//		if(sv.size()<2)
//		{
//			std::cout<<"sv size must be greater than 2"<<std::endl;
//			exit(1);
//		}
			if(option==3)
			{
				tags.push_back(ParticleTag("/px",1, c, dtype, 1));				c += tags.back().dim;
				tags.push_back(ParticleTag("/py",1, c, dtype, 1));				c += tags.back().dim;
				tags.push_back(ParticleTag("/pz",1, c, dtype, 1));				c += tags.back().dim;
			}
			else
			{
				tags.push_back(ParticleTag("/Pos3", 3, c, dtype, 1));
				c += tags.back().dim;
			}

			int dim1=6;
			if( (sv.size()==1) && ((sv[0]=="p3")||(sv[0]=="s3")||(sv[0]=="h3")) )
				dim1=3;
			if(dim1==6)
			{
				if(option==3)
				{
					tags.push_back(ParticleTag("/vx",1, c, dtype, 1));				c += tags.back().dim;
					tags.push_back(ParticleTag("/vy",1, c, dtype, 1));				c += tags.back().dim;
					tags.push_back(ParticleTag("/vz",1, c, dtype, 1));				c += tags.back().dim;
				}
				else
				{
					tags.push_back(ParticleTag("/Vel3", 3, c, dtype, 1));
					c += tags.back().dim;
				}
			}


//		c -= 6;
//		c += dims(sv[0].c_str());
//		c += dims(sv[1].c_str());
//		if (find(sv.begin(),sv.end(),std::string("fe")) != sv.end())
//		{
//			std::cout<<"found fe"<<std::endl;
//			tags.push_back(ParticleTag("/FeH",1, c, dtype, 1));
//			c += tags.back().dim;
//		}
//		if (find(sv.begin(),sv.end(),std::string("al")) != sv.end())
//		{
//			std::cout<<"found al"<<std::endl;
//			tags.push_back(ParticleTag("/Alpha", 1, c, dtype, 1));
//			c += tags.back().dim;
//		}
//		if (c < 6)
//			c = 6;
    	dim=0;
    	for(unsigned int i=0;i<sv.size();++i)
    		dim+=dims(sv[i].c_str());

	}
	tags.push_back(ParticleTag("/weight", 1, c, dtype, 1));
	c += tags.back().dim;
	tags.push_back(ParticleTag("/Density", 1, c, dtype, 0));
	c += tags.back().dim;
	tags.push_back(ParticleTag("/ID", 1, c, 3, 0));
	c += tags.back().dim;
}



//void ParticleData::convert(double *PosC)
//{
//	double pos[6],l[3];
//	int p;
//	for (int i = 0; i < nsize; ++i)
//	{
//		p=0;
////		if((sv[0]==std::string("en"))||(sv[1]==std::string("lz"))||(sv[1]==std::string("l3"))||(sv[1]==std::string("vr")))
//		if((sv[0]==std::string("e1"))||(sv[1]==std::string("l1"))||(sv[1]==std::string("l3")))
//		{
//			for (int j = 0; j < 6; ++j)
//				pos[j]=part[i].Pos[j]+PosC[j];
//		}
//		else
//		{
//			for (int j = 0; j < 6; ++j)
//				pos[j]=part[i].Pos[j];
//
//		}
//		if(sv[0]==std::string("p3"))
//		{
//			for (int j = 0; j < 3; ++j)
//				part[i].Pos[p+j]=pos[j];
//			p=p+3;
//		}
//		if(sv[0]==std::string("e1"))
//		{
////			std::cout<<"Option en not allowed"<<std::endl;
////			exit(1);
//			part[i].Pos[p]=BJmodel.energy(pos,pos+3);
//			p=p+1;
//		}
//		if(sv[1]==std::string("v1"))
//		{
//			double r2=(pos[0]*pos[0]+pos[1]*pos[1]+pos[2]*pos[2]);
//			if(r2>0)
//			part[i].Pos[p]=(pos[0]*pos[3]+pos[1]*pos[4]+pos[2]*pos[5])/sqrt(r2);
//			else
//			{
//				part[i].Pos[p]=sqrt(pos[3]*pos[3]+pos[4]*pos[4]+pos[5]*pos[5]);
//				std::cout<<r2<<std::endl;
////				exit(1);
//			}
//			p=p+1;
//		}
//		if(sv[1]==std::string("l3"))
//		{
//			cross_product(pos,pos+3,(part[i].Pos+p));
//			p=p+3;
//		}
//		if(sv[1]==std::string("l1"))
//		{
////			std::cout<<"Option lz not allowed"<<std::endl;
////			exit(1);
//			cross_product(pos,pos+3,l);
//			part[i].Pos[p]=l[2];
//			p=p+1;
//		}
//	}
//}

void ParticleData::readEbfFile(const char* fname)
{
	ebf::EbfDataInfo dinfo;
//	double posC[6];
//	ExtBinFormat ebf1;
//	ebf1.open(fname,"r","/Typelist",2);
	int nsize1=0,option=0;
	npart.clear();
	if(ebf::ContainsKey(fname,"/Typelist")==1)
    {
		ebf::Read(fname,"/Typelist",npart);
//    	npart.resize(ebf1.dim(0));
//    	ebf1.sync(&npart[0]);
    	for(size_t i=0;i<npart.size();++i)
    		nsize1+=npart[i];
    }

	if(ebf::ContainsKey(fname,"/Pos",dinfo)==1)
	  { 
	    option=1;	    
	  }
	else if(ebf::ContainsKey(fname,"/Pos3",dinfo)==1) 
	  {
	    option=2;
	  }
	else if(ebf::ContainsKey(fname,"/px",dinfo)==1)
	  {
	    option=3;
	  }
	else if(ebf::ContainsKey(fname,"/x",dinfo)==1)
	  {
	    option=4;
	    if(dim == -1)
	      {
		std::cout<<"changing dim to "<<dinfo.dim[1]<<" from "<<dim<<std::endl;  
		dim=dinfo.dim[1];
	      }
	  }

//    if(ebf1.status==0)
//    {
//    	ebf1.open(fname,"r","/Pos3",4);
//        if(ebf1.status==1)
//        	option=2;
//    }
//    if(ebf1.status==0)
//    {
//    	ebf1.open(fname,"r","/px",4);
//        if(ebf1.status==1)
//        	option=3;
//    }


    if(option > 0)
    {
    	if(npart.size()==0)
    	{
    		nsize1=dinfo.dim[0];
    		npart.push_back(nsize1);
    	}
    	initialize(nsize1,option);
    }
    else
    {
    	std::cout<<"/Pos, /Pos3 or /px absent"<<std::endl;
    	exit(1);
    }

	for (int j = 0; j < 6; ++j)
		posC[j]=0.0;
	if(ebf::ContainsKey(fname,"/Center")==1)
		ebf::Read(fname,"/Center",&posC[0],6);


//	ebf1.open(fname,"r","/Center",5);
//    if(ebf1.status==1)
//    	ebf1.sync(&posC[0]);
//    else
//    {
//    	ebf1.open(fname,"r","/Center",4);
//    	if(ebf1.status==1)
//    	{
//    		float temp[6];
//    		if(ebf1.status==1)
//    			ebf1.sync(&temp[0]);
//    		for (int j = 0; j < ebf1.dim(0); ++j)
//    			posC[j]=temp[j];
//    	}
//    }
//	ebf1.close();


	std::cout<<"Total tags  "<< tags.size()<< std::endl;
    for (size_t ig = 0; ig < tags.size(); ig++)
    {
//    	std::cout << ig << " " << tags.size() << " " << tags[ig].name << std::endl;
    	if (tags[ig].status == 1)
    		read_tag(fname,tags[ig],nsize);
    }


}


Particle::Particle()
{
}

Particle::~Particle()
{
}



void ParticleData::kdOrder( )
{
	int64_t i,iSrc,iDes;
	char * cSave= new char[sizeof(data[0])*items];
	char * cFirst= (char *)data;
	int64_t memSize=sizeof(data[0])*items;
	Particle* Part=begin();

	for ( i = 0; i < nsize; i++)
	{
		if( ((Part[i].Pos-data)/items)  != i)
		{
			memcpy(cSave,cFirst+(i*memSize),memSize);
			iDes=i;
			do
			{
				iSrc=(Part[iDes].Pos-data)/items;
				if(iSrc==i)
				{
					memcpy(cFirst+(iDes*memSize),cSave,memSize);
					Part[iDes].Pos =data+(iDes*items);
					break;
				}
				memcpy(cFirst+(iDes*memSize),cFirst+(iSrc*memSize),memSize);
				Part[iDes].Pos =data+(iDes*items);
				iDes = iSrc;
			}
			while(1);
		}
	}
	delete []cSave;
}



