//============================================================================
// Author      : Sanjib Sharma
// Version     :
// Copyright   : Copyright (c) Sanjib Sharma 2010
//============================================================================
#include <algorithm>
#include "ParticlePriorityQueue.h"
#include "Mat.h"
#include "Functions.h"


ParticlePriorityQueue::ParticlePriorityQueue( )
{
	ngblist =NULL;
	imarkA=NULL;
	imarkB=NULL;
	metric=NULL;
	searchCenter=NULL;
	searchx=NULL;
	maxSize=0;
	csize=0;
	dim=0;
	pFirst=NULL;
	hi=0;
	hj=0;
	ngbTemp.Pa=NULL;
	hinvd=0.0;
}

ParticlePriorityQueue::~ParticlePriorityQueue()
{
	delete [] ngblist;
	delete [] imarkA;
	delete [] searchCenter;
	delete [] metric;
	for(int j=0;j<dim;++j)
		delete [] searchx[j];
	delete []searchx;
}

void ParticlePriorityQueue::setImarkB(bool *imark)
{
	imarkB=imark;
}

void ParticlePriorityQueue::clearImarkB()
{
	imarkB=NULL;
}

void ParticlePriorityQueue::initialize(Particle* pBegin,Particle* pEnd,int ngbsize1,int dim1)
{
	maxSize=ngbsize1;
	dim= dim1;
	ngblist = new Neighbor[ngbsize1];
	imarkA= new bool[(pEnd-pBegin)];
	searchCenter= new double[dim];
	metric= new double[dim];
	searchx= new double*[dim];
	for(int j=0;j<dim;++j)
		searchx[j]=new double[2];
	csize=0;
	pFirst=pBegin;
	hi=0;
	hj=0;
	ngbTemp.Pa=NULL;
	for(int i=0;i<(pEnd-pBegin);++i)
		imarkA[i]=0;
	for(Particle* Part=pFirst;Part<(pFirst+maxSize);Part++)
	{
		imarkA[Part-pFirst]=1;
		ngblist[csize].Pa=Part;
		ngblist[csize].r=0.0;
		csize++;
	}
	gmatrix.resize(dim,dim);
}

void ParticlePriorityQueue::reset(const double* searchCenter1 ,const double* metric1)
{
	hinvd=1.0;
	for(int j=0;j<dim;++j)
	{
		searchCenter[j]=searchCenter1[j];
		metric[j]=metric1[j];
		hinvd*=metric[j];
	}


	for(Neighbor* ngb=begin();ngb<end();ngb++)
		ngb->r=ngb->Pa->distance(searchCenter,metric,dim);
	std::make_heap(begin(),end());

	double temp=sqrt(top().r);
    for(int j=0; j<dim; ++j)
    {
    	searchx[j][0]=searchCenter[j]-temp/metric[j];
    	searchx[j][1]=searchCenter[j]+temp/metric[j];
    }

}

void ParticlePriorityQueue::reset2(const double* searchCenter1 ,const double* matrixa)
{
	Mat mt(dim,0);
	int i=0;
	for(int k=0; k<dim; k++)
		for(int j=k; j<dim; j++)
		{
			mt.covar[k][j]=matrixa[i];			i++;
			if(k!=j)
				mt.covar[j][k]=mt.covar[k][j];
//			if(k!=j)
//				mt.covar[j][k]=0.0;
//			else
//				mt.covar[j][k]=1.0;
		}

//	for(int k=0; k<dim; k++)
//		for(int j=0; j<dim; j++)
//			cout<<"covar "<<k<<" "<<j<<" "<<mt.covar[k][j]<<endl;

//	mt.covar.print();
	mt.jacobi();
//	mt.v.print();
//	printv(mt.d.begin(),mt.d.end());

//	for(int k=0; k<dim; k++)
//		for(int j=0; j<dim; j++)
//			cout<<"covar "<<k<<" "<<j<<" "<<mt.covar[k][j]<<endl;
//	for(int k=0; k<dim; k++)
//		for(int j=0; j<dim; j++)
//			cout<<"v "<<k<<" "<<j<<" "<<mt.v[k][j]<<endl;
//	for(int j=0; j<dim; j++)
//	cout<<"d= "<<j<<" "<<mt.d[j]<<endl;

	for(int j=0; j<dim; j++)
	{
		mt.d[j]=sqrt(mt.d[j]);
	}
	mt.v.transpose();

//	for(int j=0; j<dim; j++)
//		mt.d[j]=1.0;
//	for(int k=0; k<dim; k++)
//		for(int j=k; j<dim; j++)
//		{
//			if(k!=j)
//				mt.v[j][k]=0.0;
//			else
//				mt.v[j][k]=1.0;
//		}

	for(int k=0; k<dim; k++)
	 {
		 metric[k]=0.0;
		 for(int j=0; j<dim; j++)
			 metric[k]+=mt.v[j][k]*mt.v[j][k]*(mt.d[j]*mt.d[j]);
		 metric[k]=1.0/sqrt(metric[k]);
	 }

	 for(int j=0;j<dim;++j)
		 searchCenter[j]=searchCenter1[j];

		for(int k=0; k<dim; k++)
			for(int j=0; j<dim; j++)
				gmatrix[k][j]=mt.v[k][j]/mt.d[k];

		hinvd=1.0;
		for(int j=0; j<dim; j++)
			hinvd/=mt.d[j];


		for(Neighbor* ngb=begin();ngb<end();ngb++)
		ngb->r=ngb->Pa->distance(searchCenter,gmatrix,dim);
		std::make_heap(begin(),end());

	double temp=sqrt(top().r);
    for(int j=0; j<dim; ++j)
    {
    	searchx[j][0]=searchCenter[j]-temp/metric[j];
    	searchx[j][1]=searchCenter[j]+temp/metric[j];
    }

}
//void ParticlePriorityQueue::check(Particle* Pa1)
//{
//	for(Neighbor* ngb=begin();ngb<end();ngb++)
//	{
//		if(ngb->Pa->ID(items)>12214)
//		{
//			cout<<"Prob "<<Pa1->ID(items)<<endl;
//			exit(1);
//		}
//	}
//
//}

void ParticlePriorityQueue::push1(Particle* Pa1)
{
//	int j;
	if(imarkA[(Pa1-pFirst)]==0)
	{
//			for(j=0; j<dim; ++j)
//			    {
//					if((Pa1->Pos[j])>searchx[j][1])
//			    		break;
//			    	if((Pa1->Pos[j])<searchx[j][0])
//			    		break;
//			    }
//		    if(j==dim)
//	    {
	    	double temp=Pa1->distance_pq(searchCenter,metric,dim,ngblist[0].r);
	    	if(temp<ngblist[0].r)
	    	{
	    		imarkA[(ngblist[0].Pa-pFirst)]=0;
	    		ngblist[0].Pa=Pa1;
	    		ngblist[0].r=temp;
	    		imarkA[(Pa1-pFirst)]=1;

	    		adjust_heap();

	    		temp=sqrt(top().r);
	    	    for(int k=0; k<dim; ++k)
	    	    {
	    	    	searchx[k][0]=searchCenter[k]-temp/metric[k];
	    	    	searchx[k][1]=searchCenter[k]+temp/metric[k];
	    	    }
	    	}
//	    }

	}
}

void ParticlePriorityQueue::push1(Particle* Pa1,double* lp)
{
//	int j;
	if(imarkA[(Pa1-pFirst)]==0)
	{
//			for(j=0; j<dim; ++j)
//			    {
//					if((Pa1->Pos[j])>searchx[j][1])
//			    		break;
//			    	if((Pa1->Pos[j])<searchx[j][0])
//			    		break;
//			    }
//		    if(j==dim)
//	    {
	    	double temp=Pa1->distance_pq(searchCenter,metric,dim,ngblist[0].r,lp);
	    	if(temp<ngblist[0].r)
	    	{
	    		imarkA[(ngblist[0].Pa-pFirst)]=0;
	    		ngblist[0].Pa=Pa1;
	    		ngblist[0].r=temp;
	    		imarkA[(Pa1-pFirst)]=1;

	    		adjust_heap();

	    		temp=sqrt(top().r);
	    	    for(int k=0; k<dim; ++k)
	    	    {
	    	    	searchx[k][0]=searchCenter[k]-temp/metric[k];
	    	    	searchx[k][1]=searchCenter[k]+temp/metric[k];
	    	    }
	    	}
//	    }

	}
}



void ParticlePriorityQueue::push2(Particle* Pa1)
{
	if(imarkA[(Pa1-pFirst)]==0)
	{
	    	double temp=Pa1->distance_pq(searchCenter,gmatrix,dim,ngblist[0].r);
	    	if(temp<ngblist[0].r)
	    	{
	    		imarkA[(ngblist[0].Pa-pFirst)]=0;
	    		ngblist[0].Pa=Pa1;
	    		ngblist[0].r=temp;
	    		imarkA[(Pa1-pFirst)]=1;

	    		adjust_heap();

	    		temp=sqrt(top().r);
	    	    for(int k=0; k<dim; ++k)
	    	    {
	    	    	searchx[k][0]=searchCenter[k]-temp/metric[k];
	    	    	searchx[k][1]=searchCenter[k]+temp/metric[k];
	    	    }
	    	}
	}
}

Particle* ParticlePriorityQueue::pNext()
{
	double rmin;
	Particle *pnext;
//	int i=0;
	rmin=begin()->r;
	pnext=begin()->Pa;
	for(Neighbor* ngb=begin();ngb<end();ngb++)
	{
		if((ngb->r <rmin)&&(imarkB[ngb->Pa-pFirst]!=1))
		{
			rmin=ngb->r;
			pnext=ngb->Pa;
		}
	}
		return pnext;
}

double ParticlePriorityQueue::density( )
{
	double den=0.0;
//	int i=0;
	for(Neighbor* ngb=begin();ngb<end();ngb++)
	{
		den+=(1-ngb->r)*1.0*ngb->Pa->Pos[6];
	}
	std::cout<<den<<" "<<top().r<<std::endl;
	den/=pow(top().r,dim/2.0);
		den*=0.5968;

		return den;
}

void ParticlePriorityQueue::ngbList(int parent,std::vector<int>  &ngbList,std::vector<double> &rList,int items)
{
	int i=0;
//	sort(begin(),end());
	for(Neighbor* ngb=begin();ngb<end();ngb++)
	{
//		cout<<"Here "<<ngb-begin()<<" "<<parent<<" "<<ngb->r<<" "<<ngb->Pa->ID(items)<<" "<<csize<<" "<<i<<endl;
		if(ngb->Pa->ID(items)!=parent)
		{
			ngbList[i]=ngb->Pa->ID(items);
			rList[i++]=ngb->r;
		}
	}
	if((i != (csize-1))&&(parent != -1))
	{
		std::cout<<"Aborting: parent not found"<<std::endl;
		exit(1);
	}
//	exit(1);
}



void ParticlePriorityQueue::print( )
{
	std::cout<<"Printing priority queue"<<std::endl;
	std::cout<<searchCenter[0]<<" "<<searchCenter[1]<<" "<<searchCenter[2]<<" "<<std::endl;
	std::cout<<metric[0]<<" "<<metric[1]<<" "<<metric[2]<<" "<<std::endl;
	for(Neighbor* ngb=begin();ngb<end();ngb++)
		std::cout<<(ngb->Pa)-pFirst<<"  "<<ngb->r<<"  "<<" "<<ngb->Pa->Pos[0]<<"  "<<ngb->Pa->Pos[1]<<"  "<<ngb->Pa->Pos[2]<<"  "<<std::endl;
}





