//============================================================================
// Author      : Sanjib Sharma
// Version     :
// Copyright   : Copyright (c) Sanjib Sharma 2010
//============================================================================


#include <algorithm>
#include "Splitter.h"


static Ran randomS1(17l);

ptrdiff_t random_for_shuffle (ptrdiff_t i) { return randomS1.int32()%i;}

// pointer object to it:
ptrdiff_t (*p_random_for_shuffle)(ptrdiff_t) = random_for_shuffle;



ParticleStats1::ParticleStats1(int dim1)
{
	dim=dim1;
	covar.resize(dim*dim);
	covar2= new double*[dim];
	for(int i=0;i<dim;++i)
		covar2[i]=&covar[i*dim];
	xmin.resize(dim);
	xmax.resize(dim);
	xmean.resize(dim);
	xsig.resize(dim);
}

ParticleStats1::~ParticleStats1( )
{
	delete [] covar2;
}


Splitter::~Splitter( )
{

}
SplitterEnCubic::~SplitterEnCubic( )
{
}
SplitterEn::~SplitterEn( )
{
}
SplitterEnC::~SplitterEnC( )
{
}
SplitterEnSub::~SplitterEnSub( )
{
}
Subspace::~Subspace( )
{
}

//Entropy1d::~Entropy1d( )
//{
//}
//
//
//SplitterEnC::~SplitterEnC( )
//{
//	for (int i=0; i<maxPart; i++)
//		delete [] le[i];
//	delete [] le;
//}

void ParticleStats1::print( )
{
	int j;
	std::cout.precision(4);
	std::cout<<"Covariance Matrix"<<std::endl;
	for(j=0;j<dim;++j)
	{
		for(int k=0;k<dim;++k)
			std::cout<<std::scientific<<std::setw(12)<<covar2[j][k];
		std::cout<<std::endl;
	}
//	std::cout<<std::endl;
//	if(dim==6)
//	std::cout<<en[0]<<" "<<en[1]<<" "<<en[2]<<" "<<en[3]<<" "<<en[4]<<" "<<en[5]<<" "<<std::endl;
	std::cout<<std::setw(12)<<"dim"<<std::setw(12)<<"min"<<std::setw(12) <<"max"<<std::setw(12)<<" mean"<<std::setw(12)<<"sig"<<std::endl;
	for(j=0;j<dim;++j)
		std::cout<<std::scientific<<std::setw(12)<<j<<std::setw(12)<<xmin[j]<<std::setw(12)<<xmax[j]<<std::setw(12)<<xmean[j]<<std::setw(12)<<xsig[j]<<std::setw(12)<<std::endl;

//	std::cout<<std::setw(12)<<"dim"<<std::setw(12)<<"en"<<std::setw(12) <<"enC"<<std::setw(12)<<"enSub "<<std::setw(12)<<"sig"<<std::endl;
//	for(j=0;j<dim;++j)
//		std::cout<<scientific<<std::setw(12)<<j<<std::setw(12)<<en[j]<<std::setw(12)<<enC[j]<<std::setw(12)<<enSub[j]<<std::setw(12)<<xsig[j]<<std::setw(12)<<std::endl;


}


void ParticleStats1::computeMinMaxSig(Particle *pFirst,Particle *pLast,int jbegin,int jend)
{
	int j,count;
	double temp;
	count=(pLast-pFirst);
	for(j=jbegin;j<jend;++j)
	{
		xmin[j]=pFirst->Pos[j];
		xmax[j]=pFirst->Pos[j];
		xmean[j]=pFirst->Pos[j];
		xsig[j]=(pFirst->Pos[j]*pFirst->Pos[j]);
	}


	for (Particle* Pa=pFirst+1; Pa<pLast; ++Pa)
	{
		for(j=jbegin;j<jend;++j)
		{
			temp=Pa->Pos[j];
			if(xmin[j]>temp)
				xmin[j]=(temp);
			if(xmax[j]<temp)
				xmax[j]=(temp);
			xmean[j]+=temp;
			xsig[j]+=(temp*temp);
		}
	}

//	for (int i=1; i<count; ++i)
//	{
//	for(j=jbegin;j<jend;++j)
//	{
//		temp=pFirst[i].Pos[j];
//		if(xmin[j]>temp)
//			xmin[j]=(temp);
//		if(xmax[j]<temp)
//			xmax[j]=(temp);
//		xmean[j]+=temp;
//		xsig[j]+=(temp*temp);
//	}
//}


//	count=(pLast-pFirst);
	for(j=jbegin;j<jend;++j)
	{
		xmean[j]/=count;
		xsig[j]/=count;
		xsig[j]=sqrt(xsig[j]-xmean[j]*xmean[j]);
	}
}

void ParticleStats1::computeCovar(Particle *pFirst,Particle *pLast)
{
	int j,k,count;
	double temp;

	for(j=0;j<dim;++j)
		xmean[j]=0.0;
	for(j=0;j<dim;++j)
		for(k=0;k<dim;++k)
			covar2[j][k]=0.0;


	for (Particle* Pa=pFirst; Pa<pLast; ++Pa)
	{
		for(j=0;j<dim;++j)
		{
			temp=Pa->Pos[j];
			xmean[j]+=temp;
			for(k=j;k<dim;++k)
			{
				covar2[j][k]+=temp*(Pa->Pos[k]);
//				covar[j*dim+k]+=temp*(Pa->Pos[k]);
			}
		}
	}

	count=(pLast-pFirst);
	for(j=0;j<dim;++j)
		xmean[j]/=count;

	for(j=0;j<dim;++j)
		for(k=j;k<dim;++k)
		{
			covar2[j][k]/=count;
			covar2[j][k]-=xmean[j]*xmean[k];
			if(j!=k)
				covar2[k][j]=covar2[j][k];
		}


}

//void ParticleStats1::computeAll(Particle *pFirst,Particle *pLast)
//{
//
//}

void Splitter::computeSplitParameters(Particle *pFirst,Particle *pLast,double* bnd,int prevK)
{
//	std::cout<<pFirst<<" "<<pLast-pFirst<<std::endl;
	int count=pLast-pFirst;
//	if(count>1000)
//		treeType=2;
//	else
//		treeType=1;

	if(treeType==0)
	{
		computeMinMaxSig(pFirst,pLast,0,dim);
//		std::cout<<"Sig "<<xsig[0]<<" "<<xsig[1]<<" "<<xsig[2]<<" "<<xsig[3]<<" "<<xsig[4]<<" "<<xsig[5]<<std::endl;
//		std::cout<<"Sig "<<xmin[0]<<" "<<xmin[1]<<" "<<xmin[2]<<" "<<xmin[3]<<" "<<xmin[4]<<" "<<xmin[5]<<std::endl;
//		std::cout<<"Sig "<<xsig[0]<<" "<<xmax[1]<<" "<<xmax[2]<<" "<<xmax[3]<<" "<<xmax[4]<<" "<<xmax[5]<<std::endl;
//		kSplit=max_element(xsig,xsig+dim)-xsig;
		kSplit=max_element(xsig.begin(),xsig.end())-xsig.begin();
//		kSplit=max_element(xsig.begin()+3,xsig.begin()+6)-xsig.begin();
	}
	if(treeType==1)
	{
		computeMinMaxSig(pFirst,pLast,0,dim);
		if (((pLast-pFirst)>=partBoundary)&&(bnd!=NULL))
				correctBoundary(bnd,count);
			if(prevK >= 0)
//		if(parent!=NULL)
			kSplit=(prevK+1)%dim;
		else
			kSplit=0;
	}

	xSplit=computexSplit(pFirst,pLast);
}

double Splitter::computexSplit(Particle *pFirst,Particle *pLast)
{
	double temp=0.0;
	for (Particle* Pa=pFirst; Pa<pLast; ++Pa)
	{
		temp+=(Pa->Pos[kSplit]);
	}
	temp/=(pLast-pFirst);
	return temp;
}

void Splitter::save(Particle *pFirst,Particle *pLast)
{
	int tempi;
	Pdat.resize((pLast-pFirst)*dim);
	for (Particle* Pa=pFirst; Pa<pLast; ++Pa)
	{
		tempi=dim*Pa->ID(items);
		for(int k=0;k<dim;++k)
			Pdat[tempi+k]=Pa->Pos[k];
	}
}

void Splitter::load(Particle *pFirst, Particle *pLast)
{
	int tempi;
	for (Particle* Pa=pFirst; Pa<pLast; ++Pa)
	{
		tempi=dim*Pa->ID(items);
		for(int k=0;k<dim;++k)
			Pa->Pos[k]=Pdat[tempi+k];
	}
}

//void Splitter::print( )
//{
//	if(en.size()==6)
//	std::cout<<en[0]<<" "<<en[1]<<" "<<en[2]<<" "<<en[3]<<" "<<en[4]<<" "<<en[5]<<" "<<std::endl;
//}

void Splitter::correctParticles(Particle *pFirst,Particle *pLast,double* bnd)

{
	int j=0,i=dim;
	for (int k=0; k<dim; k++)
	{
		if (  fabs((xmax[k]-xmin[k])*2/(xmax[k]+xmin[k])) <epsilon )
			j++;
	}

	if(j==dim)
	{
		std::cout<<"Correcting part"<<std::endl;
		while(i>=dim)
//			i=int(drand48()*dim);
		    i=randomS.int32()%dim;
		pFirst->Pos[i]=bnd[2*i]+(bnd[2*i+1]-bnd[2*i])*0.01;
		pLast->Pos[i]=bnd[2*i+1]-(bnd[2*i+1]-bnd[2*i])*0.01;
	}
}


void Splitter::correctBoundary(double* bnd,int count)
{
	double temp1=25.0;
	double temp2=1.0*temp1/5;
	double temp;
	for (int k=0; k<dim; k++)
	{
		if (  fabs((xmax[k]-xmin[k])*2/(xmax[k]+xmin[k])) >epsilon )
		{
			temp=(xmax[k]-xmin[k])/(count-1);
			if (((bnd[2*k+1]-xmax[k])>(temp2*temp))&&((xmin[k]-bnd[2*k]) > (temp2*temp)))
			{
				bnd[2*k+1]=xmax[k]+0.5*temp;
				bnd[2*k]=xmin[k]-0.5*temp;
			}
			else
			{
				if ((bnd[2*k+1]-xmax[k])>(temp1*temp))
					bnd[2*k+1]=xmax[k]+0.5*temp;
				if ((xmin[k]-bnd[2*k])>(temp1*temp))
					bnd[2*k]=xmin[k]-0.5*temp;
			}
		}
	}
}

void SplitterEnCubic::computeSplitParameters(Particle *pFirst,Particle *pLast,double* bnd,int prevK)
{
//	std::cout<<pFirst<<" "<<pLast-pFirst<<std::endl;
	int count=pLast-pFirst;
	compute3dEntropy(pFirst, pLast, bnd) ;
		{
			int k;
			int jmin=0,jmax=3;
			double temp1=1.0;
			double temp2=1.0;
			for(int j=0;j<3;j++)
			{
			    temp1*=((xmax[j]-xmin[j])/dx[j]);
			    temp2*=((xmax[j+3]-xmin[j+3])/dx[j+3]);
			}
			if(count<=2)
			{
				temp1=2;
				temp2=2;
			}
			if((temp1<1)&&(temp2<1))
			    if(prevK < 3)
			    {
			    	jmin=3;	    	jmax=6;
			    }

			if((temp1>=1)&&(temp2>=1))
			{
			    if(	enSub[0]	==	enSub[1])
			    {
			    	if(prevK < 3)
			    	{
			    		jmin=3;                 jmax=6;
			    	}
			    }
			    else
			    {
				    if(	enSub[1]	<	enSub[0])
			    	{
			    		jmin=3;	    		jmax=6;
			    	}
			    }
			}
			if((temp2>=1)&&(temp1<1))
			{
				jmin=3;	    		jmax=6;
			}
		//	k=max_element(xsig+jmin,xsig+jmax)-xsig;
			k=max_element(xsig.begin()+jmin,xsig.begin()+jmax)-xsig.begin();

			kSplit=k;
		}
		if(count<2)
			kSplit=(prevK+1)%dim;
		xSplit=computexSplit(pFirst,pLast);
}


void SplitterEnCubic::compute3dEntropy(Particle *pFirst, Particle *pLast,double* bnd)
{
//	int *indp=&ind[0]-1;
//	double *xp=&x[0]-1;
	int j, k;
	double temp1;
//	std::vector<double> dx(dim, 0.0);
	int64_t li, enBins1, enSubBins, count;
	count=pLast-pFirst;
	computeMinMaxSig(pFirst, pLast, 0, dim);
//	double epsilon=1e-5;
	if (((pLast-pFirst)>=partBoundary)&&(bnd!=NULL))
			correctBoundary(bnd,count);

	for (j=0; j<dim; j++)
	{
		enSub[j]=0.0;
	}



	enBins1=int(pow(count,0.333333));
	if (enBins1<2)
		enBins1=2;

	if(bnd==NULL)
		for (j=0; j<dim; j++)
		{
			temp1=(xmax[j]-xmin[j])*0.5/(count-1);
			xmax[j]+=temp1;
			xmin[j]-=temp1;
			dx[j]=(xmax[j]-xmin[j])/enBins1;
		}
	else
	for (j=0; j<dim; j++)
	{
//		temp1=(xmax[j]-xmin[j])*0.5/(count-1);
//		xmax[j]+=temp1;
//		xmin[j]-=temp1;
		//		dx[j]=(xmax[j]-xmin[j])/enBins1;
		dx[j]=(bnd[2*j+1]-bnd[2*j])/enBins1;
//		std::cout<<j<<" "<<dx[j]<<bnd[2*j+1]<<" "<<bnd[2*j]<<" "<<enBins1<<std::endl;
	}

	for (j=0; j<dim; j++)
		dx[j]=1/dx[j];

	enSubBins=1;
	for (j=0; j<3; j++)
		enSubBins*=enBins1;

	if (enSubBins>maxPart)
	{
		std::cout<<"enSubBins too large"<<enSubBins<<" maxPart="<<maxPart<<std::endl;
		exit(1);
	}
//	std::cout<<enSubBins<<std::endl;

	int64_t ii=0;
	float tempf;
	// Calculate multi-dim Shannon Entropy "en"
	for (j=0; j<2; j++)
	{
		for (k=0; k<enSubBins; k++)
			rho[k]=0;
			for (Particle* Pa=pFirst; Pa<pLast; Pa++)
			{
				li=0;
				if (j==0)
					for (k=0; k<3; ++k)
					{
						ii=int64_t((Pa->Pos[k]-xmin[k])*dx[k]);
						ii=std::min(ii,enBins1-1);
						li=li*enBins1+ii;
					}
				else
					for (k=3; k<6; ++k)
					{
						ii=int64_t((Pa->Pos[k]-xmin[k])*dx[k]);
						ii=std::min(ii,enBins1-1);
						li=li*enBins1+ii;
					}
				if (li >(enSubBins-1))
				{
//					print();
					std::cout<<"li too large "<<li<<" "<<enSubBins<<" "<<pLast-pFirst<<std::endl;
					for (k=0; k<6; ++k)
						std::cout<<"li too large "<<xmin[k]<<" "<<xmax[k]<<" "<<Pa->Pos[k]<<" "<<enBins1<<" "<<((Pa->Pos[k]-xmin[k])*dx[k])<<" "<<dx[k]<<std::endl;
					exit(1);
				}
				rho[li]++;
			}

		for (k=0; k<enSubBins; k++)
		{
			if (rho[k]>0)
			{
//				rho[k]/=count;
//				if (rho[k]>1)
//					rho[k]=1.0;
//				enSub[j]+=rho[k]*log(rho[k]);
				tempf=float(rho[k])/count;
				if (tempf<1.0)
					enSub[j]+=tempf*log(tempf);
			}
		}
		enSub[j]/=(log(1.0/enSubBins));
	}

}

void SplitterEn::computeSplitParameters(Particle *pFirst,Particle *pLast,double* bnd,int prevK)
{
	int count=pLast-pFirst;
		compute1dEntropy(pFirst,pLast,bnd) ;
		{
			int j,k=0;
			for (j=0; j<dim; j++)
				if(  abs ((xmax[j]-xmin[j])*2/(xmax[j]+xmin[j])) <epsilon )
				{
					en[j]=2.0;
				}
			if((*min_element(en.begin(),en.end()))==2.0)
			{
				std::cout<<"problem Particles having identical co-ordinates "<<std::endl;
				exit(1);
			}
			double emin=*min_element(en.begin(),en.end());
			for (j=1; j<=dim; j++)
			{
				k=(prevK+j)%dim;
				if(en[k]==emin)
					break;
			}
			kSplit=k;
		}


		if(count<2)
			kSplit=(prevK+1)%dim;
		xSplit=computexSplit(pFirst,pLast);
}


void SplitterEnC::computeSplitParameters(Particle *pFirst,Particle *pLast,double* bnd,int prevK)
{
//	std::cout<<pFirst<<" "<<pLast-pFirst<<std::endl;
	status_peak=0;
	int count=pLast-pFirst;

	if (count>=1000)
	{
		computeComplEntropy(pFirst, pLast, bnd) ;
		compute1dEntropy(pFirst,pLast,bnd) ;


		{
			int j,k;
			//	status_peak=0;
			if(status_peak==0)
		{
			for (j=0; j<dim; j++)
				en[j]=0.0;

			for (j=0; j<dim; j++)
				if(  abs ((xmax[j]-xmin[j])*2/(xmax[j]+xmin[j])) <epsilon )
				{
					en[j]=2.0;
				}
		//	if((*min_element(en,en+dim))==2.0)
			if((*min_element(en.begin(),en.end()))==2.0)
			{
				std::cout<<"problem Particles having identical co-ordinates "<<std::endl;
				exit(1);
			}
		//	double emin=*min_element(enC,enC+dim);
		//	for (j=1; j<=dim; j++)
		//	{
		//		k=(prevK+j)%dim;
		//		if(enC[k]==emin)
		//			break;
		//	}
//			int imin;
			std::vector<int> d(dim);
			for (j=0; j<dim; j++)
				d[j]=j;
		//	sort2(dim,(enC-1),(&d[0]-1));
//			sort2(dim,(&enC[0]-1),(&d[0]-1));
			mysorti2(&enC[0],&enC[dim],&d[0]);
		//	if(en[d[0]]<en[d[1]])
				k=d[0];
		//	else
		//		k=d[1];

		}
		else
		{
		//	k=max_element(troughF,troughF+dim)-troughF;
			k=max_element(troughF.begin(),troughF.end())-troughF.begin();
		}
			kSplit=k;
		}
	}
		else
		{
			kSplit=(prevK+1)%dim;
		}

	if(status_peak==0)
	{
		xSplit=computexSplit(pFirst,pLast);
	}
	else
	{
		xSplit=troughX[kSplit];
	}
}


void Splitter::compute1dEntropy(Particle *pFirst,Particle *pLast,double* bnd)
{
	int i,j,k;
	int64_t  enBins,count;
//	std::vector<double> dx(dim,0.0);
	count=pLast-pFirst;
	computeMinMaxSig(pFirst,pLast,0,dim);
	if (((pLast-pFirst)>=partBoundary)&&(bnd!=NULL))
			correctBoundary(bnd,count);



	for (j=0; j<dim; j++)
		en[j]=0.0;

		enBins=count/10;
		if (enBins<3)
			enBins=count;

if(bnd==NULL)
	for (j=0; j<dim; j++)
	{
		double temp1=(xmax[j]-xmin[j])*0.5/(count-1);
		xmax[j]+=temp1;
		xmin[j]-=temp1;
		dx[j]=(xmax[j]-xmin[j])/enBins;
	}
else
for (j=0; j<dim; j++)
		{
	//		temp1=(xmax[j]-xmin[j])*0.5/(count-1);
	//		xmax[j]+=temp1;
	//		xmin[j]-=temp1;
	//		dx[j]=(xmax[j]-xmin[j])/enBins;
			dx[j]=(bnd[2*j+1]-bnd[2*j])/enBins;
		}

		for (j=0; j<dim; j++)
			dx[j]=1/dx[j];



		float tempf;
		std::cout.precision(10);
		// Calculate 1-d Shannon Entropy "en"
	for ( j=0; j<dim; j++)
	{

		for ( k=0; k<enBins; ++k)
			rho[k]=0;

		i=0;
		if(bnd==NULL)
			for (Particle* Pa=pFirst; Pa<pLast; Pa++)
			{
				rho[  std::min(int64_t((Pa->Pos[j]- xmin[j])*dx[j]),enBins-1) ]++;
			}
		else
		for (Particle* Pa=pFirst; Pa<pLast; Pa++)
		{
			rho[ std::min(int64_t((Pa->Pos[j]- bnd[2*j])*dx[j]),enBins-1) ]++;
		}
		for ( k=0; k<enBins; k++)
		{
			if (rho[k]>0)
			{
//				rho[k]/=count;
//				en[j]+=rho[k]*log(rho[k]);
				tempf=float(rho[k])/count;
//				en[j]+=tempf*log(tempf);
				if(tempf<1.0)
					en[j]+=tempf*log(tempf);
			}
		}
		en[j]/=(log(1.0/enBins));
	}
}


void SplitterEnC::computeComplEntropy(Particle *pFirst,Particle *pLast,double* bnd)
{
	int *indp=&ind[0]-1;
	double *xp=&x[0]-1;;
	int i,j,k;
	int64_t enBins1,enSubBins,dcount,count;
	count=pLast-pFirst;
	computeMinMaxSig(pFirst,pLast,0,dim);
	if (((pLast-pFirst)>=partBoundary)&&(bnd!=NULL))
			correctBoundary(bnd,count);

	//	compute1dEntropy(pFirst,pLast);

	for (j=0; j<dim; j++)
	{
		enC[j]=0.0;
		troughX[j]=0.0;
		troughF[j]=0;
	}
	status_peak=0;

	enBins1=int(pow(count,1.0/(dim-1)));
	if (enBins1<2)
		enBins1=2;

	dcount=count/enBins1;
	double dx=enBins1*1.0/count;
//	while ((enBins1*dcount)<count)
//		dcount++;

	enSubBins=1;
	for ( j=0; j<(dim-1); j++)
		enSubBins*=enBins1;
//	std::cout<<enBins1<<" "<<enSubBins<<" "<<dcount<<" "<<count<<std::endl;

	if(enSubBins>maxPart)
	{
		std::cout<<"enSubBins too large"<<enSubBins<<" maxPart="<<maxPart<<std::endl;
		exit(1);
	}
	// Calculate inverse probability transform "en"
	for ( j=0; j<dim; j++)
	{
		i=0;
		x.resize(count);
		for (Particle* Pa=pFirst; Pa<pLast; Pa++)
		{
			ind[i]=i;
			x[i++]=Pa->Pos[j];
		}
//		sort2(count, xp, indp);
		mysorti2(xp,xp+count,indp);
		if((count>50))
		{
//			calculate_peak_split(x,troughF[j],troughX[j],1);
//			if(troughF[j]>0.1)
//				status_peak=1;
		}
		for (i=0; i<count; ++i)
		{
				//        	std::cout<<ii<<" "<<count<<" "<<ind[ii]<<" "<<x[ind[ii]]<<std::endl;
//				le[ind[i]][j]=i/dcount;
				le[ind[i]][j]=int(i*dx);
				if (le[ind[i]][j] >(enBins1-1))
				{
					std::cout<<count<<" "<<enBins1<<" "<<enSubBins
							<<" le too large "
							<<le[ind[i]][j] <<std::endl;
					std::cout<<i<<" "<<dcount<<" "<<enBins1<<"  "
							<<le[ind[i]][j] <<std::endl;
					exit(1);
				}
		}

	}

	float tempf;
	// Calculate multi-dim Shannon Entropy "en"
//	if(status_peak==0)
	for (j=0; j<dim; j++)
	{

		for (k=0; k<enSubBins; k++)
			rho[k]=0;
		int li;
		for (i=0; i<count; i++)
		{
			li=0;
			for (k=0; k<(dim-1); k++)
			{
				li=li*enBins1+le[i][(k+j+1)%dim];
			}
			if (li >(enSubBins-1))
			{
				std::cout<<"li too large"<<li<<std::endl;
				exit(1);
			}
			rho[li]++;
		}

		for (k=0; k<enSubBins; k++)
		{
			if (rho[k]>0)
			{
//				rho[k]/=count;
//				if (rho[k]>1)
//					rho[k]=1.0;
//				enC[j]+=rho[k]*log(rho[k]);
				tempf=float(rho[k])/count;
				if (tempf<1.0)
					enC[j]+=tempf*log(tempf);
//				enC[j]+=tempf*log(tempf);

			}
		}
		enC[j]/=(log(1.0/enSubBins));
		enC[j]=1-enC[j];
	}
//	std::cout<<enC[0]<<" "<<enC[1]<<" "<<enC[2]<<" "<<std::endl;
//exit(1);
}


void SplitterEnC::calculate_peak_split(std::vector<double> &x, double &e, double &ksplit,int option)
{
	int j, ngb, ngb2, mini, mini_last;
	double min, min_last, sig, sig_max;
	std::vector<double> den(x.size(), 0.0);
	std::vector<int> peak;
	ngb=100;
	ngb2=ngb/2;
	sig_max=2.0;

	if (x.size()>10000)
	//	{ngb=10; sig_max=5;}
	{
		ngb=x.size()/100;
		sig_max=5.0;
		ngb2=ngb/2;
	}

	if (x.size()<500)
	//	{ngb=10; sig_max=5;}
	{
		ngb=50;
		sig_max=3.0;
		ngb2=ngb/2;
	}
	if (x.size()<100)
	//	{ngb=10; sig_max=5;}
	{
		ngb=10;
		sig_max=4.0;
		ngb2=ngb/2;
	}
	e=0;
	if(option==0) sort(x.begin(), x.end());
	for (int i=0; i<ngb2; i++)
		den[i]=(i+ngb2)/(x[ngb]-x[0]);
	for (size_t i=x.size()-ngb2; i<x.size(); i++)
		den[i]=(x.size()-i+ngb2)/(x[x.size()]-x[x.size()-ngb]);
	for (size_t i=ngb2; i<(x.size()-ngb2); i++)
	{
		den[i]=(ngb+1)/(x[i+ngb2]-x[i-ngb2]);
//				if((i%100)==0)
//				std::cout<<scientific<<i<<" "<<x.size()<<" "<<den[i]<<" "<<x[i-ngb2]<<" "<<x[i]<<" "<<x[i+ngb2]<<" "<<std::endl;
	}
	//	exit(1);
	//	ngb2=ngb;
	min=den[ngb2];
	min_last=*max_element(den.begin(), den.end());
	mini=ngb2;
	mini_last=ngb2;
	int status;
	for (size_t i=ngb2; i<x.size()-ngb2; i++)
	{
		if (min>den[i])
		{
			min=den[i];
			mini=i;
		}

		status=1;
		for (j=1; j<=ngb2; j++)
			if ((den[i]<den[i-j])||(den[i]<den[i+j]))
			{
				status=0;
				break;
			}

		if (status)
		{
			if (peak.size()==0)
				min=den[i];
			peak.push_back(i);
			//			std::cout<<x[i]<<" "<<i<<" "<<den[i]<<" "<<std::endl;
		}

		if (peak.size()==2)
		{

			if (den[peak[0]]<den[peak[1]])
				sig=(den[peak[0]]/min-1);
			else
				sig=(den[peak[1]]/min-1);
			if ((fabs(peak[0]-mini)<ngb2)||(fabs(peak[1]-mini)<ngb2))
				sig=0.0;

			if ((sig>sig_max)&&(min<min_last))
			{
				min_last=min;
				mini_last=mini;
				e=sig;
				if ((x[mini+1]-x[mini])>(x[mini]-x[mini-1]))
					ksplit=(x[mini]+x[mini+1])*0.5;
				else
					ksplit=(x[mini]+x[mini-1])*0.5;
//								std::cout<<sig<<" check "<<i<<" "<<min<<" "<<den[peak[0]]<<" "<<den[peak[1]]<<" "<<peak[0]<<" "<<peak[1]<<std::endl;

				if (den[peak[0]]<den[peak[1]])
				{
					min=den[peak[1]];
					mini=peak[1];
				}
				else
				{
					min=den[peak[0]];
					mini=peak[0];
				}
			}

			if (den[peak[0]]<den[peak[1]])
				peak[0]=peak[1];
			peak.pop_back();

		}

	}

	ngb=0;
	if (mini_last!=ngb2)
	{
		for (size_t i=0; i<den.size(); i++)
			if (den[i]>min_last)
				ngb++;
		e=(1.0*ngb)/den.size();

	}
	else
		e=0.0;
//	std::cout<<ksplit<<" "<<e<<std::endl;
	//	e=peak.size();
	//	e=0.0;
	//		std::cout<<x.size()<<" check "<<std::endl;

	//	exit(1);
}


//---------------------------------------------------------------

int SplitterEnSub::next_comb(std::vector<int> &comb, int n)
{
	int k=comb.size();
	int i = k - 1;
	++comb[i];
	while ((i >= 1) && (comb[i] >= n - k + 1 + i))
	{
		--i;
		++comb[i];
	}

	if (comb[0] > n - k) /* Combination (n-k, n-k+1, ..., n) reached */
		return 0; /* No more combinations can be generated */

	/* comb now looks like (..., x, n, n, n, ..., n).
	 Turn it into (..., x, x + 1, x + 2, ...) */
	for (i = i + 1; i < k; ++i)
		comb[i] = comb[i - 1] + 1;
	return 1;
}

void SplitterEnSub::generate_subs_k(std::vector<Subspace> &sdv, int n,int kmax)
{
	std::vector<int> comb,comb1;
	comb.resize(kmax);
	for (size_t i = 0; i < comb.size(); ++i)
		comb[i] = i;
	comb1.resize(comb.size()*2);
	for (size_t j = 0; j < comb.size(); ++j)
	{
		comb1[2*j]=comb[j]*2;
		comb1[2*j+1]=comb[j]*2+1;
	}
//	Subspace s(comb1);
//	std::cout<<"here1"<<std::endl;
	sdv.resize(0);
	sdv.push_back(Subspace(comb1));
//	std::cout<<"here2"<<std::endl;
	while(next_comb(comb,n))
	{
		comb1.resize(comb.size()*2);
		for (size_t j = 0; j < comb.size(); ++j)
		{
			comb1[2*j]=comb[j]*2;
			comb1[2*j+1]=comb[j]*2+1;
		}
//		std::cout<<"here3"<<std::endl;
		sdv.push_back(Subspace(comb1));
	}
}


void SplitterEnSub::correct_odd(Subspace &s,int n)
{
	if(s.d.back()==(n-1))
	{
		for(size_t i=1;i<(s.d.size()-1);i=i+2)
		{
			if((s.d[i+1]-s.d[i])>1)
			{
				s.d.back()=s.d[i]+1;
//				std::cout<<"corrected "<<s.d.back()<<" "<<s.d[i]<<" "<<s.d[i+1]<<std::endl;
				break;
			}
		}
	}
	if(s.d.back()==(n-1))
	{
		s.d.back()=s.d.front()-1;
	}
}

void SplitterEnSub::shuffle_subs_all(std::vector<std::vector<Subspace> > &sv,int n)
{
	std::vector<int> d(n,0);
	for(size_t i=0;i<d.size();++i)
		d[i]=i;
	random_shuffle(d.begin(),d.end(),p_random_for_shuffle);
	for(size_t i=0;i<sv.size();++i)
		for(size_t j=0;j<sv[i].size();++j)
			sv[i][j].shuffle(d);

}

void SplitterEnSub::generate_subs_all(std::vector<std::vector<Subspace> > &sv, int n)
{
	int isOdd=0;
	if((n%2)!=0)
	{
		isOdd=1;				n++;
	}
	int kmax=n/2-1;

	for(size_t i=0;i<sv.size();++i)
		sv[i].clear();
	sv.clear();	sv.resize(kmax-1);

	for(int i=2;i<=kmax;++i)
		generate_subs_k(sv[i-2],n/2,i);

	if(isOdd)
	for(size_t i=0;i<sv.size();++i)
		for(size_t j=0;j<sv[i].size();++j)
			correct_odd(sv[i][j],n);
//	for(size_t i=0;i<sv.size();++i)
//		std::cout<<sv.size()<<" "<<sv[i][0].d.size()<<std::endl;
//	exit(1);
}

void SplitterEnSub::compute_enS(Subspace &S,Particle *pFirst,Particle *pLast)
{
//	std::cout<<"compute_enS"<<std::endl;
	float tempf;
	int bins;
	int ebins=1;
	int ib=pFirst-pRoot;
	int ie=pLast-pRoot;
	int count=ie-ib;

//	bins=int64_t(pow(count/10,1.0/(S.d.size())));
//	if(bins<2) bins=2;

	bins=int64_t(pow(count*1.0,1.0/(S.d.size())));
	if(bins>=2)
	{
	for(size_t j=0;j<S.d.size();++j)
	{
		dx[S.d[j]]=1.0/(count*1.0/bins);
		ebins*=bins;
	}
//	std::cout<<count<<" EnS "<<bins<<" "<<ebins<<" "<<ib<<" "<<ie<<std::endl;
//	std::cout<<count<<" "<<ib<<" "<<ie<<" "<<pind.size()<<std::endl;
	S.enS=0.0;
	rho.resize(ebins);
	fill(rho.begin(),rho.end(),0);

//	for(Particle* Pa=pFirst;Pa<pLast;Pa++)
//	{
//				pind[Pa->ID(items)]=0;
//	}
//	for(int j=0;j<S.d.size();++j)
//	{
//		Particle* Pa=&pl[S.d[j]][ib];
//		for(int i=0;i<count;++i)
//			{
//				pind[Pa->ID(items)]=pind[Pa->ID(items)]*bins+int(i*dx[S.d[j]]);
//				Pa++;
//			}
//	}
//	for(Particle* Pa=pFirst;Pa<pLast;Pa++)
//	{
//		if(pind[Pa->ID(items)]>rho.size())
//		{
//			std::cout<<Pa-pFirst<<" Problem "<<pind[Pa->ID(items)]<<" "<<rho.size()<<std::endl;
//			exit(1);
//		}
//		rho[pind[Pa->ID(items)]]++;
//	}


	int tempi;
	for(int i=0;i<count;++i)
	{
		tempi=0;
		for(size_t j=0;j<S.d.size();++j)
			tempi=tempi*bins+int(pln[i*dim+S.d[j]]*dx[S.d[j]]);
		rho[tempi]++;
	}


	for (size_t k=0; k<rho.size(); k++)
	{
		if (rho[k]>0)
		{
			tempf=float(rho[k])/count;
			if (tempf<1.0)
				S.enS+=tempf*log(tempf);
		}
	}
	S.enS/=(log(1.0/rho.size()));
	//changed
	//	S.enS+=pow(10.0,(log10(rho.size())*0.75-log10(count)-0.45))+2.0/count;
	S.enS+=(0.572*rho.size())/(count*log(rho.size()));
//	S.enS=min(S.enS,1.0);
	}
	else
	{
		S.enS=1.0;
	}
//	std::cout<<"compute_enS Done"<<std::endl;
}

void SplitterEnSub::swap_pind(int d_old,int d_new,int count,int bins_n,int ib)
{
	Particle* Pa1=&pl[d_old][ib];
	for(int i=0;i<count;++i)
	{
			pind[Pa1->ID(items)]-=int(i*dx[d_old])*bins_n;
			Pa1++;
	}
	Particle* Pa2=&pl[d_new][ib];
	for(int i=0;i<count;++i)
	{
		pind[Pa2->ID(items)]+=int(i*dx[d_new])*bins_n;
		Pa2++;
	}
//	for(int i=0;i<count;i++)
//	{
//		if(pind1[i]!=pind[i])
//		{
//			std::cout<<i<<" "<<pind[i]<<" "<<pind1[i]<<std::endl;
//			exit(1);
//		}
//	}
}

void SplitterEnSub::compute_enC(Subspace &S,Particle *pFirst,Particle *pLast)
{
//	std::cout<<"compute_enC"<<std::endl;
	float tempf;
	int bins;
	int ebins=1;
	int ib=pFirst-pRoot;
	int ie=pLast-pRoot;
	int count=ie-ib;
//	bins=int64_t(pow(count/10,1.0/(S.d.size()-1)));
//	if(bins<2) bins=2;
	bins=int64_t(pow(count*1.0,1.0/(S.d.size()-1)));

	if((bins>=2)&&(S.d.size()>2))
	{
	for(size_t j=0;j<S.d.size();++j)
	{
		dx[S.d[j]]=1.0/(count*1.0/bins);
		S.enC[j]=0.0;
	}
	for(size_t j=0;j<S.d.size()-1;j++)
	{
//		std::cout<<j<<std::endl;
		ebins*=bins;
	}

	rho.resize(ebins);
//	std::cout<<count<<"EnC "<<bins<<" "<<ebins<<" "<<ib<<" "<<ie<<" "<<S.d.size()<<std::endl;
	int bins_n=1.0;
	for(size_t k=0;k<S.d.size()-1;++k)
		bins_n*=bins;

	for(size_t k=0;k<S.d.size();++k)
	{

		fill(rho.begin(),rho.end(),0);
//		if(k==0)
//		{
//		for(Particle* Pa=pFirst;Pa<pLast;Pa++)
//		{
//			pind[Pa->ID(items)]=0;
//		}
//
//		for(int j=0;j<S.d.size();++j)
//		{
//			if(j!=k)
//			{
//			Particle* Pa=&pl[S.d[j]][ib];
//			for(int i=0;i<count;++i)
//				{
//					pind[Pa->ID(items)]=pind[Pa->ID(items)]*bins+int(i*dx[S.d[j]]);
//					Pa++;
//				}
//			}
//		}
//		}
//		else
//		{
//			swap_pind(S.d[k],S.d[k-1],count,bins_n,ib);
//		}
//		bins_n=bins_n/bins;
//
//
//		for(Particle* Pa=pFirst;Pa<pLast;Pa++)
//		{
//			if(pind[Pa->ID(items)]>rho.size())
//			{
//				std::cout<<Pa-pFirst<<" Problem "<<pind[Pa->ID(items)]<<" "<<rho.size()<<std::endl;
//				exit(1);
//			}
//			rho[pind[Pa->ID(items)]]++;
//		}

		int tempi;
		for(int i=0;i<count;++i)
		{
			tempi=0;
			for(size_t j=0;j<S.d.size();++j)
				if(j!=k)
				tempi=tempi*bins+int(pln[i*dim+S.d[j]]*dx[S.d[j]]);
			rho[tempi]++;
		}


		for (size_t i=0; i<rho.size(); i++)
		{
			if (rho[i]>0)
			{
				tempf=float(rho[i])/count;
				if (tempf<1.0)
					S.enC[k]+=tempf*log(tempf);
			}
		}
		S.enC[k]/=(log(1.0/rho.size()));
		S.enC[k]=1-S.enC[k];

	}
//	sort2(S.d.size(),(&(S.enC[0])-1),(&(S.d[0])-1));
	mysorti2(&S.enC[0],&S.enC[S.d.size()],&S.d[0]);
}
	else
	{
		for(size_t j=0;j<S.d.size();++j)
			S.enC[j]=1.0;

	}
	//std::cout<<"compute_enC Done"<<std::endl;
}


void SplitterEnSub::update_subspace(std::vector<Subspace> &sv,Subspace &S)
{
	int i=sv[0].d.size()-S.d.size();
	if(sv[i]<S)
		sv[i]=S;
}

void SplitterEnSub::calculate_peak_split_all(Particle *pFirst,Particle *pLast,int prevK)
{
	tempx.resize(pLast-pFirst);
	prevK=randomS.int32()%dim;
	for(int j=0;j<dim;++j)
	{
		int j1=(j+prevK+1)%dim;
//		std::cout<<j1<<std::endl;
		for(int i=0;i<(pLast-pFirst);++i)
		{
			tempx[i]=pFirst[i].Pos[j1];
			rho[i]=i;
		}
//		sort2((pLast-pFirst),&tempx[0]-1,&rho[0]-1);
		mysorti2(&tempx[0],&tempx[pLast-pFirst],&rho[0]);
		calculate_peak_split(tempx,j1);
		if(status_peak==1)
		{
			tr[0].print(tempx);
			break;
		}
	}
}

void SplitterEnSub::identifySubspace(Particle *pFirst,Particle *pLast,int prevK)
{
//	std::cout<<pLast-pFirst<<std::endl;
	tempx.resize(pLast-pFirst);
//	if(prevK==(-1))
	    prevK=randomS.int32()%dim;

	for(int j=0;j<dim;++j)
	{
		int j1=(j+prevK+1)%dim;
//		std::cout<<j1<<std::endl;
		for(int i=0;i<(pLast-pFirst);++i)
		{
			tempx[i]=pFirst[i].Pos[j1];
			rho[i]=i;
		}
//		sort2((pLast-pFirst),&tempx[0]-1,&rho[0]-1);
		mysorti2(&tempx[0],&tempx[pLast-pFirst],&rho[0]);
		calculate_peak_split(tempx,j1);
		if(status_peak==1)
		{
			tr[0].print(tempx);
			break;
		}
		for (int i = 0; i < (pLast-pFirst); ++i)
			pln[rho[i] * dim + j1] = i;
	}




	//	std::cout<<"identifySubspace"<<std::endl;
//	if(pl.size()==0)
//	{
//		pl.resize(dim,std::vector<Particle>(pLast-pFirst));
//		for(int j=0;j<dim;++j)
//		{
//			for(int i=0;i<(pLast-pFirst);++i)
//				pl[j][i]=pFirst[i];
//		}
//		for(int j=0;j<dim;++j)
//			sort(pl[j].begin(),pl[j].end(),par_less_than(j));
//		pind.resize(pLast-pFirst);
//	}

	Sv.clear();
	Subspace baseS(dim);
	Sv.push_back(baseS);


	if((status_peak==0)&&(dim>2))
	{
//changed taken out
		//	Sv.clear();
//	Subspace baseS(dim);
//	Sv.push_back(baseS);

	//	baseS.print();
	compute_enS(Sv[0],pFirst,pLast);
	compute_enC(Sv[0],pFirst,pLast);
	for(int j=Sv[0].d.size()-1;j>=2;--j)
	{
		Sv.push_back(Subspace(Sv[0].d.begin(),Sv[0].d.begin()+j));
		compute_enS(Sv.back(),pFirst,pLast);
	}

	int dim1=dim;
	if((dim%2)!=0)
		dim1++;

	if((dim1-2)>=4)
	{
	for(int j=dim1-2;j>=4;j=j-2)
	{
		if((dim1-2)!=(dim-1))
		{
			std::vector<Subspace> Sdv=sAll[j/2-2];
			for(size_t k=0;k<Sdv.size();++k)
				compute_enS(Sdv[k],pFirst,pLast);
			update_subspace(Sv,*min_element(Sdv.begin(),Sdv.end()));
			compute_enC(Sv[dim1-j],pFirst,pLast);
			Sdv.clear();
		}
		for(int k=Sv[dim1-j].d.size()-1;k>=2;--k)
		{
// seems like bug
//			Subspace S1(Subspace(Sv[0].d.begin(),Sv[0].d.begin()+j));
// should be
			Subspace S1(Subspace(Sv[dim1-j].d.begin(),Sv[dim1-j].d.begin()+k));
			compute_enS(S1,pFirst,pLast);
			update_subspace(Sv,S1);
		}
	}
	}
	}
//	std::cout<<"identifySubspace Done"<<std::endl;

}




void SplitterEnSub::computeSplitParameters(Particle *pFirst, Particle *pLast,double* bnd, int prevK)
{
	status_peak = 0;
	int count = pLast - pFirst;
	//	if(pl.size()==0)
	//	{
	//		pl.resize(dim,std::vector<Particle>(pLast-pFirst));
	//		for(int j=0;j<dim;++j)
	//		{
	//			for(int i=0;i<(pLast-pFirst);++i)
	//				pl[j][i]=pFirst[i];
	//		}
	//		for(int j=0;j<dim;++j)
	//			sort(pl[j].begin(),pl[j].end(),par_less_than(j));
	//	}


	if (count >= 2)
	{

		if((count >= 32)&&(dim>1))
		{
			//			calculate_peak_split(pFirst, pLast);
			//			std::cout<<"status peak "<<" "<<status_peak<<std::endl;
			compute1dEntropy(pFirst, pLast, bnd);
			identifySubspace(pFirst, pLast, prevK);
			if (status_peak == 0)
			{
				std::vector<Subspace>::iterator sub = min_element(Sv.begin(),Sv.end());
				compute_enC(*sub,pFirst,pLast);
//				if(count>512)
//				{
//					std::cout<<"Count "<<count<<std::endl;
//					sub->print();
//				}
				//				kSplit = Sv[0].d[0];
				if ((sub->enS >= 1.0)||(dim==2))
				{
//					std::cout<<"kSplit "<<sub->enS<<" "<<dim<<std::endl;
					kSplit = min_element(en.begin(), en.end()) - en.begin();
				}
				else
				{
					kSplit = sub->d[min_element(sub->enC.begin(), sub->enC.end()) - sub->enC.begin()];
//					if((*min_element(sub->enC.begin(), sub->enC.end()))==0.0)
//					{
//					kSplit = sub->d[0];
//					double temp = en[sub->d[0]];
//					for (size_t j = 1; j < sub->d.size(); ++j)
//					{
//						if (temp > en[sub->d[j]])
//						{
//							kSplit = sub->d[j];
//							temp = en[sub->d[j]];
//						}
//					}
//					}

				}
				xSplit = computexSplit(pFirst, pLast);
				//				for (int j = 0; j < Sv.size(); ++j)
				//					Sv[j].print();
			}
			else
			{
				kSplit = tr[0].kSplit;
				xSplit = tr[0].xSplit;
			}
			//			for (int j = 0; j < dim; ++j)
			//				stable_partition(pl[j].begin() + (pFirst - pRoot),
			//						pl[j].begin() + (pLast - pRoot), data_less_than(kSplit,xSplit));

			//			std::cout<<"Compute Split "<<pLast-pFirst<<" "<<pind.size()<<" "<<kSplit<<" "<<xSplit<<std::endl;
		}
		else
		{
			compute1dEntropy(pFirst, pLast, bnd);
			kSplit = min_element(en.begin(), en.end()) - en.begin();
//			kSplit = min_element(en.begin(), en.begin()+3) - en.begin();
//			kSplit = (prevK+1)%dim;
			xSplit = computexSplit(pFirst, pLast);
		}


	}
	else
	{
		if(dim>1)
		{
		computeMinMaxSig(pFirst, pLast, 0, dim);
		if (((pLast - pFirst) >= partBoundary) && (bnd != NULL))
			correctBoundary(bnd, count);
		if (prevK >= 0)
			kSplit = (prevK + 1) % dim;
		else
			kSplit = 0;
		}
		else
			kSplit = 0;

//				if(prevK<0)
//					kSplit=0;
//				else
//					kSplit=(prevK+1)%dim;
//
		xSplit = computexSplit(pFirst, pLast);
	}

//	if(kSplit>2)
//	{
//		compute1dEntropy(pFirst, pLast, bnd);
//	//						std::cout<<"kSplit"<<kSplit<<" "<<count<<std::endl;
//		kSplit = min_element(en.begin(), en.begin()+3) - en.begin();
//	//						std::cout<<"kSplit"<<kSplit<<" "<<count<<std::endl;
//		xSplit = computexSplit(pFirst, pLast);
//
//	}


	//	std::cout<<" "<<kSplit<<" "<<xSplit<<" "<<pLast-pFirst<<std::endl;

}

void Trough::set(int k1,int binsize1,std::vector<double> &tempx)
{
	xSplit=0.0; kSplit=k1; tempx_size=tempx.size();
	statusl=-1;		statusr=-1;		 binsize=binsize1;
	sigl=0.0;		sigr=0.0;			denl1=0.0;		denr1=0.0;
//		countl=0;		countr=0; ibegin=ibegin1; iend=iend1;
}

Trough::Trough()
{

}

Trough::~Trough()
{

}


void SplitterEnSub::calculate_peak_split(std::vector<double> &tempx,int j)
{
	int binsize=25;
//	binsize=10;

//	if(tempx.size()<1000)
//		binsize=10;

	status_peak=0;
	tr[0].set(j,binsize,tempx);
	tr[0].calculate(tempx,status_peak);
//	std::cout<<"Peak "<<j<<" "<<status_peak<<std::endl;
//	std::vector<Trough>:: iterator it=min_element(tr.begin(),tr.end());
}


void Trough::calculate(std::vector<double> &tempx, int &status_peak)
{
	double max_sig=8.0;
	tempx_size=tempx.size();

	if(tempx_size<100000)
		max_sig=7.0;
	if(tempx_size<10000)
		max_sig=6.0;
	if(tempx_size<1000)
		max_sig=5.0;
	if(tempx_size<100)
		max_sig=4.0;

	int i1 = 0;
	double den = 0.0;
	int count_half = tempx.size() / 2;
	for (int i = binsize; i < (count_half - binsize); i = i + 2 * binsize)
	{
		i1 = i;
//		if(tempx[i1 + binsize]>tempx[i1 - binsize])
//			den = binsize * 2 / (tempx[i1 + binsize] - tempx[i1 - binsize]);
//		else
		{
			int binsize1=binsize;
			while(tempx[i1 + binsize1]==tempx[i1 - binsize])
				binsize1*=2;
			den = (binsize+binsize1) / (tempx[i1 + binsize1] - tempx[i1 - binsize]);

		}
		if ((denl1 < den) && (statusl < 1))
		{
			denl1 = den;
			il1 = i1;
			denl0 = den;
			il0 = i1;
			denl2 = den;
			il2 = i1;
		}
		if (denl0 > den)
		{
			denl0 = den;
			il0 = i1;
			denl2 = den;
			il2 = i1;
		}
		if (denl2 < den)
		{
			denl2 = den;
			il2 = i1;
		}
		if (statusl < 0)
		{
			sigl = log(denl1 / denl0) * sqrt(binsize * 2 + 1);
			if (sigl > max_sig)
				statusl++;
		}
		else
		{
			sigl = log(denl2 / denl0) * sqrt(binsize * 2 + 1);
			if (sigl > max_sig)
			{
				statusl++;
				status_peak = 1;
			}
		}

		i1 = tempx.size() - 1 - i;
//		den = binsize * 2 / (tempx[i1 + binsize] - tempx[i1 - binsize]);
		{
			int binsize1=binsize;
			while(tempx[i1 + binsize]==tempx[i1 - binsize1])
				binsize1*=2;
			den = (binsize +binsize1) / (tempx[i1 + binsize] - tempx[i1 - binsize1]);

		}

		if ((denr1 < den) && (statusr < 1))
		{
			denr1 = den;
			ir1 = i1;
			denr0 = den;
			ir0 = i1;
			denr2 = den;
			ir2 = i1;
		}
		if (denr0 > den)
		{
			denr0 = den;
			ir0 = i1;
			denr2 = den;
			ir2 = i1;
		}
		if (denr2 < den)
		{
			denr2 = den;
			ir2 = i1;
		}

		if (statusr < 0)
		{
			sigr = log(denr1 / denr0) * sqrt(binsize * 2 + 1);
			if (sigr > max_sig)
				statusr++;
		}
		else
		{
			sigr = log(denr2 / denr0) * sqrt(binsize * 2 + 1);
			if (sigr > max_sig)
			{
				statusr++;
				status_peak = 1;
			}
		}
		if (status_peak == 1)
			break;
	}

	if(status_peak==0)
	{
		if(denl0>denr0)
		{
			denl0=denr0;
			il0=ir0;
		}
		sigl = log(denl1 / denl0) * sqrt(binsize * 2 + 1);
		sigr = log(denr1 / denl0) * sqrt(binsize * 2 + 1);
		sigl=std::min(sigl,sigr);
		if(sigl>max_sig)
		{
			statusl=1;
			status_peak=1;
		}
	}
	if(status_peak==1)
		xSplit=get_xSplit(tempx);
//	std::cout<<kSplit<<" "<<tempx.front()<<" "<<tempx.back()<<std::endl;

}

double Trough::get_xSplit(std::vector<double> &tempx)
{
	if ((il0 * statusl) > ((tempx_size-ir0) * statusr))
		return (tempx[il0] + tempx[il0 + 1]) / 2.0;
	else
		return (tempx[ir0] + tempx[ir0 + 1]) / 2.0;
}






