//============================================================================
// Author      : Sanjib Sharma
// Version     :
// Copyright   : Copyright (c) Sanjib Sharma 2010
//============================================================================

#include <algorithm>
#include "gfextra.h"
#include "Timer.h"
#include "ebfvector.hpp"


double rtbis(double ( *func)(double ),const double x0, const double x1, const double x2)
{
	double xacc=1e-5;
	const int JMAX=50;
	double dx,xmid,rtb;
	double f=func(x1)-x0;
	double fmid=func(x2)-x0;
	if (f*fmid >= 0.0) throw("Root must be bracketed for bisection in rtbis");
	rtb = f < 0.0 ? (dx=x2-x1,x1) : (dx=x1-x2,x2);
	for (int j=0;j<JMAX;j++)
	{
		fmid=func(xmid=rtb+(dx *= 0.5))-x0;
		if (fmid <= 0.0) rtb=xmid;
		if (abs(dx) < xacc || fmid == 0.0) return rtb;
	}
	throw("Too many bisections in rtbis");
}
float calculate_sigma(int d)
{
	double sigma[]=	{1.0954457, 1.1547005, 1.1952286, 1.2247449, 1.2472191, 1.2649111,
			1.2792043, 1.2909945, 1.3008873, 1.3093074, 1.3165613, 1.3228758,
			1.3284225, 1.3333335, 1.3377124, 1.3416411, 1.3451858, 1.3484001,
			1.3513283, 1.3540070};
	if((d<=20)&&(d>0))
	{
		return sigma[d-1];
	}
	else
	{
		std::cout<<"value of dimension d not in range 1-20"<<std::endl;
		exit(1);
	}
}
void load_table_fm(std::vector<float> &tb_fm,std::vector<float> &tb_cr)
{
    tb_fm.resize(13);
    tb_cr.resize(13);
    tb_fm[0]=-6;    tb_cr[0]=0.597059;
    tb_fm[1]=-5.7;  tb_cr[1]=0.624571;
    tb_fm[2]=-5;    tb_cr[2]=0.693416;
    tb_fm[3]=-4.7;  tb_cr[3]=0.725328;
    tb_fm[4]=-4;    tb_cr[4]=0.804926 ;
    tb_fm[5]=-3.7;  tb_cr[5]=0.841510 ;
    tb_fm[6]=-3;    tb_cr[6]=0.929884 ;
    tb_fm[7]=-2.7;  tb_cr[7]=0.967204;
    tb_fm[8]=-2;    tb_cr[8]=1.03405;
    tb_fm[8]=-1.7;  tb_cr[9]=1.04632;
    tb_fm[10]=-1;   tb_cr[10]=1.05000;
    tb_fm[11]=-0.7; tb_cr[11]=1.05000;
    tb_fm[12]=0.0;  tb_cr[12]=1.05000;
}


void load_plist(const char *buf, std::vector<class Plist*> &pl,
		std::vector<class Plist> &plv, std::vector<class GParticle> &Par)
{
	Timer timer1;
	std::cout << "Reading Den,NgbList" <<std::endl;

	{
		ebf::EbfVector<float> tempx(buf,"/DENSITY");
		Par.clear();
		Par.resize(tempx.size());
		for(size_t i=0;i<tempx.size();++i)
			Par[i].Density=tempx[i];
	}

	{
		ebf::EbfVector<float> tempx(buf,"/MASS");
		for(size_t i=0;i<tempx.size();++i)
			Par[i].mass()=tempx[i];
	}

	plv.clear();
	plv.resize(Par.size());

	pl.clear();
	pl.resize(Par.size());


	ebf::EbfFile ebf_density;
	ebf_density.Open(buf,"/NGB_LIST");

	for (int i = 0; i < ebf_density.dim(0); i++)
		plv[i].nlist.resize(ebf_density.dim(1) - 1);

	int Debug=1;
	if (Debug)
		std::cout << std::fixed << "Memory for plv\t\t\t" << "\t" << std::setw(12)
				<< plv.size() * sizeof(plv[0]) / (1024 * 1024.0) << "MB" <<std::endl;
	if (Debug)
		std::cout << std::fixed << "Memory for pl\t\t\t" << "\t" << std::setw(12) << pl.size()
				* sizeof(pl[0]) / (1024 * 1024.0) << "MB" <<std::endl;
	if (Debug)
		std::cout << std::fixed << "Memory for Par\t\t\t" << "\t" << std::setw(12)
				<< Par.size() * sizeof(Par[0]) / (1024 * 1024.0) << "MB" <<std::endl;
	if (Debug)
		std::cout << std::fixed << "Memory for nlist\t\t" << "\t" << std::setw(12)
				<< plv.size() * plv[0].nlist.capacity()
						* sizeof(plv[0].nlist[0]) / (1024 * 1024) << "MB"
				<<std::endl;


	for (int i = 0; i < ebf_density.dim(0); i++)
	{
		pl[i] = &plv[i];
		ebf_density.Read(&(plv[i].pi),1);
		Par[plv[i].pi].ID = 0;
		//	Par[plv[i].pi].n=0;
		ebf_density.Read(&(plv[i].nlist[0]), (ebf_density.dim(1) - 1));
	}
	ebf_density.Close();

}

//bool cmp1( class Plist *a, class Plist *b )
//{  return Par[a->pi].Density > Par[b->pi].Density; }


class cmp_plist : public std::binary_function<class Plist *,class Plist *,bool>
{
	GParticle* Par1;
public:
	explicit cmp_plist(GParticle* Par) :Par1(Par) { }
	bool operator()(const class Plist *a, const class Plist *b ) const
	{  return Par1[a->pi].Density > Par1[b->pi].Density; }
};

void sort_plist(const char *filename)
{
	std::vector<class Plist*> pl;
	std::vector<class Plist> plv;
	std::vector<GParticle> Par;
	load_plist(filename, pl, plv, Par);
	std::cout << "Sorting Par according to density  ....";
	fflush(stdout);
	//    Timer t1;
	//    t1.start();
	//	#pragma omp parallel
	//    {
	//        int tnum=omp_get_num_threads();
	//        int tid=omp_get_thread_num();
	//    	std::vector<class Plist*>::iterator pBegin=pl.begin()+tid*(pl.size()/tnum);
	//    	std::vector<class Plist*>::iterator pEnd=pl.begin()+(tid+1)*(pl.size()/tnum);
	//    	if((tid+1)==tnum)
	//    		pEnd=pl.end();
	//    	sort(pBegin,pEnd,cmp1);
	//    }
	//    t1.print("time1 ");
	//    t1.start();
	//	#pragma omp parallel
	//    {
	//    	int tnum=omp_get_num_threads();
	//    	int tid=omp_get_thread_num();
	//    	if(tid==0)
	//    		for(int i=1;i<tnum;++i)
	//    		{
	//    			std::vector<class Plist*>::iterator pBegin=pl.begin();
	//    			std::vector<class Plist*>::iterator pMiddle=pl.begin()+(i)*(pl.size()/tnum);
	//    			std::vector<class Plist*>::iterator pEnd=pl.begin()+(i+1)*(pl.size()/tnum);
	//    			if((i+1)==tnum)
	//    				pEnd=pl.end();
	//    			inplace_merge(pBegin,pMiddle,pEnd,cmp1);
	//    		}
	//    }
	//    t1.printC("time2 ");

//	sort(pl.begin(), pl.end(), cmp1());
	sort(pl.begin(), pl.end(), cmp_plist(&Par[0]));
	write_plist(filename, pl, plv, Par);
//	std::cout << "Done" <<std::endl;
}

void load_Par(const char *buf, std::vector<GParticle> &Par, int &NumPart,std::vector<int> &nx,std::vector<int> &sortedList)
{
	Par.clear();
	std::string sbuf=buf;
//	ExtBinFormat ebf_density;

//	ebf_density.open(buf, "r", "/Voxels", 4);
//	NumPart=1;
//	nx.resize(ebf_density.rank());
//	for(int i=0;i<ebf_density.rank();++i)
//	{
//			NumPart*=ebf_density.dim(i);
//			nx[ebf_density.rank()-1-i]=ebf_density.dim(i);
//	}
//	Par.resize(NumPart);
//	ebf_density.sync(&Par[0].Density, sizeof(Par[0].Density),sizeof(Par[0]));

	{
		ebf::EbfVector<float> tempx(sbuf, "r", "/Voxels");
		Par.resize(tempx.size());
		for(size_t i=0;i<tempx.size();++i)
			Par[i].Density=tempx[i];
	}


		for (size_t i = 0; i < Par.size(); i++)
			Par[i].ID = 0;
		for (size_t i = 0; i < Par.size(); i++)
			Par[i].mass() = 1.0;
		for (size_t i = 0; i < Par.size(); i++)
			if(Par[i].Density <= 0)
				Par[i].Density=1e-6;

		sortedList.resize(Par.size(),0);
//		ebf_density.open(buf, "r", "/SortedList", 2);
//		ebf_density.sync(&sortedList[0]);
		//		ebf_density.close();
		ebf::Read(buf,"/SortedList",&sortedList[0],Par.size());


}


void load_plist1(const char *buf, std::vector<class Plist> &plv, std::vector<
		GParticle> &Par, int &nmerge, int &NumPart, int &nc, FILE* &fd)
{
	std::string sbuf=buf;
	//	int blklen, i,
	int j;
	//	int status = 0, flag_swap = 0;
//	int patterns[] = { 4, 1, 0 }, 	patternh[] ={ 4, 1, 0 };
	//    double t0=second();
	int imax;
	if (nc == 0)
	{
		//	sprintf(buf,"%s%s%s",AllG.InitCondFile,AllG.DeSuffix,".ebf");
		//    std::cout<<buf<<std::endl;
//		ebf_density.open(buf, "r", "/DENSITY", 4);

		//	ebf_header ebfh=ebf_read_header("/DENSITY",4,buf,status);

		{
			ebf::EbfVector<double> tempx(sbuf,"/DENSITY","r");
			Par.clear();
			Par.resize(tempx.size());
			for(size_t i=0;i<tempx.size();++i)
				Par[i].Density=tempx[i];
		}
		NumPart = Par.size();
		plv.clear();
		// choose betwween size gt 50
		if (NumPart < 200000)
			plv.resize(NumPart);
		else
			plv.resize(100000);

		//	plv.resize(50000);


		{
			ebf::EbfVector<double> tempx(sbuf,"/MASS");
			for(size_t i=0;i<tempx.size();++i)
				Par[i].Mass=tempx[i];
		}

		ebf::EbfFile ebf_density;
		ebf_density.Open(buf,"/NGB_LIST");
		if(ebf_density.ecode!=0)
		{
			std::cout<<buf<<":/NGB_LIST not found, run with -dl option"<<std::endl;
			exit(1);
		}

		//    ebfh=ebf_read_header("/DENSITY",4,buf,status);
		//	if(status)
		//	    ebf_rw(buf,"r",&Par[0].Density,sizeof(Par[0].Density),sizeof(Par[0]),ebfh);
		//
		//	ebfh=ebf_read_header("/MASS",4,buf,status);
		//	if(status)
		//	    ebf_rw(buf,"r",&Par[0].Mass,sizeof(Par[0].Mass),sizeof(Par[0]),ebfh);


		//	ebfh=ebf_read_header("/NGB_LIST",2,buf,status);
		//	if(status)
		//	    fd=ebf_open(buf,"r",ebfh,flag_swap);
		//		fd = ebf_density.fp();


		for (size_t i = 0; i < plv.size(); i++)
		{
			plv[i].nlist.resize(ebf_density.dim(1) - 1);
		}

		for (size_t i = 0; i < Par.size(); i++)
			Par[i].ID = 0;
		ebf_density.Close();
	}


	ebf::EbfFile ebf_density;
	ebf_density.Open(buf,"/NGB_LIST");


	nmerge = ebf_density.dim(1) - 1;

	imax = nc + plv.size();
	if (imax > NumPart)
		imax = NumPart;

//	int64_t offset1=nc * ebf_density.dim(1);

	j = 0;
//	std::cout<<"nc= "<<nc<<" "<<imax<<" "<<nmerge<<" "<<offset1<<" "<<ebf_density.getfpos()<<" "<<ebf_density.dataSize()*ebf_density.dim(0)*ebf_density.dim(1)<<std::endl;
//	std::cout<<nc * ebf_density.dataSize()*ebf_density.dim(1)<<" "<<ebf_density.dataSize()*ebf_density.dim(1)<<std::endl;
	ebf_density.Seek(nc * ebf_density.dim(1));
	for (int i = nc; i < imax; i++)
	{
//		std::cout<<i<<std::endl;
		ebf_density.Read(&plv[j].pi,1);
		ebf_density.Read(&(plv[j].nlist[0]),nmerge);
//		ebf_density.random_read(&plv[j].pi,4,offset1+(nmerge+1)*j*ebf_density.dataSize());
//		ebf_density.random_read(&(plv[j].nlist[0]),4*nmerge,(offset1+((nmerge+1)*j+1)*ebf_density.dataSize()));
		j++;
	}

	nc = imax;
	//	if (nc >= NumPart)
	ebf_density.Close();
	//    	fclose(fd);
}


void write_plist(const char *buf, std::vector<class Plist*> &pl,
		std::vector<class Plist> &plv, std::vector<class GParticle> &Par)
{
	//#define BLKLEN my_fwrite1(&blklen,patterns,1,AllG.flag_swap,fd);
	//    int   blklen,n_tot,n_merge,i,j;

	std::cout << "Writing Den,NgbList " <<std::endl;

	//    sprintf(buf,"%s%s%s",AllG.InitCondFile,AllG.DeSuffix,".ebf");
	//    std::cout<<buf<<std::endl;

	ebf::EbfFile ebf_density;
	ebf_density.Open(buf,"/DENSITY", "w",4);
	for(size_t i=0;i<Par.size();++i)
		ebf_density.Write(&(Par[i].Density),1);
	ebf_density.Close();


	ebf_density.Open(buf,"/MASS", "a",4);
	for(size_t i=0;i<Par.size();++i)
		ebf_density.Write(&(Par[i].mass()),1);
	ebf_density.Close();

	//    ebfh=ebf_create_header("/MASS",4,4,1);    ebfh.dim[0]=int(Par.size());
	//    ebf_rw(buf,"a",&Par[0].Mass,sizeof(Par[0].Mass),sizeof(Par[0]),ebfh);
	//    ebfh=ebf_create_header("/NGB_LIST",2,4,2);    ebfh.dim[0]=int(Par.size()); ebfh.dim[1]=plv[0].nlist.size()+1;
	//    fd=ebf_open(buf,"a",ebfh,flag_swap);

	std::vector<int64_t> dims;
	dims.push_back(Par.size());
	dims.push_back(plv[0].nlist.size() + 1);
	ebf_density.Open(buf, "/NGB_LIST","a", 2, "",2,&dims[0]);
	for (size_t i = 0; i < Par.size(); i++)
	{
		ebf_density.Write(&(pl[i]->pi),1);
		ebf_density.Write(&(pl[i]->nlist[0]),plv[0].nlist.size());
	}

	ebf_density.Close();

	//    ebf_close(fd,ebfh);
	//    ebf_density.ebf_close(fd,ebfh);


	for (int i = 0; i < int(plv.size()); i++)
		plv[i].nlist.clear();

	plv.clear();
	pl.clear();
	Par.clear();

}





